/*
 * arch/arm/mach-tcc92x/clock.c
 *
 * Clock driver for TCC92xx
 *
 * Copyright (C) 2009 Telechips, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/list.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/spinlock.h>

#include <mach/clock.h>
#include <mach/globals.h>
#include <mach/tca_ckc.h>
#include <linux/delay.h>

#if 1
#define dprintk(msg...) pr_debug(msg)
#else
#define dprintk(msg...)	pr_info(msg)
#endif

struct freqclk_table {
	unsigned int freq;
	int clk_src;
};


#define CALLED_BY_CLOCK_C_FILE
#if defined(CONFIG_DRAM_MDDR)
#include "tcc_clocktbl_lpddr.h"
#elif defined(CONFIG_DRAM_DDR2)
#include "tcc_clocktbl_ddr2.h"
#else
#error Select the DDR type
#endif

#define NUM_FREQS ARRAY_SIZE(clk_table)


void tcc_cpufreq_set_voltage_limit(int uV, int type);

static LIST_HEAD(clock_list);
static DEFINE_SPINLOCK(clock_lock);
static DEFINE_MUTEX(clock_mutex);

static int _clk_set_rate(struct clk *clk, unsigned long rate);

/* XIN */
static struct clk clk_xin = {
	.name = "xin",
        .rate = 12000,
};

/* PLL0 */
static int pll0_clk_set_rate(struct clk *clk, unsigned long rate)
{
        if (clk->rate == rate)
                return 0;

        dprintk("clock: set PLL0 clock to %ldMHz\n", rate/1000000);

	tca_ckc_setpll(rate/100, 0);
	return 0;
}

static struct clk clk_pll0 = {
	.name = "pll0",
	.flags = CLK_ALWAYS_ENABLED,
	.set_rate = pll0_clk_set_rate,
};

/* PLL1 */
static struct clk clk_pll1 = {
	.name = "pll1",
	.flags = CLK_ALWAYS_ENABLED,
};

/* PLL2 */
static struct clk clk_pll2 = {
	.name = "pll2",
	.flags = CLK_ALWAYS_ENABLED,
};

/* PLL3 */
static struct clk clk_pll3 = {
	.name = "pll3",
	.flags = CLK_ALWAYS_ENABLED,
};

/* CPU clock */
static unsigned long cpu_clk_get_rate(struct clk *clk)
{
	return tca_ckc_getcpu() * 100;
}

static int cpu_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int n;
	unsigned long pll;

	if (clk->rate == rate)
		return 0;

	dprintk("clock: set CPU frequency to %ldMHz\n", rate / 1000000);
	pll = tca_ckc_getpll(0);

	//dprintk("clock: pll =  %ldhz, rate = %d\n", pll, rate);

	rate /= 100;
	n = (rate*16) / pll;
	if (!n) ++n;	// n must be greater than 0

	//dprintk("clock: n =  %d\n", n);

	tca_ckc_setcpu(n);
	return 0;
}

static struct clk clk_cpu = {
	.name = "cpu",
	.parent = &clk_pll0,
	.get_rate = cpu_clk_get_rate,
	.set_rate = cpu_clk_set_rate,
};

/* Memory bus clock */
static int membus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	struct clk *pclk;

	list_for_each_entry(pclk, &clock_list, list) {
		if (pclk->parent == clk && pclk->rate > rate) {
                rate = pclk->rate;
		}
	}
	if (clk->rate != rate) {
		tca_ckc_set_mem_freq(rate / 1000);
		dprintk("clock: set memory bus clock to %ldMHz\n", rate / 1000000);
	}
	clk->rate = rate;
	return 0;
}

static struct clk clk_membus = {
	.name = "membus",
	.parent = &clk_pll3,
	.set_rate = membus_clk_set_rate,
};

#if 1 //GRP-BUS
static int gpu_clk_enable(struct clk *clk)
{
	int ret;

	ret = tca_ckc_setpmupwroff(PMU_GRAPHICBUS, 1);
	tca_ckc_enable(CLKCTRL3, 1);

	if(ret == 1)
	{
		tca_ckc_setgrppwdn(0, 0, 1);
	}
	
	return 0;
}

static void gpu_clk_disable(struct clk *clk)
{
	tca_ckc_enable(CLKCTRL3, 0);
	tca_ckc_setpmupwroff(PMU_GRAPHICBUS, 0);
}

/* GRP bus clock */
static int gpubus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int i, matched = 0;
	struct clk *pclk;
	
	list_for_each_entry(pclk, &clock_list, list) {
		if (pclk->parent == clk && pclk->rate > rate) {
                rate = pclk->rate;
		}
	}
	
	for(i = 0; i < NUM_FREQS ; i++)
	{
		if(rate <= clk_table[i].freq)
		{
			rate = clk_table[i].freq;
			matched = 1;
			break;
		}
	}	

	if (clk->rate != rate && matched/* && clk->usecount > 0*/) 
	{		
		tca_ckc_setfbusctrl(CLKCTRL3, NOCHANGE, NORMAL_MD, clk_table[i].freq, clk_table[i].clk_src); /*FBUS_GRP*/
		clk->rate = rate;

		dprintk("clock: set gpu bus clock to %ldMHz\n", rate / 10000);		
	}
	else
	{
		return 1;
	}

	return 0;
}	

static struct clk clk_gpu = {
	.name = "gpu",
	.parent = &clk_pll2,
	.flags = CLK_AUTO_OFF,	
	.enable = gpu_clk_enable,
	.disable = gpu_clk_disable,
	.set_rate = gpubus_clk_set_rate,
};
#endif

#if 1 //VPU-BUS
static int vpubus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int i, matched = 0;
	struct clk *pclk;
	
	list_for_each_entry(pclk, &clock_list, list) {
		if (pclk->parent == clk && pclk->rate > rate) {
				rate = pclk->rate;
		}
	}

	for(i = 0; i < NUM_FREQS ; i++)
	{
		if(rate <= clk_table[i].freq)
		{
			rate = clk_table[i].freq;
			matched = 1;
			break;
		}
	}	

	if (clk->rate != rate && matched /*&& clk->usecount > 0 && (rate > 60000)*/) 
	{
		tca_ckc_setfbusctrl(CLKCTRL5, NOCHANGE, NORMAL_MD, clk_table[i].freq, clk_table[i].clk_src); /*FBUS_VBUS*/
		
		clk->rate = rate;
		dprintk("clock: set vpu bus clock to %ldMHz\n", rate / 10000);		
	}
	else
	{
		return 1;
	}

	return 0;
}	

/* Bus clock for video processing */
static int vbus_clk_enable(struct clk *clk)
{
	int ret;

	ret = tca_ckc_setpmupwroff(PMU_VIDEOBUS, 1);
	mdelay(1);
	tca_ckc_enable(CLKCTRL5, 1);
	if(ret == 1)
	{
		mdelay(1);
		tca_ckc_setvbuspwdn(0, 0, 1);
	}

	return 0;
}

static void vbus_clk_disable(struct clk *clk)
{
	tca_ckc_setvbuspwdn(0, 1, 1);
	mdelay(1);

	tca_ckc_enable(CLKCTRL5, 0);
	tca_ckc_setpmupwroff(PMU_VIDEOBUS, 0);
}

static struct clk clk_vbus = {
	.name = "vbus",
	.parent = &clk_pll2,
	.flags = CLK_AUTO_OFF,	
	.enable = vbus_clk_enable,
	.disable = vbus_clk_disable,
	.set_rate = vpubus_clk_set_rate,	
};

/* VIDEO-CODEC pwdn in VIDEO-BUS */
static int vpu_vcache_clk_set_rate(struct clk *clk, unsigned long rate)
{
	clk->rate = rate;
	if(rate == 0)
		return 0;
	
	return _clk_set_rate(clk->parent, rate);
}

static int vcache_clk_enable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_VIDEOCACHE, 1, 0);
	return 0;
}

static void vcache_clk_disable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_VIDEOCACHE, 0, 0);
}

static struct clk clk_vcache = {
	.name = "vcache",
    .parent = &clk_vbus,
    .flags = CLK_AUTO_OFF,
    .enable = vcache_clk_enable,
    .disable = vcache_clk_disable,
    .set_rate = vpu_vcache_clk_set_rate,
};

/* VIDEO-CODEC pwdn in VIDEO-BUS */
static int vpu_vcodec_clk_set_rate(struct clk *clk, unsigned long rate)
{
	clk->rate = rate;
	if(rate == 0)
		return 0;
	
	return _clk_set_rate(clk->parent, rate);
}

static int vcodec_clk_enable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_VIDEOCODEC, 1, 0);
	return 0;
}

static void vcodec_clk_disable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_VIDEOCODEC, 0, 0);
}

static struct clk clk_vcodec = {
	.name = "vcodec",
    .parent = &clk_vbus,
    .flags = CLK_AUTO_OFF,
    .enable = vcodec_clk_enable,
    .disable = vcodec_clk_disable,
	.set_rate = vpu_vcodec_clk_set_rate,
};


/* JPEG Decode pwdn in VIDEO-BUS */
static int vpu_jpegd_clk_set_rate(struct clk *clk, unsigned long rate)
{
	clk->rate = rate;
	if(rate == 0)
		return 0;
	
	return _clk_set_rate(clk->parent, rate);
}

static int jpegd_clk_enable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_JPEGDECODE, 1, 0);
	return 0;
}

static void jpegd_clk_disable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_JPEGDECODE, 0, 0);
}

static struct clk clk_jpegd = {
	.name = "jpegd",
    .parent = &clk_vbus,
    .flags = CLK_AUTO_OFF,
    .enable = jpegd_clk_enable,
    .disable = jpegd_clk_disable,
	.set_rate = vpu_jpegd_clk_set_rate,    
};

/* JPEG Encode pwdn in VIDEO-BUS */
static int vpu_jpege_clk_set_rate(struct clk *clk, unsigned long rate)
{
	clk->rate = rate;
	if(rate == 0)
		return 0;
	
	return _clk_set_rate(clk->parent, rate);
}

static int jpege_clk_enable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_JPEGENCODE, 1, 0);
	return 0;
}

static void jpege_clk_disable(struct clk *clk)
{
    tca_ckc_setvbuspwdn(VBUSPWDN_JPEGENCODE, 0, 0);
}

static struct clk clk_jpege = {
	.name = "jpege",
    .parent = &clk_vbus,
    .flags = CLK_AUTO_OFF,
    .enable = jpege_clk_enable,
    .disable = jpege_clk_disable,    
	.set_rate = vpu_jpege_clk_set_rate,    
};
#endif

/* Core clock for video codec */
static int vcorebus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int i, matched = 0;

	for(i = 0; i < NUM_FREQS ; i++)
	{
		if(rate <= clk_table[i].freq)
		{
			rate = clk_table[i].freq;
			matched = 1;
			break;
		}
	}

	if (clk->rate != rate && matched /*&& clk->usecount > 0 && (rate > 60000)*/) 
	{
		tca_ckc_setfbusctrl(CLKCTRL6, NOCHANGE, NORMAL_MD, clk_table[i].freq, clk_table[i].clk_src); /*FBUS_VCore*/
		
		clk->rate = rate;
		dprintk("clock: set vpu core clock to %ldMHz\n", rate / 10000);		
	}
	else
	{
		return 1;
	}

	return 0;
}

static int vcore_clk_enable(struct clk *clk)
{
	tca_ckc_enable(CLKCTRL6, 1);
	return 0;
}

static void vcore_clk_disable(struct clk *clk)
{
	tca_ckc_enable(CLKCTRL6, 0);
}

static struct clk clk_vcore = {
	.name = "vcore",
	.parent = &clk_pll2,
	.flags = CLK_AUTO_OFF,	
	.enable = vcore_clk_enable,
	.disable = vcore_clk_disable,
	.set_rate = vcorebus_clk_set_rate,    	
};

#if 1 // DDI-BUS
/* Bus clock for display devices */
static int ddi_clk_enable(struct clk *clk)
{
	int ret;
	
	ret = tca_ckc_setpmupwroff(PMU_DDIBUS, 1);
	tca_ckc_enable(CLKCTRL1, 1);

	if(ret == 1)
	{
		tca_ckc_setddipwdn(0, 0, 1);
	}
	
	return 0;
}

static void ddi_clk_disable(struct clk *clk)
{
	tca_ckc_enable(CLKCTRL1, 0);
	tca_ckc_setpmupwroff(PMU_DDIBUS, 0);
}


/* DDI bus clock */
static int ddibus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int i, matched = 0;
	struct clk *pclk;
	
	list_for_each_entry(pclk, &clock_list, list) {
		if (pclk->parent == clk && pclk->rate > rate) {
                rate = pclk->rate;
		}
	}

	for(i = 0; i < NUM_FREQS ; i++)
	{
		if(rate <= clk_table[i].freq)
		{
			rate = clk_table[i].freq;
			matched = 1;
			break;
		}
	}

	if (clk->rate != rate && matched /*&& clk->usecount > 0*/) 
	{		
		tca_ckc_setfbusctrl(CLKCTRL1, NOCHANGE, NORMAL_MD, clk_table[i].freq, clk_table[i].clk_src); /*FBUS_DDI*/
		clk->rate = rate;

		dprintk("clock: set ddi bus clock to %ldMHz\n", rate / 10000);		
	}

	return 0;
}

static struct clk clk_ddi = {
	.name = "ddi",
	.parent = &clk_pll2,
	.flags = CLK_AUTO_OFF,	
	.enable = ddi_clk_enable,
	.disable = ddi_clk_disable,
	.set_rate = ddibus_clk_set_rate,
};
#endif

static int iobus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int i, matched = 0;
	
	for(i = 0; i < NUM_FREQS ; i++)
	{
		if(rate <= clk_table[i].freq)
		{
			rate = clk_table[i].freq;
			matched = 1;
			break;
		}
	}
	
	if (clk->rate != rate && matched) 
	{		
		tca_ckc_setfbusctrl(CLKCTRL4,  ENABLE, NORMAL_MD, clk_table[i].freq, clk_table[i].clk_src); /*FBUS_IO*/
		clk->rate = rate;

		dprintk("clock: set io bus clock to %ldMHz\n", rate / 10000);		
	}

	return 0;
}

static struct clk clk_io = {
	.name = "iobus",
	.parent = &clk_pll1,
	.flags = CLK_ALWAYS_ENABLED,	
	.set_rate = iobus_clk_set_rate,
};

static int smubus_clk_set_rate(struct clk *clk, unsigned long rate)
{
	tca_ckc_setfbusctrl(CLKCTRL7,  ENABLE, NORMAL_MD, rate, DIRECTPLL0); /*FBUS_SMU*/
	clk->rate = rate;
	
	dprintk("clock: set smu bus clock to %ldMHz\n", rate / 10000);

	return 0;
}

static struct clk clk_smu = {
	.name = "smu",
	.parent = &clk_pll0,
	.flags = CLK_ALWAYS_ENABLED,	
	.set_rate = smubus_clk_set_rate,
};

static int sdhc_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_SDMMCCONTROLLER,1);
	return 0;
}

static void sdhc_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_SDMMCCONTROLLER,0);
}

static struct clk clk_sdhc = {
	.name = "sdhc",
	.enable = sdhc_clk_enable,
	.disable = sdhc_clk_disable,
};


/* SD/MMC channel 0 clock */
static int sdhc0_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_SDMMC0,1);
	return 0;
}

static void sdhc0_clk_disable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_SDMMC0,0);
}

static struct clk clk_sdhc0 = {
	.name = "sdhc0",
	.parent = &clk_sdhc,
	.enable = sdhc0_clk_enable,
	.disable = sdhc0_clk_disable,
};

/* SD/MMC channel 0 slot 0 clock */
static int sdhc0_slot0_clk_enable(struct clk *clk)
{
	BITSET(HwSDCORE0SLOT0->CLK, HwSDCLKSEL_SCK_EN);
	return 0;
}

static void sdhc0_slot0_clk_disable(struct clk *clk)
{
	BITCLR(HwSDCORE0SLOT0->CLK, HwSDCLKSEL_SCK_EN);
}

static struct clk clk_sdhc0_slot0 = {
	.name = "sdhc0_slot0",
	.parent = &clk_sdhc0,
	.enable = sdhc0_slot0_clk_enable,
	.disable = sdhc0_slot0_clk_disable,
};

/* SD/MMC channel 0 slot 1 clock */
static int sdhc0_slot1_clk_enable(struct clk *clk)
{
	BITSET(HwSDCORE0SLOT1->CLK, HwSDCLKSEL_SCK_EN);
	return 0;
}

static void sdhc0_slot1_clk_disable(struct clk *clk)
{
	BITCLR(HwSDCORE0SLOT1->CLK, HwSDCLKSEL_SCK_EN);
}

static struct clk clk_sdhc0_slot1 = {
	.name = "sdhc0_slot1",	
	.parent = &clk_sdhc0,
	.enable = sdhc0_slot1_clk_enable,
	.disable = sdhc0_slot1_clk_disable,
};

/* SD/MMC channel 1 clock */
static int sdhc1_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_SDMMC1,1);
	return 0;
}

static void sdhc1_clk_disable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_SDMMC1,0);
}

static struct clk clk_sdhc1 = {
	.name = "sdhc1",
	.parent = &clk_sdhc,
	.enable = sdhc1_clk_enable,
	.disable = sdhc1_clk_disable,
};

/* SD/MMC channel 1 slot 0 clock */
static int sdhc1_slot0_clk_enable(struct clk *clk)
{
	BITSET(HwSDCORE1SLOT0->CLK, HwSDCLKSEL_SCK_EN);
	return 0;
}

static void sdhc1_slot0_clk_disable(struct clk *clk)
{
	BITCLR(HwSDCORE1SLOT0->CLK, HwSDCLKSEL_SCK_EN);
}

static struct clk clk_sdhc1_slot0 = {
	.name = "sdhc1_slot0",
	.flags = CLK_AUTO_OFF,
	.parent = &clk_sdhc1,
	.enable = sdhc1_slot0_clk_enable,
	.disable = sdhc1_slot0_clk_disable,
};

/* SD/MMC channel 1 slot 1 clock */
static int sdhc1_slot1_clk_enable(struct clk *clk)
{
	BITSET(HwSDCORE1SLOT1->CLK, HwSDCLKSEL_SCK_EN);
	return 0;
}

static void sdhc1_slot1_clk_disable(struct clk *clk)
{
	BITSET(HwSDCORE1SLOT0->CLK, HwSDCLKSEL_SCK_EN);
}

static struct clk clk_sdhc1_slot1 = {
	.name = "sdhc1_slot1",
	.flags = CLK_AUTO_OFF,
	.parent = &clk_sdhc1,
	.enable = sdhc1_slot1_clk_enable,
	.disable = sdhc1_slot1_clk_disable,
};

/* USB 1.1 host controller clock */
static int usb_host_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_USB11H, 1);
	return 0;
}

static void usb_host_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_USB11H, 0);
}

static struct clk clk_usb_host = {
	.name = "usb_host",
#ifndef CONFIG_USB_HID
	.flags = CLK_AUTO_OFF,	
#endif
	.enable = usb_host_clk_enable,
	.disable = usb_host_clk_disable,
};

/* USB 2.0 OTG controller clock */
static int usb_otg_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_USB20OTG, 1);
	return 0;
}

static void usb_otg_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_USB20OTG, 0);
}

static struct clk clk_usb_otg = {
	.name = "usb_otg",
	.flags = CLK_AUTO_OFF,	
	.enable = usb_otg_clk_enable,
	.disable = usb_otg_clk_disable,
};

/* IDE controller clock */
static int ide_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_IDECONTROLLER, 1);
	return 0;
}

static void ide_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_IDECONTROLLER, 0);
}

static struct clk clk_ide = {
	.name = "ide_clk",
	.flags = CLK_AUTO_OFF,	
	.enable = ide_clk_enable,
	.disable = ide_clk_disable,
};

/* SATA controller clock */
static int sata_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_SATAHCONTROLLER, 1);
	return 0;
}

static void sata_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_SATAHCONTROLLER, 0);
}

static struct clk clk_sata = {
	.name = "sata_hclk",
	.flags = CLK_AUTO_OFF,	
	.enable = sata_clk_enable,
	.disable = sata_clk_disable,
};

/* Memory Stick controller clock */
static int mstick_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_MSTICK, 1);
	tca_ckc_setiobus(RB_MEMORYSTICKCONTROLLER, 1);
	return 0;
}

static void mstick_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_MEMORYSTICKCONTROLLER, 0);
	tca_ckc_pclk_enable(PERI_MSTICK, 0);
}

static struct clk clk_mstick = {
	.name = "mstick_clk",
	.flags = CLK_AUTO_OFF,	
	.enable = mstick_clk_enable,
	.disable = mstick_clk_disable,
};

/* UART0 clock */
static int uart0_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_UART0, 1);
	tca_ckc_setiobus(RB_UARTCONTROLLER0, 1);
	return 0;
}

static void uart0_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_UARTCONTROLLER0, 0);
	tca_ckc_pclk_enable(PERI_UART0, 0);
}

static struct clk clk_uart0  = {
	.name = "uart0",
	.flags = CLK_AUTO_OFF,	
	.enable = uart0_clk_enable,
	.disable = uart0_clk_disable,
};

/* UART1 clock */
static int uart1_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_UART1, 1);
	tca_ckc_setiobus(RB_UARTCONTROLLER1, 1);
	return 0;
}

static void uart1_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_UARTCONTROLLER1, 0);
	tca_ckc_pclk_enable(PERI_UART1, 0);
}

static struct clk clk_uart1 = {
	.name = "uart1",
	.flags = CLK_AUTO_OFF,	
	.enable = uart1_clk_enable,
	.disable = uart1_clk_disable,
};

/* UART2 clock */
static int uart2_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_UART2, 1);
	tca_ckc_setiobus(RB_UARTCONTROLLER2, 1);
	return 0;
}

static void uart2_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_UARTCONTROLLER2, 0);
	tca_ckc_pclk_enable(PERI_UART2, 0);
}

static struct clk clk_uart2 = {
	.name = "uart2",
	.flags = CLK_AUTO_OFF,	
	.enable = uart2_clk_enable,
	.disable = uart2_clk_disable,
};

/* UART3 clock */
static int uart3_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_UART3, 1);
	tca_ckc_setiobus(RB_UARTCONTROLLER3, 1);
	return 0;
}

static void uart3_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_UARTCONTROLLER3, 0);
	tca_ckc_pclk_enable(PERI_UART3, 0);
}

static struct clk clk_uart3 = {
	.name = "uart3",
	.flags = CLK_AUTO_OFF,	
	.enable = uart3_clk_enable,
	.disable = uart3_clk_disable,
};

/* UART4 clock */
static int uart4_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_UART4, 1);
	tca_ckc_setiobus(RB_UARTCONTROLLER4, 1);
	return 0;
}

static void uart4_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_UARTCONTROLLER4, 0);
	tca_ckc_pclk_enable(PERI_UART4, 0);
}

static struct clk clk_uart4 = {
	.name = "uart4",
	.flags = CLK_AUTO_OFF,	
	.enable = uart4_clk_enable,
	.disable = uart4_clk_disable,
};

/* UART5 clock */
static int uart5_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_UART5, 1);
	tca_ckc_setiobus(RB_UARTCONTROLLER5, 1);
	return 0;
}

static void uart5_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_UARTCONTROLLER5, 0);
	tca_ckc_pclk_enable(PERI_UART5, 0);
}

static struct clk clk_uart5 = {
	.name = "uart5",
	.flags = CLK_AUTO_OFF,	
	.enable = uart5_clk_enable,
	.disable = uart5_clk_disable,
};


/* GPSB channel 0 clock */
static int gpsb0_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_GPSB0, 1);
	tca_ckc_setiobus(RB_GPSBCONTROLLER0, 1);
	return 0;
}

static void gpsb0_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSBCONTROLLER0, 0);
	tca_ckc_pclk_enable(PERI_GPSB0, 0);
}
static struct clk clk_gpsb0 = {
	.name = "gpsb0",
	.flags = CLK_AUTO_OFF,
	.enable = gpsb0_clk_enable,
	.disable = gpsb0_clk_disable,
};

/* GPSB channel 1 clock */
static int gpsb1_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_GPSB1, 1);
	tca_ckc_setiobus(RB_GPSBCONTROLLER1, 1);
	return 0;
}

static void gpsb1_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSBCONTROLLER1, 0);
	tca_ckc_pclk_enable(PERI_GPSB1, 0);
}
static struct clk clk_gpsb1 = {
	.name = "gpsb1",
	.flags = CLK_AUTO_OFF,
	.enable = gpsb1_clk_enable,
	.disable = gpsb1_clk_disable,
};

/* GPSB channel 2 clock */
static int gpsb2_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_GPSB2, 1);
	tca_ckc_setiobus(RB_GPSBCONTROLLER2, 1);
	return 0;
}

static void gpsb2_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSBCONTROLLER2, 0);
	tca_ckc_pclk_enable(PERI_GPSB2, 0);
}
static struct clk clk_gpsb2 = {
	.name = "gpsb2",
	.flags = CLK_AUTO_OFF,
	.enable = gpsb2_clk_enable,
	.disable = gpsb2_clk_disable,
};

/* GPSB channel 3 clock */
static int gpsb3_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_GPSB3, 1);
	tca_ckc_setiobus(RB_GPSBCONTROLLER3, 1);
	return 0;
}

static void gpsb3_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSBCONTROLLER3, 0);
	tca_ckc_pclk_enable(PERI_GPSB3, 0);
}
static struct clk clk_gpsb3 = {
	.name = "gpsb3",
	.flags = CLK_AUTO_OFF,
	.enable = gpsb3_clk_enable,
	.disable = gpsb3_clk_disable,
};

/* GPSB channel 4 clock */
static int gpsb4_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_GPSB4, 1);
	tca_ckc_setiobus(RB_GPSBCONTROLLER4, 1);
	return 0;
}

static void gpsb4_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSBCONTROLLER4, 0);
	tca_ckc_pclk_enable(PERI_GPSB4, 0);
}
static struct clk clk_gpsb4 = {
	.name = "gpsb4",
	.flags = CLK_AUTO_OFF,
	.enable = gpsb4_clk_enable,
	.disable = gpsb4_clk_disable,
};

/* GPSB channel 5 clock */
static int gpsb5_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_GPSB5, 1);
	tca_ckc_setiobus(RB_GPSBCONTROLLER5, 1);
	return 0;
}

static void gpsb5_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSBCONTROLLER5, 0);
	tca_ckc_pclk_enable(PERI_GPSB5, 0);
}
static struct clk clk_gpsb5 = {
	.name = "gpsb5",
	.flags = CLK_AUTO_OFF,
	.enable = gpsb5_clk_enable,
	.disable = gpsb5_clk_disable,
};

/* I2C clock */
static int i2c_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_I2C, 1);
	tca_ckc_setiobus(RB_I2CCONTROLLER, ENABLE);
	return 0;
}

static void i2c_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_I2CCONTROLLER, DISABLE);
	tca_ckc_pclk_enable(PERI_I2C, 0);
}

static struct clk clk_i2c = {
	.name = "i2c",
	.parent = &clk_xin,
	.flags = CLK_AUTO_OFF,	
	.enable = i2c_clk_enable,
	.disable = i2c_clk_disable,
};

/* ADC clock */
static int adc_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_ADC, 1);
	tca_ckc_setiobus(RB_TSADCCONTROLLER, 1);
	return 0;
}

static void adc_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_TSADCCONTROLLER, 0);
	tca_ckc_pclk_enable(PERI_ADC, 0);
}

static struct clk clk_adc = {
	.name = "adc",
        .parent = &clk_xin,
        .flags = CLK_AUTO_OFF,
        .enable = adc_clk_enable,
        .disable = adc_clk_disable,
};

/* SPDIF clock */
static int spdif_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_SPDIF, 1);
	tca_ckc_setiobus(RB_SPDIFTXCONTROLLER, 1);
	return 0;
}

static void spdif_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_SPDIFTXCONTROLLER, 0);
	tca_ckc_pclk_enable(PERI_SPDIF, 0);
}

static struct clk clk_spdif = {
	.name = "spdif_clk",
	.flags = CLK_AUTO_OFF,
	.enable = spdif_clk_enable,
	.disable = spdif_clk_disable,
};

/* EHI channel 0 clock */
static int ehi0_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_EHI0, 1);
	tca_ckc_setiobus(RB_EXTHCONTROLLER0, 1);
	return 0;
}

static void ehi0_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_EXTHCONTROLLER0, 0);
	tca_ckc_pclk_enable(PERI_EHI0, 0);
}

static struct clk clk_ehi0 = {
	.name = "ehi0",
	.flags = CLK_AUTO_OFF,
	.enable = ehi0_clk_enable,
	.disable = ehi0_clk_disable,
};

/* EHI channel 1 clock */
static int ehi1_clk_enable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_EHI1, 1);
	tca_ckc_setiobus(RB_EXTHCONTROLLER1, 1);
	return 0;
}

static void ehi1_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_EXTHCONTROLLER1, 0);
	tca_ckc_pclk_enable(PERI_EHI1, 0);
}

static struct clk clk_ehi1 = {
	.name = "ehi1",
	.flags = CLK_AUTO_OFF,
	.enable = ehi1_clk_enable,
	.disable = ehi1_clk_disable,
};

/* GPS controller clock */
static int gps_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSCONTROLLER, 1);
	return 0;
}

static void gps_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_GPSCONTROLLER, 0);
}

static struct clk clk_gps = {
	.name = "gps_hclk",
	.flags = CLK_AUTO_OFF,
	.enable = gps_clk_enable,
	.disable = gps_clk_disable,
};

/* Reserved clock */
static int reserved_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_RESERVEDCONTROLLER, 1);
	return 0;
}

static void reserved_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_RESERVEDCONTROLLER, 0);
}

static struct clk clk_reserved = {
	.name = "reserved",
	.flags = CLK_AUTO_OFF,
	.enable = reserved_clk_enable,
	.disable = reserved_clk_disable,
};

/* CAN clock */
static int can_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_CANCONTROLLER, 1);
	return 0;
}

static void can_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_CANCONTROLLER, 0);
}

static struct clk clk_can = {
	.name = "can_hclk",
	.flags = CLK_AUTO_OFF,
	.enable = can_clk_enable,
	.disable = can_clk_disable,
};

/* MPE_FEC controller clock */
static int mpe_fec_clk_enable(struct clk *clk)
{
	tca_ckc_setiobus(RB_MPE_FECCONTROLLER, 1);
	return 0;
}

static void mpe_fec_clk_disable(struct clk *clk)
{
	tca_ckc_setiobus(RB_MPE_FECCONTROLLER, 0);
}

static struct clk clk_mpe_fec = {
	.name = "mpe_fec_hclk",
	.flags = CLK_AUTO_OFF,
	.enable = mpe_fec_clk_enable,
	.disable = mpe_fec_clk_disable,
};

/* SRAM controller clock */
static int sram_clk_enable(struct clk *clk)
{
//	tca_ckc_setiobus(RB_SRAMCONTROLLER, 1);
	return 0;
}

static void sram_clk_disable(struct clk *clk)
{
//	tca_ckc_setiobus(RB_SRAMCONTROLLER, 0);
}

static struct clk clk_sram = {
	.name = "sram_clk",
	.enable = sram_clk_enable,
	.disable = sram_clk_disable,
};

/* Audio DMA clock */
static struct clk clk_adma = {
	.name = "adma",
};

/* DAI clock */
static struct clk clk_dai = {
	.name = "dai",
};

/* Mali clock */
volatile unsigned char mali_regs[0x4000];		/* Mali 16KB registers address space */

static int mali_clk_enable(struct clk *clk)
{
	tca_ckc_setgrppwdn(GRPPWDN_MALI, 1, 0);
	/* XXX: restore Mali registsers */
//	memcpy((void *) &HwPIXELPROCESSOR_BASE, (void *) mali_regs, sizeof(mali_regs));
	return 0;
}

static void mali_clk_disable(struct clk *clk)
{
	/* XXX: backup Mali registers */
//	memcpy((void *) mali_regs, (void *) &HwPIXELPROCESSOR_BASE, sizeof(mali_regs));
	tca_ckc_setgrppwdn(GRPPWDN_MALI, 0, 0);
}

static struct clk clk_mali = {
	.name = "mali_clk",
	.parent = &clk_gpu,	
	.flags = CLK_AUTO_OFF,
	.enable = mali_clk_enable,
	.disable = mali_clk_disable,
};


/* overlay mixer clock */
static int overlay_clk_enable(struct clk *clk)
{
	tca_ckc_setgrppwdn(GRPPWDN_OVERLAY, 1, 0);
	return 0;
}

static void overlay_clk_disable(struct clk *clk)
{
	tca_ckc_setgrppwdn(GRPPWDN_OVERLAY, 0, 0);
}

static struct clk clk_overlay = {
	.name = "overlay",
	.parent = &clk_gpu,
	.flags = CLK_AUTO_OFF,
	.enable = overlay_clk_enable,
	.disable = overlay_clk_disable,		
};

/* -----------DDI clock -----------*/
/* camera cif clock*/
static int cif_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_CIF, 1, 0);
	return 0;
}

static void cif_clk_disable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_CIF, 0, 0);
}

static struct clk clk_cif = {
	.name = "cif",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = cif_clk_enable,
	.disable = cif_clk_disable,		
};

/* video enhancer clock*/
static int viqe_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_VIQE, 1, 0);
	return 0;
}

static void viqe_clk_disable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_VIQE, 0, 0);
}

static struct clk clk_viqe = {
	.name = "viqe",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = viqe_clk_enable,
	.disable = viqe_clk_disable, 	

};



/* ddi bus clock for HDMI */
static int mem_lcdc0_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int ret = 0;
	struct clk *clk_parent = clk->parent;
	
	clk->rate = rate;
	if(rate == 0)
	{
		clk_parent->flags = 0;
	}
	else
	{
		if(clk_parent->flags & CLK_FIXED_RATE)
		{
			clk_parent->flags = 0;
			ret = _clk_set_rate(clk_parent, rate);
			clk_parent->flags = CLK_FIXED_RATE;
		}
		else
		{
			ret = _clk_set_rate(clk_parent, rate);
			clk_parent->flags = CLK_FIXED_RATE;
		}
	}
	
	return ret;
}

static int mem_lcdc1_clk_set_rate(struct clk *clk, unsigned long rate)
{
	int ret = 0;
	struct clk *clk_parent = clk->parent;
	
	clk->rate = rate;
	if(rate == 0)
	{
		clk_parent->flags = 0;
	}
	else
	{
		if(clk_parent->flags & CLK_FIXED_RATE)
		{
			clk_parent->flags = 0;
			ret = _clk_set_rate(clk_parent, rate);
			clk_parent->flags = CLK_FIXED_RATE;
		}
		else
		{
			ret = _clk_set_rate(clk_parent, rate);
			clk_parent->flags = CLK_FIXED_RATE;
		}
	}
	
	return ret;
}



/* lcdc 0 clock*/
static int lcdc0_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_LCDC0, 1, 0);
	tca_ckc_pclk_enable(PERI_LCD0, 1);
//	HwLCD->LCTRL |= HwLCTRL_LEN;
	return 0;
}

static void lcdc0_clk_disable(struct clk *clk)
{
	HwLCD->LCTRL &= ~HwLCTRL_LEN;
	tca_ckc_pclk_enable(PERI_LCD0, 0);
	tca_ckc_setddipwdn(DDIPWDN_LCDC0, 0, 0);
}

static struct clk clk_lcdc0 = {
	.name = "lcdc0",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = lcdc0_clk_enable,
	.disable = lcdc0_clk_disable, 	
	.set_rate = mem_lcdc0_clk_set_rate,

};

/* lcdc 1 clock*/
static int lcdc1_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_LCDC1, 1, 0);
	tca_ckc_pclk_enable(PERI_LCD1, 1);
//	HwLCD1->LCTRL |= HwLCTRL_LEN;
	return 0;
}

static void lcdc1_clk_disable(struct clk *clk)
{
	HwLCD1->LCTRL &= ~HwLCTRL_LEN;
	tca_ckc_pclk_enable(PERI_LCD1, 0);
	tca_ckc_setddipwdn(DDIPWDN_LCDC1, 0, 0);
}

static struct clk clk_lcdc1 = {
	.name = "lcdc1",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = lcdc1_clk_enable,
	.disable = lcdc1_clk_disable, 	
	.set_rate = mem_lcdc1_clk_set_rate,

};

/* lcd system interface clock*/
static int lcdsi_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_LCDSI, 1, 0);
	tca_ckc_pclk_enable(PERI_LCDSI, 1);
	return 0;
}

static void lcdsi_clk_disable(struct clk *clk)
{
	tca_ckc_pclk_enable(PERI_LCDSI, 0);
	tca_ckc_setddipwdn(DDIPWDN_LCDSI, 0, 0);
}
static struct clk clk_lcdsi = {
	.name = "lcdsi",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = lcdsi_clk_enable,
	.disable = lcdsi_clk_disable, 	

};

/* m2m scaler 0 clock*/
static int m2m0_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_MSCL0, 1, 0);
	return 0;
}

static void m2m0_clk_disable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_MSCL0, 0, 0);
}

static struct clk clk_m2m0 = {
	.name = "m2m0",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = m2m0_clk_enable,
	.disable = m2m0_clk_disable, 	

};

/* m2m scaler 1 clock*/
static int m2m1_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_MSCL1, 1, 0);
	return 0;
}

static void m2m1_clk_disable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_MSCL1, 0, 0);
}

static struct clk clk_m2m1 = {
	.name = "m2m1",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = m2m1_clk_enable,
	.disable = m2m1_clk_disable, 	

};

/* ddi cache clock*/
/* m2m scaler 1 clock  */
static int ddicache_clk_enable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_DDIC, 1, 0);
	return 0;
}

static void ddicache_clk_disable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_DDIC, 0, 0);
}

static struct clk clk_ddi_cache = {
	.name = "ddi_cache",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = ddicache_clk_enable,
	.disable = ddicache_clk_disable, 	

};

/* hdmi clock*/
static int hdmi_clk_enable(struct clk *clk)
{
	tca_ckc_setpmupwroff(PMU_HDMIPHY, 1);
	tca_ckc_setddipwdn(DDIPWDN_HDMI, 1, 0);
	return 0;
}

static void hdmi_clk_disable(struct clk *clk)
{
	tca_ckc_setddipwdn(DDIPWDN_HDMI, 0, 0);
	tca_ckc_setpmupwroff(PMU_HDMIPHY, 0);
}

static struct clk clk_hdmi = {
	.name = "hdmi",
	.parent = &clk_ddi,
	.flags = CLK_AUTO_OFF,
	.enable = hdmi_clk_enable,
	.disable = hdmi_clk_disable, 	
};

/* CIF Internal clock */
static int cifmc_clk_enable(struct clk *clk)
{
    tca_ckc_pclk_enable(PERI_CIFMC, 1);
	return 0;
}

static void cifmc_clk_disable(struct clk *clk)
{
    tca_ckc_pclk_enable(PERI_CIFMC, 0);
}

static struct clk clk_cifmc = {
	.name = "cifmc",
    .parent = &clk_pll2,
    .flags = CLK_AUTO_OFF,
    .enable = cifmc_clk_enable,
    .disable = cifmc_clk_disable,
};

/* CIF Scaler clock */
static int cifsc_clk_enable(struct clk *clk)
{
    tca_ckc_pclk_enable(PERI_CIFSC, 1);
	return 0;
}

static void cifsc_clk_disable(struct clk *clk)
{
    tca_ckc_pclk_enable(PERI_CIFSC, 0);
}

static struct clk clk_cifsc = {
	.name = "cifsc",
    .parent = &clk_pll2,
    .flags = CLK_AUTO_OFF,
    .enable = cifsc_clk_enable,
    .disable = cifsc_clk_disable,
};

/* dac clock*/
static int videodac_enable(struct clk *clk)
{
	tca_ckc_setpmupwroff(PMU_VIDEODAC, 1);
	return 0;
}

static void videodac_disable(struct clk *clk)
{
	tca_ckc_setpmupwroff(PMU_VIDEODAC, 0);
}

static struct clk clk_videodac = {
	.name = "videodac",
	.flags = CLK_AUTO_OFF,
	.enable = videodac_enable,
	.disable = videodac_disable, 	
};

static struct clk *onchip_clocks[] = {
	&clk_pll0,
	&clk_pll1,
	&clk_pll2,
	&clk_pll3,
	&clk_cpu,
	&clk_membus,

//videodac!!
	&clk_videodac, //This should be placed here. Don't move.

//vbus!!
	&clk_vcache,
	&clk_vcodec,
	&clk_jpegd,
	&clk_jpege,
	&clk_vbus,
	
//vcore!!
	&clk_vcore,
	
//gpu!!
    &clk_mali,
	&clk_overlay,
	&clk_gpu,

//ddi!!
	&clk_cif,
	&clk_viqe,
	&clk_lcdc0,
	&clk_lcdc1,
	&clk_lcdsi,
	&clk_m2m0,
	&clk_m2m1,
	&clk_ddi_cache,
	&clk_hdmi,
	&clk_ddi,

//io!!
	&clk_io,
//smu!!
	&clk_smu,

//periperal!!
	&clk_usb_host,
	&clk_usb_otg,
	&clk_ide,
	&clk_sata,
	&clk_mstick,
	&clk_cifmc,
	&clk_cifsc,	
	&clk_sdhc,
	&clk_sdhc0,
	&clk_sdhc0_slot0,
	&clk_sdhc0_slot1,
	&clk_sdhc1,
	&clk_sdhc1_slot0,
	&clk_sdhc1_slot1,	
	&clk_uart0,
	&clk_uart1,
	&clk_uart2,
	&clk_uart3,
	&clk_uart4,
	&clk_uart5,
	&clk_i2c,
	&clk_gpsb0,
	&clk_gpsb1,
	&clk_gpsb2,
	&clk_gpsb3,
	&clk_gpsb4,
	&clk_gpsb5,
	&clk_adc,
	&clk_spdif,
	&clk_ehi0,
	&clk_ehi1,
	&clk_gps,
	&clk_reserved,
	&clk_can,
	&clk_mpe_fec,
	&clk_sram,
	&clk_adma,
	&clk_dai,
};

struct clk *clk_get(struct device *dev, const char *id)
{
	struct clk *clk;

	mutex_lock(&clock_mutex);
	list_for_each_entry(clk, &clock_list, list) {
		if (!strcmp(id, clk->name) && clk->dev == dev)
			goto found;
		if (!strcmp(id, clk->name) && clk->dev == NULL)
			goto found;
	}
	clk = ERR_PTR(-ENOENT);
found:
	mutex_unlock(&clock_mutex);
	return clk;
}
EXPORT_SYMBOL(clk_get);

void clk_put(struct clk *clk)
{
	/* do nothing */
}
EXPORT_SYMBOL(clk_put);

static int _clk_enable(struct clk *clk)
{
	int ret = 0;

	clk->usecount++;
	if (clk->usecount == 1) {
		if (clk->parent)
			ret = _clk_enable(clk->parent);

		if (clk->enable)
			ret = clk->enable(clk);
		else
			pr_warning("clock: %s: unimplemented function 'clk_enable' called\n", clk->name);
	}
	return ret;
}

int clk_enable(struct clk *clk)
{
	unsigned long flags;
	int ret = 0;

	dprintk("clock: %s: enable\n", clk->name);

	BUG_ON(clk == 0);

	spin_lock_irqsave(&clock_lock, flags);
	ret = _clk_enable(clk);
	spin_unlock_irqrestore(&clock_lock, flags);
	return ret;
}
EXPORT_SYMBOL(clk_enable);

static void _clk_disable(struct clk *clk)
{
	if (clk->flags & CLK_ALWAYS_ENABLED)
		return;

	clk->usecount--;
	if (clk->usecount == 0) {
		if (clk->disable)
			(*clk->disable)(clk);
		else
			pr_warning("clock: %s: unimplemented function 'clk_disable' called\n", clk->name);
		if (clk->parent)
			_clk_disable(clk->parent);
	}	
}

void clk_disable(struct clk *clk)
{
	unsigned long flags;

	dprintk("clock: %s: disable\n", clk->name);

//	if (strcmp(clk->name, "mali_clk")==0) {
//		clk->usecount = 1;
//	}

	BUG_ON(clk->usecount == 0);

	spin_lock_irqsave(&clock_lock, flags);
	_clk_disable(clk);
	spin_unlock_irqrestore(&clock_lock, flags);
}
EXPORT_SYMBOL(clk_disable);

unsigned long clk_get_rate(struct clk *clk)
{
	unsigned long rate;

	mutex_lock(&clock_mutex);
	if (clk->get_rate)
		rate = (*clk->get_rate)(clk);
	else
		rate = clk->rate;
	mutex_unlock(&clock_mutex);
	return rate;
}
EXPORT_SYMBOL(clk_get_rate);

/*
 * Set the clock rate for a clock source
 */
static int _clk_set_rate(struct clk *clk, unsigned long rate)
{
	int ret = 0;
	unsigned long flags;

	if (clk->flags & CLK_FIXED_RATE)
		return 0;

	dprintk("clock: %s: clock rate = %ld\n", clk->name, rate);

	spin_lock_irqsave(&clock_lock, flags);
	if (clk->set_rate) {
		ret = (*clk->set_rate)(clk, rate);
	} else {
		pr_warning("clock: unimplemented function 'clk_set_rate' called\n");
	}
	spin_unlock_irqrestore(&clock_lock, flags);
	return ret;
}

int clk_set_rate(struct clk *clk, unsigned long rate)
{
	int ret;

	ret = _clk_set_rate(clk, rate);
	
	if(ret == 0){
		clk->rate = rate;
	}
	
	return ret;
}
EXPORT_SYMBOL(clk_set_rate);

/*
 * Adjust a rate to the exact rate a clock can provide
 */
long clk_round_rate(struct clk *clk, unsigned long rate)
{
	//TODO
	return rate;
}
EXPORT_SYMBOL(clk_round_rate);

/*
 * Set the parent clock source for the clock
 */
int clk_set_parent(struct clk *clk, struct clk *parent)
{
	unsigned long flags;
	spin_lock_irqsave(&clock_lock, flags);
	clk->parent = parent;
	spin_unlock_irqrestore(&clock_lock, flags);
	return 0;
}
EXPORT_SYMBOL(clk_set_parent);

/*
 * Get the parent clock source for the clock
 */
struct clk *clk_get_parent(struct clk *clk)
{
	BUG_ON(clk == 0);

	return clk->parent;
}
EXPORT_SYMBOL(clk_get_parent);

int __init clk_init(void)
{
	struct clk **clk;

	spin_lock_init(&clock_lock);
	mutex_lock(&clock_mutex);

	tca_ckc_init();

	for (clk = onchip_clocks; clk < onchip_clocks + ARRAY_SIZE(onchip_clocks); clk++) {
		list_add_tail(&(*clk)->list, &clock_list);
	}
	mutex_unlock(&clock_mutex);

	pr_info("TCC clock driver initialized\n");

	return 0;
}

arch_initcall(clk_init);

void clk_disable_unused(void)
{
	struct clk *clk;
	int count = 0;

	list_for_each_entry(clk, &clock_list, list) {
		if (clk->flags & CLK_AUTO_OFF && !clk->usecount && clk->disable) {
			clk->disable(clk);
			count++;
			pr_debug("clock: clock '%s' disabled\n", clk->name);
		}
	}
	pr_debug("clock: disabled %d unused clocks\n", count);
}
EXPORT_SYMBOL(clk_disable_unused);

static int __init clock_late_init(void)
{
	clk_disable_unused();
	return 0;
}

late_initcall(clock_late_init);

#include <linux/proc_fs.h>
#include <linux/seq_file.h>

static void *tcc92xx_ck_start(struct seq_file *m, loff_t *pos)
{
	return *pos < 1 ? (void *)1 : NULL;
}

static void *tcc92xx_ck_next(struct seq_file *m, void *v, loff_t *pos)
{
	++*pos;
	return NULL;
}

static void tcc92xx_ck_stop(struct seq_file *m, void *v)
{
}

static int tcc92xx_ck_show(struct seq_file *m, void *v)
{
	struct clk *cp;
	
	list_for_each_entry(cp, &clock_list, list)
		if(!strcmp(cp->name,"cpu") || !strcmp(cp->name,"membus"))
			seq_printf(m,"%d ",cp->rate);
		else if(!strcmp(cp->name,"ddi") || !strcmp(cp->name,"gpu") || !strcmp(cp->name,"iobus") || 
				!strcmp(cp->name,"vbus") || !strcmp(cp->name,"vcore") || !strcmp(cp->name,"smu"))
			seq_printf(m,"%d ",cp->rate*100);

	seq_printf(m,"\n");

	return 0;
}

static const struct seq_operations tcc92xx_ck_op = {
	.start	= tcc92xx_ck_start,
	.next	= tcc92xx_ck_next,
	.stop	= tcc92xx_ck_stop,
	.show	= tcc92xx_ck_show
};

static int tcc92xx_ck_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &tcc92xx_ck_op);
}

static const struct file_operations proc_tcc92xx_ck_operations = {
	.open		= tcc92xx_ck_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

static int __init tcc92xx_ck_proc_init(void)
{
	proc_create("clocks", 0, NULL, &proc_tcc92xx_ck_operations);
	return 0;

}
__initcall(tcc92xx_ck_proc_init);
