/* linux/drivers/ide/arm/bast-ide.c
 *
 * Copyright (c) 2003-2004 Simtec Electronics
 *  Ben Dooks <ben@simtec.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
*/

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/errno.h>

#include <asm/io.h>
#include <asm/irq.h>

#include <asm/arch/tcc79x.h>
#include <asm/plat-tcc/tcc_ckc.h>

/* list of registered interfaces */
static ide_hwif_t *ifs[1];

#ifdef __KERNEL__
#   define TCC_ATAPI_DELAY(X) mdelay(X)
#else
#   define TCC_ATAPI_DELAY(X) Sleep(X)
#endif

#pragma pack(push, 4)
struct tca_ide_regs {
    volatile unsigned long PIOCTRL, UDMACTRL,
             IDMACTRL, IDMASA, IDMASP, IDMACSA, IDMADA, IDMADP, IDMACDA,
             IDEINT, UDMATCNT, UDMAIN, UDMAOUT, UDMACRC, UDMACTCNT;
};
#pragma pack(pop)

struct tca_ide_handle {
    volatile struct tca_ide_regs *regs;
    int irq;
    unsigned long cs0_addr;
    unsigned long cs1_addr;
};
static struct tca_ide_handle *p_ide_handle = NULL;



static struct tca_ide_handle *tca_new_ide(int irq)
{
    unsigned long base_addr = (unsigned long)(&(HwIDE_BASE));
    struct tca_ide_handle *h = NULL;

#ifdef __KERNEL__
    h = (struct tca_ide_handle *)kmalloc(sizeof(struct tca_ide_handle), GFP_KERNEL);
#else
    h = (struct tca_ide_handle *)malloc(sizeof(struct tca_ide_handle));
#endif
    if (h) {
        memset(h, 0, sizeof(struct tca_ide_handle));

        h->regs = (volatile struct tca_ide_regs *)(base_addr + 0x40);
        memset((void *)(h->regs), 0, sizeof(struct tca_ide_regs));

        h->cs0_addr = base_addr;
        h->cs1_addr = base_addr + 0x20;
        h->irq = irq;
    }
    return h;
}

static struct tca_ide_handle *tca_delete_ide(struct tca_ide_handle *h)
{
    if (h) {
#ifdef __KERNEL__
        kfree(h);
#else
        free(h);
#endif
    }
    return NULL;
}

static int tca_ide_pio_config(struct tca_ide_handle *h)
{
    unsigned bus_total_cycle, setup, pw, hold;
    unsigned int bus_clk = 0;

    bus_clk = tca_ckc_getcurrentbus();
    printk("bus_clk :  BUS-%d\n", bus_clk);

    bus_total_cycle = 120 * bus_clk / 10000000L;
    if (bus_total_cycle <= 10) {
        setup   = 2;    // 3 cycle
        pw      = 5;    // 6 cycle
        hold    = 0;    // 1 cycle
    } else {
        bus_total_cycle = bus_total_cycle - 2;  // assume Hold = 2

        hold    = 1;    // 1 means 2 cycles
        pw      = bus_total_cycle * 75 / 100 - 1;   // PW = 75% of remain cycles
        setup   = bus_total_cycle - pw - 1;     // Stp = 25% of remain cycles                
    }

    printk("@@@ setup[%d], pw[%d]. hold[%d]\n", setup, pw, hold);

    /* PIO mode setting */
    //h->regs->PIOCTRL = (HwPIOCTRL_SYNC_2 | (setup << 13) | (pw << 7) | (hold << 1));
    h->regs->PIOCTRL = (HwPIOCTRL_SYNC_2 | (setup << 13) | (pw << 7) | (hold << 1) | HwPIOCTRL_RDY_EXT);
    return 0;
}

static int tca_ide_port_config(struct tca_ide_handle *h)
{
    /* HDD port config */
    BITCSET(HwPORTCFG2, 0xF0FFFFFF, 0x40444444);
    HwPORTCFG3 = 0x44444444;
    BITCSET(HwPORTCFG13, 0xF, 0x4);

    /* RESET port config (GPIOF_25) */
    BITCSET(HwPORTCFG2, 0x0F000000, 0x01000000);
    BITSET(HwGPFEN, Hw25);

    /* Externel intr GPIO (F_27) config */
    BITCSET(HwPORTCFG1, 0xF, 0x1);
    BITCLR(HwGPFEN, Hw27);
    BITCLR(HwGPFDAT, Hw27);

#ifdef __KERNEL__
    /* EINT setting (GPIOF_27: EINT3) */
    BITCSET(HwEINTSEL0, 0xFF000000, SEL_GPIOF27 << 24);

    BITSET(HwSEL, HwINT_EI3);
    BITCLR(HwPOL, HwINT_EI3);
    BITSET(HwMODE, HwINT_EI3);
    BITSET(HwIRQMSK, HwINT_EI3);
#else
    /* EINT setting (GPIOF_27: EINT6 => SEL3) */
    BITCSET(HwEINTSEL1, Hw21 - Hw16, SEL_GPIOF27 << 16);
    BITSET(HwIRQSEL, HwIRQSEL_SEL3_EINT6);

    BITSET(HwSEL, HwINT_SEL3);
    BITCLR(HwPOL, HwINT_SEL3);
    BITSET(HwMODE, HwINT_SEL3);
    BITSET(HwIRQMSK, HwINT_SEL3);
#endif
    return 0;
}

static int tca_ide_hw_reset(struct tca_ide_handle *h, int deley_ms)
{
#if 0
    BITCLR(HwGPFDAT, Hw25);
    TCC_ATAPI_DELAY(1);
    BITSET(HwGPFDAT, Hw25);
    TCC_ATAPI_DELAY(deley_ms);
#else
    BITSET(HwGPFDAT, Hw25);
    TCC_ATAPI_DELAY(1);
    BITCLR(HwGPFDAT, Hw25);
    TCC_ATAPI_DELAY(deley_ms);
#endif
    return 0;
}

static int __init tcc_ide_register(struct tca_ide_handle *h, ide_hwif_t **hwif)
{
	hw_regs_t hw;
	int i;
    unsigned long cs0 = h->cs0_addr;
    unsigned long cs1 = h->cs1_addr;

	memset(&hw, 0, sizeof(hw));

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw.io_ports[i] = cs0;
		cs0 += 4;
	}

	hw.io_ports[IDE_CONTROL_OFFSET] = cs1 + (6 * 4);
	//hw.io_ports[IDE_IRQ_OFFSET] = cs1 + (7 * 4);

	hw.irq = h->irq;

	ide_register_hw(&hw, NULL, 0, hwif);

	return 0;
}

static int __init tcc_ide_init(void)
{
    tca_ckc_com_setperibus(RB_HDD, 1);

    p_ide_handle = tca_new_ide(INT_EI3);
    if (p_ide_handle) {
        tca_ide_pio_config(p_ide_handle);
        tca_ide_port_config(p_ide_handle);
        tca_ide_hw_reset(p_ide_handle, 1000);

        printk("Telechips: IDE driver (using IORDY) [%s - %s]\n", __DATE__, __TIME__);

        tcc_ide_register(p_ide_handle, &ifs[0]);
        return 0;
    }

    p_ide_handle = tca_delete_ide(p_ide_handle);
    printk("Telechips: IDE driver fail !!!!! [%s - %s]\n", __DATE__, __TIME__);
    return -1;
}

module_init(tcc_ide_init);

MODULE_AUTHOR("linux@telechips.com");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("TCC79X IDE driver");

