/*
 * linux/drivers/serial/tcc_ts_ak.c
 *
 * Author: <linux@telechips.com>
 * Created:     June 10, 2008
 * Description: Touchscreen driver for AK4183 on Telechips TCC Series
 *
 * Copyright (C) 2008-2009 Telechips 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/device.h>

#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/sched.h>


#include <linux/slab.h>
#include <linux/irq.h>

#include <linux/spinlock.h>

#include <asm/hardware.h>
#include <asm/io.h>
#ifdef CONFIG_MACH_TCC7901S
#include <asm/arch/tcc79x.h>
#else
#include <asm/arch/tcc83x_virt.h>
#endif
#include "tca_tchcontrol.h"
#include "tca_tchhwctl.h"
#include "tca_ak4183tchctl.h"

#ifdef CONFIG_I2C
#   include <linux/i2c.h>
#else
#   include "tcc83xx_i2c.h"
#endif


//==============================================================
#define DEVICE_NAME	"ak4183_ts"

#define CMD_SLEEP_NO 0x80
#define CMD_ADC_ON 0x04

#define PEN_DOWN 1
#define PEN_RELEASE 0

#ifdef CONFIG_MACH_TCC7901S
#define READ_PENIRQ() (HwGPEDAT & Hw31)
#define AK_I2C_NUM 0
#else
#define READ_PENIRQ() (HwGDATA_B & Hw6)
#define AK_I2C_NUM 1
#endif

struct ts_event {
	unsigned short pressure;
	unsigned short x;
	unsigned short y;
};

struct ak4183_ts {
    struct input_dev *input;
    unsigned int x, y, pressure, valid;
    int pen_status, opened, running;
    int irq;
    struct mutex mutex;
    //wait_queue_head_t wait_q;
	struct work_struct work_q;
};

struct ak4183_ts *ak4183_ts;
static struct timer_list ts_timer;
static int irq_no;

#ifdef CONFIG_I2C
static unsigned short normal_i2c[] = { 0x48, I2C_CLIENT_END };
static unsigned short dummy[] = {I2C_CLIENT_END};

static struct i2c_client_address_data addr_data = {
	.normal_i2c = normal_i2c,
	.probe      = dummy,
	.ignore     = dummy,
};

static struct i2c_driver ak_i2c_driver;
static struct i2c_client *ak_i2c_client;

static void TOUCH_SEND_CMD(unsigned char reg, unsigned char val)
{
	unsigned char cmd[1];
	cmd[0] = val;
	i2c_master_send(ak_i2c_client, cmd, 1);
}

static void TOUCH_READ_DAT(unsigned char reg, unsigned int val)
{
	unsigned char *buf;
	buf = (unsigned char *)val;
	i2c_master_recv(ak_i2c_client, buf, 2);
}

static int ak_i2c_probe(struct i2c_adapter *adap, int addr, int kind)
{
    struct i2c_client *i2c;
    int ret;

    i2c = kzalloc(sizeof(*i2c), GFP_KERNEL);
    if (!i2c)
        return -ENOMEM;

    strcpy(i2c->name, "AK4183");
    i2c->flags = 0;
    i2c->addr = addr;
    i2c->adapter = adap;
    i2c->driver = &ak_i2c_driver;

    ak_i2c_client = i2c;

    /* attach i2c client device */
    ret = i2c_attach_client(i2c);
    if (ret < 0) {
        printk("%s: failed to attach codec at addr %x\n", __func__, addr);
        goto err;
    }

    return ret;
err:
    kfree(i2c);
    return ret;
}

static int ak_i2c_detach(struct i2c_client *client)
{
    i2c_detach_client(client);
    kfree(ak_i2c_client);
    return 0;
}

static int ak_i2c_attach(struct i2c_adapter *adap)
{
    int ret = -1;
	if (adap->nr == AK_I2C_NUM) {
		ret = i2c_probe(adap, &addr_data, ak_i2c_probe);
    }
    return ret;
}

/* ak4183 i2c codec control layer */
static struct i2c_driver ak_i2c_driver = {
    .driver = {
        .name   = "tcc-ak4183-ts",
        .owner  = THIS_MODULE,
    },
    .id             = I2C_DRIVERID_AK4183,
    .attach_adapter = ak_i2c_attach,
    .detach_client  = ak_i2c_detach,
    .command        = NULL,
};
#endif

static void write_cmd_ak(unsigned char cmd)
{
	TOUCH_SEND_CMD(0, cmd);
	
#ifndef CONFIG_I2C
	udelay(1);
#endif
}

static short read_ak(int cmd)
{
	unsigned char data[]={0,0};
	short value;
	
	cmd |= (CMD_SLEEP_NO | CMD_ADC_ON);
    write_cmd_ak(cmd);
	
	TOUCH_READ_DAT(0, (unsigned int)(&data[0]));
	
	value = (data[0] & 0x00ff) << 4;
	value |= ((data[1] & 0x00ff) >> 4);
	value &= 0xfff;
	
#ifndef CONFIG_I2C
	udelay(1);
#endif
	return(value>>1);
}

void tea_tch_openi2c(void)
{
    int ret;
#ifdef CONFIG_I2C
	ret = i2c_add_driver(&ak_i2c_driver);
	if (ret != 0)
		printk("%s: can't add i2c driver", __func__);
#else
	TOUCH_IIC_INIT();
#endif
}

short tea_tch_readi2c(int channel)
{
    short ret;
    unsigned char buf[2];
    int cmd;
    tca_ak4183_setreadcmd(channel, buf);
    //printk("tea_tch_readi2c, channel = (%x), buf = (%x)\n", channel, buf[0]);
    cmd = (int) buf[0];
    ret = read_ak(cmd);
    return ret;
}

int tea_tch_writei2c(unsigned char command)
{
	/* debug code sgbaek */
	if(ak_i2c_client == NULL)
		printk("tea_tch_writei2c(tcc_ts_ak.c): ak_i2c_client is NULL \n");
    i2c_master_send(ak_i2c_client, &command, 1);
    return 0;
}

static inline void ak4183_adc_on(void)
{
	u8	command = CMD_SLEEP_NO | CMD_ADC_ON;
	i2c_master_send(ak_i2c_client, &command, 1);
}

static inline void ak_pen_release(struct ak4183_ts* ts_data, struct input_dev *dev)
{
    if (ts_data->pen_status != PEN_RELEASE) {
        /* printk("PEN UP\n"); */
        ts_data->pen_status = PEN_RELEASE;

        input_report_key(dev, BTN_TOUCH, PEN_RELEASE);
        input_sync(dev);
        /* wake_up_interruptible(&ts_data->wait_q); */
    }
}

static inline void ak_pen_pressure(struct ak4183_ts* ts_data, struct input_dev *dev)
{
    /* printk("PEN DOWN %u, %u, %u\n", ts_data->x, ts_data->y, ts_data->pressure); */
    ts_data->pen_status = PEN_DOWN;

    input_report_key(dev, BTN_TOUCH, PEN_DOWN);
    input_report_abs(dev, ABS_X, ak4183_ts->x);
    input_report_abs(dev, ABS_Y, ak4183_ts->y);
    input_sync(dev);

    /* wake_up_interruptible(&ts_data->wait_q); */
}

static void ts_fetch_thread(struct work_struct *work)
{
    struct ak4183_ts* ts_data = container_of(work, struct ak4183_ts, work_q);
    struct input_dev *dev = ak4183_ts->input;
    int flag, valid;
   
    ts_data->running = 1;

    tca_tchhw_disableirq();

    if (READ_PENIRQ()) {
        ak_pen_release(ts_data, dev);
        ts_data->running = 0;
        tca_tchhw_enableirq();
    } else {
        ak4183_adc_on();
        flag = tca_getrawdata(&(ts_data->x), &(ts_data->y));
        tca_tch_poweroff();
        tca_tch_poweroff();

        valid = ts_data->x | ts_data->y;
        if ((flag == 0) && valid) {
            ak_pen_pressure(ts_data, dev);
            ts_timer.expires = jiffies + 1;
            add_timer(&ts_timer);
        } else {
            ak_pen_release(ts_data, dev);
            ts_data->running = 0;
            tca_tchhw_enableirq();
        }
    }
}


static irqreturn_t ak4183_irq_handler(int irq, void *dev_id)
{
    struct ak4183_ts *ts_data = (struct ak4183_ts *)dev_id;

    if (ts_data->running == 0) {
        ts_data->running = 1;
        ts_timer.expires = jiffies;
        add_timer(&ts_timer);
    }

	return IRQ_HANDLED;
}

static void ts_timer_handler(unsigned long data)
{
    if(ak4183_ts->opened){
        if (schedule_work(&(ak4183_ts->work_q)) == 0) {
            printk("cannot schedule work !!!\n");
            ak4183_ts->running = 0;
        }
    }else {
        tca_tchhw_disableirq();
    }
}



#ifdef CONFIG_PM
static int ak4183ts_suspend(struct platform_device *dev, pm_message_t state)
{
    return 0;
}

static int ak4183ts_resume(struct platform_device *dev)
{
    return 0;
}
#else
#define ak4183ts_suspend     NULL
#define ak4183ts_resume      NULL
#endif

static int __devinit ak4183ts_probe(struct platform_device *pdev)
{
    struct input_dev *input_dev;
    int err = -ENOMEM;

    printk("%s\n", __func__);

    tca_tch_init();

    /* INPUT DEVICE */
    ak4183_ts = kzalloc(sizeof(struct ak4183_ts), GFP_KERNEL);

    input_dev = input_allocate_device();

    if (!ak4183_ts || !input_dev)
    {
        goto fail1;
    }

    platform_set_drvdata(pdev, ak4183_ts);

    ak4183_ts->running = 0;
    ak4183_ts->opened = 1;
    ak4183_ts->pen_status = PEN_RELEASE;

    ak4183_ts->input = input_dev;

    input_dev->name = "tcc-ak4183-ts";
    input_dev->phys = "ak4183/input0";
    input_dev->id.version = 0x0001;
    input_dev->dev.parent = &pdev->dev;

    input_dev->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
    input_dev->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);
    input_set_abs_params(input_dev, ABS_X, 0, 2000, 0, 0);
    input_set_abs_params(input_dev, ABS_Y, 0, 2000, 0, 0);
    input_set_abs_params(input_dev, ABS_PRESSURE, 0, 0, 0, 0);

    INIT_WORK(&(ak4183_ts->work_q), ts_fetch_thread);
    //init_waitqueue_head(&(ak4183_ts->wait_q));
    mutex_init(&(ak4183_ts->mutex));

    init_timer(&ts_timer);
    ts_timer.data = (unsigned long)ak4183_ts;
    ts_timer.function = ts_timer_handler;

    irq_no = tca_tchhw_getirqnum();
	/* Enable touch interrupt */
	if (request_irq(irq_no, ak4183_irq_handler, IRQF_DISABLED, DEVICE_NAME, ak4183_ts))
    {
        err = -EBUSY;
        goto fail1;
    }

	tca_tch_poweroff();

    err = input_register_device(ak4183_ts->input);
    if (err)
        goto fail2;

    ak4183_adc_on();

    printk("ak4183 driver is successfully loaded\n");
	return 0;

fail2: free_irq(irq_no, ak4183_ts);
fail1: input_free_device(input_dev);
       kfree(ak4183_ts);
       return err;
}

static int __devexit ak4183ts_remove(struct platform_device *pdev)
{
    ak4183_ts->opened = 0;
    tca_tchhw_disableirq();
	del_timer_sync(&ts_timer);
    flush_scheduled_work();
	free_irq(irq_no, ak4183_ts);
    input_unregister_device(ak4183_ts->input);
    kfree(ak4183_ts);

#ifdef CONFIG_I2C
    i2c_del_driver(&ak_i2c_driver);
    kfree(ak_i2c_client);
#endif
    return 0;
}

static struct platform_driver ak4183_driver = {
    .driver     = {
        .name   = "tcc-ak4183-ts",
    },
    .probe      = ak4183ts_probe,
    .remove     = ak4183ts_remove,
    .suspend    = ak4183ts_suspend,
    .resume     = ak4183ts_resume,
};


static int __init ak4183ts_init(void)
{

    int res;
    printk("Telechips AK4183 Touch Driver\n");
    res =  platform_driver_register(&ak4183_driver);
    if(res)
    {
        printk("fail : platform driver %s (%d) \n", ak4183_driver.driver.name, res);
        return res;
    }

    return 0;
}


void __exit ak4183ts_exit(void)
{
    platform_driver_unregister(&ak4183_driver);
}

module_init(ak4183ts_init);
module_exit(ak4183ts_exit);

MODULE_AUTHOR("linux@telechips.com");
MODULE_DESCRIPTION("Telechips ak4183 touch driver");
MODULE_LICENSE("GPL");

