/*
 * linux/drivers/spi/spi_tcc79x.c
 *
 * Author:  <linux@telechips.com>
 * Created: 10th Jun, 2008 
 * Description: Driver for TCC79X SPI Controllers
 *
 * Copyright (c) Telechips, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>

#include <linux/spi/spi.h>
#include <linux/spi/spi_bitbang.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <mach/hardware.h>

#include <mach/tcc79x.h>
#include <linux/workqueue.h>

#include <asm/plat-tcc/tcc_ckc.h>

#include "tca_spi_hwset.h"

#define MILLISEC_TO_TICK(X) (((X) * HZ) / 1000)

struct tcc79x_peri_id_table {
    unsigned int bus_id;
    unsigned int clk_id;
};

struct tcc79x_peri_id_table sa_peri_id_table[] = {
    { RB_GPSB0, PERI_GPSB0 },
    { RB_GPSB1, PERI_GPSB1 },
    { RB_GPSB2, PERI_GPSB2 },
    { RB_GPSB3, PERI_GPSB3 },
};

struct tca_spi_pri_handle {
    struct platform_device *pdev;
    int rx_len, rx_cnt;
    int tx_len, tx_cnt;
    struct {
        int xfer_pos;
        int value;
    } done;

    unsigned char stopping;
    wait_queue_head_t wait_q;
    struct list_head queue;
    struct spi_message *current_message;
    spinlock_t lock;
    //struct mutex pm_mutex;
};

static void tea_free_dma_linux(struct tea_dma_buf *tdma)
{
    if (tdma) {
        if (tdma->dma_addr != 0) {
            dma_free_writecombine(0, tdma->buf_size, tdma->v_addr, tdma->dma_addr);
        }
        memset(tdma, 0, sizeof(struct tea_dma_buf));
    }
}

static int tea_alloc_dma_linux(struct tea_dma_buf *tdma, unsigned int size)
{
    int ret = -1;
    if (tdma) {
        tea_free_dma_linux(tdma);
        tdma->buf_size = size;
        tdma->v_addr = dma_alloc_writecombine(0, tdma->buf_size, &tdma->dma_addr, GFP_KERNEL);
        printk("Malloc DMA buffer @0x%X(Phy=0x%X), size:%d\n",
               (unsigned int)tdma->v_addr,
               (unsigned int)tdma->dma_addr,
               tdma->buf_size);
        ret = tdma->v_addr ? 0 : 1;
    }
    return ret;
}

static int tcc79x_spi_start_rxtx(struct spi_master *master, unsigned int flen)
{
    struct tca_spi_handle *tspi = spi_master_get_devdata(master);
    struct tca_spi_pri_handle *tpri = (struct tca_spi_pri_handle *)tspi->private_data;

    tpri->done.value = -EIO;
    //mutex_lock(&(tpri->pm_mutex));

    if ((flen > 0) && (flen <= tspi->rx_dma.buf_size)) {
        tpri->rx_len = flen;
        tpri->tx_len = flen;
        tpri->rx_cnt = 0;
        tpri->tx_cnt = 0;

        tspi->clear_fifo_packet(tspi);
        tspi->set_packet_cnt(tspi, flen);
        tspi->dma_start(tspi);

        if (wait_event_interruptible_timeout((tpri->wait_q),
                                             (tpri->rx_cnt == tpri->rx_len),
                                             msecs_to_jiffies(WAIT_TIME_FOR_DMA_DONE)) == 0) {
            printk("[%s:%d] wait_event timeout  (%dms) !!!\n",
                   __func__, __LINE__, WAIT_TIME_FOR_DMA_DONE);
        } else {
            if (tpri->rx_cnt == tpri->rx_len) {
                tpri->done.value = 0;
            }
        }
    }

    tpri->current_message = NULL;
    //mutex_unlock(&(tpri->pm_mutex));

    return tpri->done.value;
}

static void tcc79x_spi_next_message(struct spi_master *master)
{
    struct tca_spi_handle *tspi = NULL;
    struct spi_message *msg = NULL;
    struct spi_device *spi = NULL;
    struct spi_transfer *xfer = NULL;
    unsigned int flen = 0;
    unsigned int copy_len = 0, total_len = 0;
    struct tca_spi_pri_handle *tpri = NULL;
    unsigned int bit_width = 0;

    tspi = spi_master_get_devdata(master);
    tpri = (struct tca_spi_pri_handle *)tspi->private_data;

    BUG_ON(tpri->current_message);

    msg = list_entry(tpri->queue.next, struct spi_message, queue);
    list_del_init(&(msg->queue));
    tpri->current_message = msg;

    spi = msg->spi;

    list_for_each_entry(xfer, &(msg->transfers), transfer_list) {
        tpri->done.xfer_pos = 0;
        tpri->done.value = 0;
        total_len = xfer->len;
        do {
            copy_len = (total_len > tspi->rx_dma.buf_size) ? tspi->rx_dma.buf_size : total_len;
            if (copy_len & 0x3) {
                copy_len %= 4;
                bit_width = tspi->regs->MODE;
                bit_width = ((bit_width >> 8) & 0x1F) + 1;
                tspi->set_bit_width(tspi, 8);
            }

            if (xfer->tx_buf) {
                memcpy(tspi->tx_dma.v_addr, xfer->tx_buf + tpri->done.xfer_pos, copy_len);
            }

            if (tcc79x_spi_start_rxtx(master, copy_len)) {
                tspi->clear_fifo_packet(tspi);
                goto lb_return;
            }

            if (xfer->rx_buf) {
                memcpy(xfer->rx_buf + tpri->done.xfer_pos, tspi->rx_dma.v_addr, copy_len);
            }

            if (bit_width) {
                tspi->set_bit_width(tspi, bit_width);
                bit_width = 0;
            }

            total_len -= copy_len;
            tpri->done.xfer_pos += copy_len;
        } while (total_len);
        flen += xfer->len;
    }

lb_return:
    msg->status = tpri->done.value;
    if (msg->status == 0) {
        msg->actual_length = flen;
    }
    msg->complete(msg->context);

    if (list_empty(&(tpri->queue)) || tpri->stopping) {
        ;
    } else {
        tcc79x_spi_next_message(master);
    }
}

static irqreturn_t tcc_spi_isr(int irq, void *dev_id)
{
    struct spi_master *master = dev_id;
    struct tca_spi_handle *tspi = spi_master_get_devdata(master);
    //unsigned long flags = 0;
    unsigned long dma_done_reg = 0;
    struct tca_spi_pri_handle *tpri = (struct tca_spi_pri_handle *)tspi->private_data;

    //local_save_flags(flags);
    //local_irq_disable();

    dma_done_reg = tspi->regs->DMAICR;

    if (dma_done_reg & Hw29) {
        tspi->dma_stop(tspi);
        tpri->tx_cnt = tpri->tx_len;
        tpri->rx_cnt = tpri->rx_len;
        wake_up(&(tpri->wait_q));
    }

    //local_irq_restore(flags);
    return IRQ_HANDLED;
}

#define MODEBITS (SPI_CPOL | SPI_CPHA | SPI_CS_HIGH | SPI_LSB_FIRST)
static int tcc79x_spi_setup(struct spi_device *spi)
{
    struct tca_spi_handle *tspi = NULL;
    unsigned int bits = spi->bits_per_word;
    struct tca_spi_pri_handle *tpri = NULL;
    //unsigned int mode = 0;

    tspi = spi_master_get_devdata(spi->master);
    tpri = (struct tca_spi_pri_handle *)tspi->private_data;

    if (tpri->stopping)
        return -ESHUTDOWN;
    if (spi->chip_select > spi->master->num_chipselect) {
        dev_dbg(&spi->dev,
                "setup: invalid chipselect %u (%u defined)\n",
                spi->chip_select, spi->master->num_chipselect);
        return -EINVAL;
    }

    if (bits == 0)
        bits = 8;
    if (bits != 8 && bits != 16 && bits != 32) {
        dev_dbg(&spi->dev, "setup: invalid bits_per_word %u\n", bits);
        return -EINVAL;
    }
    tspi->set_bit_width(tspi, bits);

    if (spi->mode & ~MODEBITS) {
        dev_dbg(&spi->dev, "setup: unsupported mode bits %x\n",
                spi->mode & ~MODEBITS);
        return -EINVAL;
    }

    //mode = tspi->regs->MODE;
    tca_spi_setCPOL(tspi->regs, (spi->mode & SPI_CPOL));
    tca_spi_setCPHA(tspi->regs, (spi->mode & SPI_CPHA));
    tca_spi_setCS_HIGH(tspi->regs, (spi->mode & SPI_CS_HIGH));
    tca_spi_setLSB_FIRST(tspi->regs, (spi->mode & SPI_LSB_FIRST));
    //tspi->regs->MODE = mode;

    return 0;
}

static int tcc79x_spi_transfer(struct spi_device *spi, struct spi_message *msg)
{
    struct tca_spi_handle *tspi = NULL;
    //unsigned long flags = 0;
    struct tca_spi_pri_handle *tpri = NULL;

    tspi = (struct tca_spi_handle *)spi_master_get_devdata(spi->master);
    tpri = (struct tca_spi_pri_handle *)tspi->private_data;

    if (unlikely(list_empty(&(msg->transfers)))) {
        return -EINVAL;
    }

    if (tpri->stopping) {
        return -ESHUTDOWN;
    }

    msg->status = -EINPROGRESS;
    msg->actual_length = 0;

    //spin_lock_irqsave(&(tpri->lock), flags);
    list_add_tail(&(msg->queue), &(tpri->queue));
    if (!tpri->current_message) {
        tcc79x_spi_next_message(spi->master);
    }
    //spin_unlock_irqrestore(&(tpri->lock), flags);
    return 0;
}

static void tcc79x_spi_cleanup(struct spi_device *spi)
{
    if (!spi->controller_state)
        return;
    dev_dbg(&spi->dev, "tcc7900_spi_cleanup\n");
}

static int __init tcc79x_spi_probe(struct platform_device *pdev)
{
    struct resource *regs = NULL;
    int irq = 0, ret = 0;
    struct spi_master *master = NULL;
    struct tca_spi_handle *tspi = NULL;
    struct tca_spi_pri_handle *tpri = NULL;

    printk("[SPI] probe  ^^ [%s:%d]\n", __func__, __LINE__);

    regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
    if (!regs) {
        return -ENXIO;
    }
    irq = platform_get_irq(pdev, 0);
    if (irq < 0) {
        return -ENXIO;
    }

    master = spi_alloc_master(&(pdev->dev), sizeof(struct tca_spi_handle));
    if (!master) {
        ret = -ENOMEM;
        goto LB_ret_put;
    }

    tpri = (struct tca_spi_pri_handle *)kmalloc(sizeof(struct tca_spi_pri_handle), GFP_KERNEL);
    if (tpri == NULL) {
        ret = -ENOMEM;
        goto LB_ret_put;
    }
    memset(tpri, 0, sizeof(struct tca_spi_pri_handle));

    tca_ckc_com_setperibus(sa_peri_id_table[pdev->id].bus_id, 1);
    tca_ckc_com_setperi(sa_peri_id_table[pdev->id].clk_id, KHz, 40000, DIRECTPLL0);

    master->bus_num = pdev->id;
    master->num_chipselect = 1;
    master->setup = tcc79x_spi_setup;
    master->transfer = tcc79x_spi_transfer;
    master->cleanup = tcc79x_spi_cleanup;
    platform_set_drvdata(pdev, master);

    tspi = spi_master_get_devdata(master);
    if (tca_spi_init(tspi,
                     (volatile struct tca_spi_regs *)regs->start,
                     irq,
                     tea_alloc_dma_linux,
                     tea_free_dma_linux,
                     pdev->id,
                     0)) {
        goto LB_ret_put;
    }

    tspi->private_data = (void *)tpri;
    tpri->pdev = pdev;

    spin_lock_init(&(tpri->lock));
    INIT_LIST_HEAD(&(tpri->queue));
    //mutex_init(&(tpri->pm_mutex));

    tspi->clear_fifo_packet(tspi);
    tspi->dma_stop(tspi);
    init_waitqueue_head(&(tpri->wait_q));

    ret = spi_register_master(master);
    if (ret) { goto LB_ret_free; }
    
    tspi->hw_init(tspi);

    ret = request_irq(tspi->irq, tcc_spi_isr, IRQF_SHARED, pdev->dev.bus_id, master);
    if (ret) { goto LB_ret_free; }

    return 0;

LB_ret_free:
    tca_spi_clean(tspi);

LB_ret_put:
    spi_master_put(master);
    if (tpri) { kfree(tpri); }

    return ret;
}

static int __exit tcc79x_spi_remove(struct platform_device *pdev)
{
    struct spi_master *master = platform_get_drvdata(pdev);
    struct tca_spi_handle *tspi = NULL;
    struct spi_message *msg = NULL;
    struct tca_spi_pri_handle *tpri = NULL;

    tspi = spi_master_get_devdata(master);
    tpri = (struct tca_spi_pri_handle *)tspi->private_data;

    spin_lock_irq(&(tpri->lock));
    tpri->stopping = 1;
    spin_unlock_irq(&tpri->lock);

    list_for_each_entry(msg, &(tpri->queue), queue) {
        msg->status = -ESHUTDOWN;
        msg->complete(msg->context);
    }

    free_irq(tspi->irq, master);

    //tea_free_dma_linux(&(tspi->tx_dma));
    //tea_free_dma_linux(&(tspi->rx_dma));
    
    if (tspi->private_data) { kfree(tspi->private_data); }

    tca_spi_clean(tspi);

    spi_unregister_master(master);
    tca_ckc_com_setperibus(sa_peri_id_table[pdev->id].bus_id, 0);

    return 0;
}

#ifdef CONFIG_PM
static int tcc79x_spi_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct spi_master *master = platform_get_drvdata(pdev);
	struct tca_spi_handle *tspi = spi_master_get_devdata(master);

    if (tspi) {
        (void)(state);
    }
	return 0;
}

static int tcc79x_spi_resume(struct platform_device *pdev)
{
	struct spi_master *master = platform_get_drvdata(pdev);
	struct tca_spi_handle *tspi = spi_master_get_devdata(master);

    if (tspi) {
        tspi->clear_fifo_packet(tspi);

#if 0
        if (tspi->is_enable_dma(tspi)) {
            tspi->dma_stop(tspi);
            tspi->dma_start(tspi);
        }
#endif

        if (tspi->id == 0) {
            HwGPSB_PCFG = (HwGPSB_PCFG & ~0x000000FF) | (0x6);
        } else {
            HwGPSB_PCFG = (HwGPSB_PCFG & ~0x0000FF00) | (0x500);
        }
        //mutex_unlock(&(tpri->pm_mutex));
    }

	return 0;
}
#else
#define	tcc79x_spi_suspend	NULL
#define	tcc79x_spi_resume	NULL
#endif

static struct platform_driver tcc79x_spidrv = {
    .remove = __exit_p(tcc79x_spi_remove),
	.suspend	= tcc79x_spi_suspend,
	.resume		= tcc79x_spi_resume,
    .driver = {
        .name = "tcc79x-spi",
        .owner = THIS_MODULE,
    },
};

static int __init tcc79x_spi_init(void)
{
    return platform_driver_probe(&tcc79x_spidrv, tcc79x_spi_probe);
}

static void __exit tcc79x_spi_exit(void)
{
    platform_driver_unregister(&tcc79x_spidrv);
}

module_init(tcc79x_spi_init);
module_exit(tcc79x_spi_exit);

MODULE_DESCRIPTION("TCC79X SPI Driver");
MODULE_LICENSE("GPL");

