
/* linux/drivers/spi/spi_tcc83x.c */

#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>

#include <linux/spi/spi.h>
#include <linux/spi/spi_bitbang.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/hardware.h>

#include <asm/arch/tcc83x_virt.h>
#include <linux/workqueue.h>

#include <asm/plat-tcc/tcc_ckc.h>

#define TCC_MAX_FRAME_BYTE 256
#define TCC_TX_FIFO_CNT 8
#define TCC_TX_FIFO_UNIT_SIZE 4
#define TCC_TX_FIFO_THRESHOLD 4
#define TCC_TX_FIFO_WRITEABLE_CNT (TCC_TX_FIFO_CNT - TCC_TX_FIFO_THRESHOLD)
#define TCC_RX_FIFO_THRESHOLD 1

#define MAX_WAIT_TIME 5000
#define MILLISEC_TO_TICK(X) (((X) * HZ) / 1000)

#define TCC_MAX_DMA_BYTE 4096

struct tcc_gdma_info {
    void *tx_base_addr;
    void *rx_base_addr;
    int intr_num;
    unsigned long reg_pos;
    unsigned long sdo_phy_addr;
    unsigned long sdi_phy_addr;
};

typedef struct _tcc_dma_buf_t {
    char *addr;
    dma_addr_t dma_addr;
    int  buf_size; // total size of DMA
} tcc_dma_buf_t;

#pragma pack(push, 4)
struct tcc83x_spi_regs {
    volatile unsigned long SDO, SDI, SBCR, SCR, SPFRMPOS, SPCTRL, SPSTS;
};

struct tcc83x_dma_regs {
    volatile unsigned long ST_SADR, SPARAM, DUMMY, C_SADR, ST_DADR, DPARAM, DUMMY2;
    volatile unsigned long C_DADR, HCOUNT, CHCTRL, RPTCTRL, EXTREQ;
};
#pragma pack(pop)

struct tcc83x_spi {
    spinlock_t lock;
    int id;
    int tx_irq, rx_irq, dma_irq;
    struct platform_device *pdev;
    volatile struct tcc83x_spi_regs *spi_regs;
    volatile struct tcc83x_dma_regs *dma_tx_regs;
    volatile struct tcc83x_dma_regs *dma_rx_regs;
    unsigned long sdo_phy_addr;
    unsigned long sdi_phy_addr;

    u8 stopping;
    struct list_head queue;
    struct spi_message *current_message;

    int rx_len, rx_cnt;
    int tx_len, tx_cnt;

    int rx_threshold;
    wait_queue_head_t wait_q;

    struct {
        int xfer_pos;
        int value;
    } done;

    tcc_dma_buf_t in_buffer;
    tcc_dma_buf_t out_buffer;
};

/* #set RCFul, ROR, RUR  #clear RCFul, REmp, RFul, ROR, RUR at SPCTRL */
#define tcc83x_en_rxint(tspi) BITSET((tspi)->spi_regs->SPCTRL, Hw12 | Hw9 | Hw8)
#define tcc83x_dis_rxint(tspi) BITCLR((tspi)->spi_regs->SPCTRL, Hw12 | Hw11 | Hw10 | Hw9 | Hw8)

/* #set TCEmpEn, TOREn, TUREn  #clear TCEmpEn, TEmpEn, TFulEn, TOREn, TUREn at SPCTRL */
#define tcc83x_en_txint(tspi) BITSET((tspi)->spi_regs->SPCTRL, Hw4 | Hw1 | Hw0)
#define tcc83x_dis_txint(tspi) BITCLR((tspi)->spi_regs->SPCTRL, Hw4 | Hw3 | Hw2 | Hw1 | Hw0)

/* #set EN  #clear EN at SCR */
#define tcc83x_en_spi(tspi) BITSET((tspi)->spi_regs->SCR, Hw31)
#define tcc83x_dis_spi(tspi) BITCLR((tspi)->spi_regs->SCR, Hw31)

/* #set IEN, EN  #clear IEN, EN at CHCTRL */
#define tcc83x_spi_dma_tx_start(tspi) BITSET((tspi)->dma_tx_regs->CHCTRL, Hw0 | Hw2)
#define tcc83x_spi_dma_tx_stop(tspi)  BITCLR((tspi)->dma_tx_regs->CHCTRL, Hw0 | Hw2)

/* #set IEN, EN  #clear IEN, EN at CHCTRL */
#define tcc83x_spi_dma_rx_start(tspi) BITSET((tspi)->dma_rx_regs->CHCTRL, Hw0 | Hw2)
#define tcc83x_spi_dma_rx_stop(tspi)  BITCLR((tspi)->dma_rx_regs->CHCTRL, Hw0 | Hw2)

static void tcc83x_set_rx_threshold(struct tcc83x_spi *tspi, unsigned int th)
{
    th &= 0x7;
    tspi->rx_threshold = th;
 
    /* clear RxCnt and set RxCnt */
    BITCLR(tspi->spi_regs->SPCTRL, Hw13 | Hw14 | Hw15);
    BITSET(tspi->spi_regs->SPCTRL, (th << 13));
}

static void tcc83x_set_tx_threshold(struct tcc83x_spi *tspi, unsigned int th)
{
    BITCLR(th, ~(Hw0 | Hw1 | Hw2));

    /* clear TxCnt and set TxCnt */
    BITCLR(tspi->spi_regs->SPCTRL, Hw5 | Hw6 | Hw7);
    BITSET(tspi->spi_regs->SPCTRL, (th << 5));
}

static void tcc83x_set_frame_length(struct tcc83x_spi *tspi, unsigned int len)
{
    len = (len - 1) & 0xFF;

    /* clear FRM_LEN and set FRM_LEN */
    BITCLR(tspi->spi_regs->SCR,  Hw16 | Hw17 | Hw18 | Hw19 | Hw20 | Hw21 | Hw22 | Hw23);
    BITSET(tspi->spi_regs->SCR, len << 16);
}

static void tcc83x_clear_fifos(struct tcc83x_spi *tspi)
{
    unsigned int reg = 0;
    /* #set TEmpEn  #clear TEmpEnat SPCTRL */
    BITSET(tspi->spi_regs->SPCTRL, Hw3);
    tcc83x_en_spi(tspi);
    while (ISCLR(tspi->spi_regs->SPSTS, Hw3)) {
        schedule();
    }
    BITCLR(tspi->spi_regs->SPCTRL, Hw3);

    udelay(2000);
    udelay(1200);

    tcc83x_dis_spi(tspi);
    /* #set REmpEn  #clear REmpEnat SPCTRL */
    BITSET(tspi->spi_regs->SPCTRL, Hw11);
    while (ISCLR(tspi->spi_regs->SPSTS, Hw11)) {
        reg = HwSDI;
        schedule();
    }
    BITCLR(tspi->spi_regs->SPCTRL, Hw11);
}

static void *tcc_free_dma_buf(tcc_dma_buf_t *tdma)
{
    if (tdma) {
        if (tdma->dma_addr != 0) {
            dma_free_writecombine(0, tdma->buf_size, tdma->addr, tdma->dma_addr);
        }
        memset(tdma, 0, sizeof(tcc_dma_buf_t));
    }
    return NULL;
}

static void *tcc_malloc_dma_buf(tcc_dma_buf_t *tdma, int buf_size)
{
    if (tdma) {
        tcc_free_dma_buf(tdma);
        tdma->buf_size = buf_size;
        tdma->addr = dma_alloc_writecombine(0, tdma->buf_size, &tdma->dma_addr, GFP_KERNEL);
        printk("Malloc DMA buffer @0x%X(Phy=0x%X), size:%d\n",
               (unsigned int)tdma->addr,
               (unsigned int)tdma->dma_addr,
               tdma->buf_size);
        return tdma->addr;
    }
    return NULL;
}

static void spi_set_in_dma_buf(struct tcc83x_spi *tspi, unsigned int length)
{
    /* clear EN at CHCTRL */
    BITCLR(tspi->dma_rx_regs->CHCTRL, Hw0);

    tspi->dma_rx_regs->ST_SADR = tspi->sdi_phy_addr;
    tspi->dma_rx_regs->ST_DADR = tspi->in_buffer.dma_addr;
    tspi->dma_rx_regs->SPARAM = 0;

    if (length >= 4) {
        tspi->dma_rx_regs->DPARAM = 0x0 | 4;
        tspi->dma_rx_regs->HCOUNT = length / 4;
    } else {
        tspi->dma_rx_regs->DPARAM = 0x0 | length;
        tspi->dma_rx_regs->HCOUNT = 1;
    }
        
    tspi->dma_rx_regs->CHCTRL &= 0x00000001;
    /* set FLAG, WSIZE(32bit), TYPE(transfer with level),  at CHCTRL */
    BITSET(tspi->dma_rx_regs->CHCTRL, (Hw3 | Hw5 | Hw8 | Hw9));

    /* clear DRI at RPTCTRL */
	BITCLR(tspi->dma_rx_regs->RPTCTRL, Hw31);

    /* set SPI[0,1] Receive */
    if (tspi->id == 0) {
        BITSET(tspi->dma_rx_regs->EXTREQ, Hw9);
    } else {
        BITSET(tspi->dma_rx_regs->EXTREQ, Hw12);
    }
}

static void spi_set_out_dma_buf(struct tcc83x_spi *tspi, unsigned int length)
{
    /* clear EN at CHCTRL */
	BITCLR(tspi->dma_tx_regs->CHCTRL, Hw0);
	tspi->dma_tx_regs->ST_SADR = tspi->out_buffer.dma_addr;
    tspi->dma_tx_regs->ST_DADR = tspi->sdo_phy_addr;

    tspi->dma_tx_regs->DPARAM = 0;

    if (length >= 4) {
        (tspi)->dma_tx_regs->SPARAM = 0x0 | 4;
        (tspi)->dma_tx_regs->HCOUNT = length / 4;
    } else {
        (tspi)->dma_tx_regs->SPARAM = 0x0 | length;
        (tspi)->dma_tx_regs->HCOUNT = 1;
    }

    tspi->dma_tx_regs->CHCTRL &= 0x00000001;
    /* set FLAG, WSIZE(32bit), TYPE(transfer with level),  at CHCTRL */
    BITSET(tspi->dma_tx_regs->CHCTRL, (Hw3 | Hw5 | Hw8 | Hw9));

    /* clear DRI at RPTCTRL */
	BITCLR(tspi->dma_tx_regs->RPTCTRL, Hw31);

    /* set SPI[0,1] Transmite */
    if (tspi->id == 0) {
        BITSET(tspi->dma_tx_regs->EXTREQ, Hw6);
    } else {
        BITSET(tspi->dma_tx_regs->EXTREQ, Hw11);
    }
}

static int tcc83x_spi_start_rxtx(struct spi_master *master, unsigned int flen)
{
    struct tcc83x_spi *tspi = spi_master_get_devdata(master);

    tspi->done.value = -EIO;
    if ((flen > 0) && (flen <= TCC_MAX_DMA_BYTE)) {
        tspi->rx_len = flen;
        tspi->tx_len = flen;
        tspi->rx_cnt = 0;
        tspi->tx_cnt = 0;

        spi_set_in_dma_buf(tspi, flen);
        spi_set_out_dma_buf(tspi, flen);

        tcc83x_spi_dma_rx_start(tspi); 
        tcc83x_spi_dma_tx_start(tspi); 

        tcc83x_set_frame_length(tspi,
                                (flen > TCC_MAX_FRAME_BYTE) ? TCC_MAX_FRAME_BYTE : flen);
        tcc83x_en_rxint(tspi);
        tcc83x_en_txint(tspi);

        tcc83x_en_spi(tspi);


        if (wait_event_interruptible_timeout((tspi->wait_q),
                                             (tspi->rx_cnt == tspi->rx_len),
                                             MILLISEC_TO_TICK(MAX_WAIT_TIME)) == 0) {
            printk("[%s:%d] wait_event timeout  (%dms) !!!\n",
                   __func__, __LINE__, MAX_WAIT_TIME);
        } else {
            if (tspi->rx_cnt == tspi->rx_len) {
                tspi->done.value = 0;
            }
        }
    }

    tspi->current_message = NULL;

    tcc83x_dis_spi(tspi);

    tcc83x_dis_rxint(tspi);
    tcc83x_dis_txint(tspi);

    tcc83x_spi_dma_rx_stop(tspi); 
    tcc83x_spi_dma_tx_stop(tspi); 

    return tspi->done.value;
}

static void tcc83x_spi_next_message(struct spi_master *master)
{
    struct tcc83x_spi *tspi = NULL;
    struct spi_message *msg = NULL;
    struct spi_device *spi = NULL;
    struct spi_transfer *xfer = NULL;
    unsigned int flen = 0;
    unsigned int copy_len = 0, total_len = 0;

    tspi = spi_master_get_devdata(master);

    BUG_ON(tspi->current_message);

    msg = list_entry(tspi->queue.next, struct spi_message, queue);
    list_del_init(&(msg->queue));
    tspi->current_message = msg;

    spi = msg->spi;

    list_for_each_entry(xfer, &(msg->transfers), transfer_list) {
        tspi->done.xfer_pos = 0;
        tspi->done.value = 0;
        total_len = xfer->len;
        do {
            copy_len = (total_len > TCC_MAX_DMA_BYTE) ? TCC_MAX_DMA_BYTE : total_len;
            if ((copy_len % TCC_TX_FIFO_UNIT_SIZE) && (copy_len > TCC_TX_FIFO_UNIT_SIZE)) {
                copy_len -= (copy_len % TCC_TX_FIFO_UNIT_SIZE);
            }

            if (copy_len > TCC_MAX_FRAME_BYTE) {
                if (copy_len % TCC_MAX_FRAME_BYTE) {
                    copy_len -= (copy_len % TCC_MAX_FRAME_BYTE);
                }
            }

            if (xfer->tx_buf) {
                memcpy(tspi->out_buffer.addr, xfer->tx_buf + tspi->done.xfer_pos, copy_len);
            }

            if (tcc83x_spi_start_rxtx(master, copy_len)) {
                tcc83x_clear_fifos(tspi);
                goto lb_return;
            }

            if (xfer->rx_buf) {
                memcpy(xfer->rx_buf + tspi->done.xfer_pos, tspi->in_buffer.addr, copy_len);
            }

            total_len -= copy_len;
            tspi->done.xfer_pos += copy_len;
        } while (total_len);
        flen += xfer->len;
    }

lb_return:
    msg->status = tspi->done.value;
    if (msg->status == 0) {
        msg->actual_length = flen;
    }
    msg->complete(msg->context);

    if (list_empty(&(tspi->queue)) || tspi->stopping) {
        ;
    } else {
        tcc83x_spi_next_message(master);
    }
}

static irqreturn_t tcc83x_spi_irq_dma(int irq, void *dev_id)
{
    struct spi_master *master = dev_id;
    struct tcc83x_spi *tspi = spi_master_get_devdata(master);
    unsigned long flags = 0;

    local_save_flags(flags);
    local_irq_disable();

	if ((tspi)->dma_tx_regs->CHCTRL & Hw3) {
        (tspi)->dma_tx_regs->CHCTRL |= Hw3;
        tcc83x_spi_dma_tx_stop(tspi);
        tspi->tx_cnt = tspi->tx_len;
    }

    if ((tspi)->dma_rx_regs->CHCTRL & Hw3) {
        (tspi)->dma_rx_regs->CHCTRL |= Hw3;
        tcc83x_spi_dma_rx_stop(tspi);
        tspi->rx_cnt = tspi->rx_len;
        wake_up(&(tspi->wait_q));
    }

    local_irq_restore(flags);
    return IRQ_HANDLED;
}

#define MODEBITS (SPI_CPOL | SPI_CPHA | SPI_CS_HIGH | SPI_LSB_FIRST)
static int tcc83x_spi_setup(struct spi_device *spi)
{
    struct tcc83x_spi *tspi = NULL;
    unsigned int bits = spi->bits_per_word;
    unsigned int scr = 0;

    tspi = spi_master_get_devdata(spi->master);

    if (tspi->stopping)
        return -ESHUTDOWN;
    if (spi->chip_select > spi->master->num_chipselect) {
        dev_dbg(&spi->dev,
                "setup: invalid chipselect %u (%u defined)\n",
                spi->chip_select, spi->master->num_chipselect);
        return -EINVAL;
    }

    if (bits == 0)
        bits = 8;
    if (bits != 8 && bits != 16 && bits !=32) {
        dev_dbg(&spi->dev, "setup: invalid bits_per_word %u\n", bits);
        return -EINVAL;
    }

    if (spi->mode & ~MODEBITS) {
        dev_dbg(&spi->dev, "setup: unsupported mode bits %x\n",
                spi->mode & ~MODEBITS);
        return -EINVAL;
    }

    scr = (tspi)->spi_regs->SCR;
    scr = (spi->mode & SPI_CPOL) ? (scr & ~Hw29) : (scr | Hw29);
    scr = (spi->mode & SPI_CPHA) ? (scr | Hw28) : (scr & ~Hw28);
    scr = (spi->mode & SPI_CS_HIGH) ? (scr | Hw12) : (scr & ~Hw12);
    scr = (spi->mode & SPI_LSB_FIRST) ? (scr & ~Hw27) : (scr | Hw27);
    (tspi)->spi_regs->SCR = scr;

    return 0;
}

static int tcc83x_spi_transfer(struct spi_device *spi, struct spi_message *msg)
{
    struct tcc83x_spi *tspi = NULL;
    unsigned long flags = 0;

    tspi = (struct tcc83x_spi *)spi_master_get_devdata(spi->master);

    if (unlikely(list_empty(&(msg->transfers)))) {
        return -EINVAL;
    }

    if (tspi->stopping) {
        return -ESHUTDOWN;
    }

    msg->status = -EINPROGRESS;
    msg->actual_length = 0;

    spin_lock_irqsave(&(tspi->lock), flags);
    list_add_tail(&(msg->queue), &(tspi->queue));
    if (!tspi->current_message) {
        tcc83x_spi_next_message(spi->master);
    }
    spin_unlock_irqrestore(&(tspi->lock), flags);
    return 0;
}

static void tcc83x_spi_cleanup(struct spi_device *spi)
{
    if (!spi->controller_state)
        return;
    dev_dbg(&spi->dev, "tcc8300_spi_cleanup\n");
}

static int __init tcc83x_spi_probe(struct platform_device *pdev)
{
    struct resource *regs = NULL;
    int tx_irq = 0, rx_irq = 0, ret = 0;
    struct spi_master *master = NULL;
    struct tcc83x_spi *tspi = NULL;
    unsigned long dma_reg_pos = 0;

    static struct tcc_gdma_info gs_gdma_info[2] = {
        { (void *)0xF0004000, (void *)0xF0004030, IRQ_DMA, Hw13, 0x9000C000, 0x9000C004 },
        { (void *)0xF000A000, (void *)0xF000A030, IRQ_GDMA1, Hw3, 0x9000E000, 0x9000E004 },
    };


    if (pdev->id > 1) {
        printk("[TCC SPI] - Cannot support spi[%d]\n", pdev->id);
        return -1;
    }

    printk("[TCC SPI] - DMA mode DMA\n");
    regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
    if (!regs) {
        return -ENXIO;
    }
    tx_irq = platform_get_irq(pdev, 0);
    rx_irq = platform_get_irq(pdev, 1);
    if (tx_irq < 0 || rx_irq < 0) {
        return -ENXIO;
    }

    master = spi_alloc_master(&(pdev->dev), sizeof(struct tcc83x_spi));
    if (!master) {
        ret = -ENOMEM;
        goto out_free;
    }

    master->bus_num = pdev->id;
    master->num_chipselect = 1;
    master->setup = tcc83x_spi_setup;
    master->transfer = tcc83x_spi_transfer;
    master->cleanup = tcc83x_spi_cleanup;
    platform_set_drvdata(pdev, master);

    tspi = spi_master_get_devdata(master);

    memset(tspi, 0, sizeof(struct tcc83x_spi));

    tspi->id = pdev->id;

    spin_lock_init(&(tspi->lock));
    INIT_LIST_HEAD(&(tspi->queue));

    tspi->pdev = pdev;
    tspi->spi_regs = (volatile struct tcc83x_spi_regs *)regs->start;
    tspi->tx_irq = tx_irq;
    tspi->rx_irq = rx_irq;
    if (tspi->spi_regs == 0) {
        goto out_free_buffer;
    }
    init_waitqueue_head(&(tspi->wait_q));

    tspi->dma_irq = gs_gdma_info[tspi->id].intr_num;
    tspi->dma_tx_regs = (volatile struct tcc83x_dma_regs *)(gs_gdma_info[tspi->id].tx_base_addr);
    tspi->dma_rx_regs = (volatile struct tcc83x_dma_regs *)(gs_gdma_info[tspi->id].rx_base_addr);
    tspi->sdo_phy_addr = gs_gdma_info[tspi->id].sdo_phy_addr;
    tspi->sdi_phy_addr = gs_gdma_info[tspi->id].sdi_phy_addr;

    if (tspi->id == 0) {
        BITCLR(HwNGSEL_B, Hw3 | Hw2 | Hw1 | Hw0); /* GPIO_B[3:0] is not general I/O */
        BITCLR(HwGSEL_B, Hw10 | Hw18); /* GPIO_B[3:0] is not use UART1 */
        BITCLR(HwGSEL_F, Hw14);
        BITSET(HwGSEL_B, Hw14); /* GPIO_B[3:0] is SPI0 */

        BITCLR(HwIEN, Hw9 | Hw28);

        BITSET(HwBCLKCTR, Hw24); /* SPI0 Master0 Clock Enable */
        BITSET(HwSWRESET, Hw24); /* SPI0 Master0 Reset Enable */
        BITCLR(HwSWRESET, Hw24); /* SPI0 Master0 Reset Disable */

        tca_ckc_setperi(PERI_SPIMS0, KHz, 40000, DIRECTPLL0);
    } else {
        BITCLR(HwNGSEL_B, Hw12 | Hw13 | Hw14 | Hw15); /* GPIO_B[15:12] is not general I/O */
        BITCLR(HwGSEL_B, Hw13 | Hw20); /* GPIO_B[15:12] is not use UART1 or CD-Interface */
        BITCLR(HwGSEL_F, Hw15);
        BITSET(HwGSEL_B, Hw15); /* GPIO_B[15:12] is SPI1 */

        BITCLR(HwIEN_B, Hw10 | Hw11);

        BITSET(HwBCLKCTR, Hw30); /* SPI1 Master0 Clock Enable */
        BITSET(HwSWRESET, Hw30); /* SPI1 Master0 Reset Enable */
        BITCLR(HwSWRESET, Hw30); /* SPI1 Master0 Reset Disable */

        tca_ckc_setperi(PERI_SPIMS1, KHz, 40000, DIRECTPLL0);
    }

    dma_reg_pos = gs_gdma_info[tspi->id].reg_pos;

    /* set DMA interrupt flag */
    BITCLR(HwIEN, dma_reg_pos);
    BITSET(HwSEL, dma_reg_pos);
    BITSET(HwMODE, dma_reg_pos);
    BITCLR(HwPOL, dma_reg_pos);
    BITSET(HwIEN, dma_reg_pos);


    tspi->spi_regs->SBCR = 0;
    tcc83x_set_rx_threshold(tspi, TCC_RX_FIFO_THRESHOLD);
    tcc83x_set_tx_threshold(tspi, TCC_TX_FIFO_THRESHOLD);
    tspi->spi_regs->SCR = (Hw30 | /* Master */
                           Hw29 | /* SCK Polarity */
                           Hw27 | /* MSB/LSB */
                           Hw26 | /* TDO Enable */
                           (TCC_MAX_FRAME_BYTE - 1) << 16 | /* frame length */
                           Hw14 /* Tx Mode Enable */);

    BITSET((tspi)->spi_regs->SCR, Hw10); /* Clear RX FIFO */
    BITSET((tspi)->spi_regs->SCR, Hw9);  /* Clear TX FIFO */
    BITSET((tspi)->spi_regs->SCR, Hw26); /* Enable TX Data Out */

    tcc83x_spi_dma_tx_stop(tspi);
    tcc83x_spi_dma_rx_stop(tspi);

    ret = request_irq(tspi->dma_irq, tcc83x_spi_irq_dma, IRQF_SHARED, pdev->dev.bus_id, master);
    if (ret) { goto out_unmap_regs1; }

    tcc83x_clear_fifos(tspi);
    tcc83x_dis_spi(tspi);

    ret = spi_register_master(master);
    if (ret) { goto out_reset_hw; }
    
    if (!tcc_malloc_dma_buf(&(tspi->out_buffer), TCC_MAX_DMA_BYTE)) {
        ret = -ENOMEM;
        goto out_reset_hw;
    }
    if (!tcc_malloc_dma_buf(&(tspi->in_buffer), TCC_MAX_DMA_BYTE)) {
        ret = -ENOMEM;
        goto out_reset_hw;
    }

    return 0;

out_reset_hw:
    free_irq(tspi->dma_irq, master);
out_unmap_regs1:
out_free_buffer:
out_free:
    spi_master_put(master);

    tcc_free_dma_buf(&(tspi->out_buffer));
    tcc_free_dma_buf(&(tspi->in_buffer));

    return ret;
}

static int __exit tcc83x_spi_remove(struct platform_device *dev)
{
    struct spi_master *master = platform_get_drvdata(dev);
    struct tcc83x_spi *tspi = NULL;
    struct spi_message *msg = NULL;

    tspi = spi_master_get_devdata(master);

    spin_lock_irq(&(tspi->lock));
    tspi->stopping = 1;
    spin_unlock_irq(&tspi->lock);

    list_for_each_entry(msg, &(tspi->queue), queue) {
        msg->status = -ESHUTDOWN;
        msg->complete(msg->context);
    }

    free_irq(tspi->dma_irq, master);

    tcc_free_dma_buf(&(tspi->out_buffer));
    tcc_free_dma_buf(&(tspi->in_buffer));
    
    spi_unregister_master(master);

    return 0;
}

static struct platform_driver tcc83x_spidrv = {
    .remove = __exit_p(tcc83x_spi_remove),
    .driver = {
        .name = "tcc83x-spi",
        .owner = THIS_MODULE,
    },
};

static int __init tcc83x_spi_init(void)
{
    return platform_driver_probe(&tcc83x_spidrv, tcc83x_spi_probe);
}

static void __exit tcc83x_spi_exit(void)
{
    platform_driver_unregister(&tcc83x_spidrv);
}

module_init(tcc83x_spi_init);
module_exit(tcc83x_spi_exit);

MODULE_DESCRIPTION("TCC83X SPI Driver");
MODULE_LICENSE("GPL");

