/*
 * linux/drivers/usb/gadget/tcc_udc.c
 *
 * Author:  <linux@telechips.com>
 * Created: 10th Jun, 2008 
 * Description: TCC83x on-chip full speed USB device controllers
 *
 * Copyright (C) 2008 Telechips, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * This driver handles the USB Device Controller (UDC) in TCC series processors.
 * There are three endpoints, in addition to ep0.
 *
 * Such controller drivers work with a gadget driver.  The gadget driver
 * returns descriptors, implements configuration and data protocols used
 * by the host to interact with this device, and allocates endpoints to
 * the different protocol interfaces.  The controller driver virtualizes
 * usb hardware so that the gadget drivers will be more portable.
 */

#ifdef CONFIG_ARCH_TCC79X
#include <mach/tcc79x.h>
#elif CONFIG_ARCH_TCC83X
#include <mach/tcc83x_virt.h>
#else
#error unsupported arch!
#endif
#include "tcc_udc.h"
#include <linux/kthread.h>
#include <linux/delay.h>

#define	DRIVER_VERSION	"v1.0"
#define	DRIVER_DESC	"Telechips USB Device Controller driver"


#ifdef CONFIG_ARCH_TCC79X
#define USE_MCU_ACCESS
#define EP_WRITE_INT
//#define USE_DMA_ACCESS
//#define USB_DMA_CHANNEL		3
//#define USE_DMA_IRQ
#elif CONFIG_ARCH_TCC83X
//#define USE_MCU_ACCESS
#define EP_WRITE_INT
#define USE_DMA_ACCESS
#define USE_DMA_IRQ
#endif

static const char driver_name [] = "tcc-udc";

static const char ep0name [] = "ep0";

static u32 s_usbt_ep0_length=0;
static u8 ALIGN_32 s_usbt_ep0_buffer[64];


static inline int usbt_is_stalled(struct usbt_ep_t *ep)
{
	int isStall;
	u32 IR_Save;
	u32 nEndPoint;

	if( !ep->desc )
		return 0;

	nEndPoint = ep->index;

	IR_Save	= HwIR;
	HwIR = nEndPoint;

	if ( nEndPoint )
		isStall = ( HwECR & HwECR_ESS_SET ) ? 0x01 : 0x00;
	else
		isStall = ( HwEP0CR & HwEP0CR_ESS_SET ) ? 0x01 : 0x00;

	HwIR	= IR_Save;

	return isStall;
}

static void usbt_unstall(struct usbt_ep_t *ep)
{
	u32 dwIRSave;
	u32 nEndPoint;

	if( !ep->desc )
		return;

	nEndPoint = ep->index;

	dwIRSave	= HwIR;
	HwIR		= nEndPoint;

	if ( nEndPoint )
	{
		HwECR		&= HwECR_ESS_CLR;

		HwECR	= HwECR;
		HwESR	= HwESR;
		HwSSR	= HwSSR;
	}
	else
	{
		HwEP0CR		&= HwEP0CR_ESS_CLR;
	}

	HwIR = dwIRSave;
}

static void usbt_stall(struct usbt_ep_t *ep)
{
	u32 dwIRSave;
	u32 nEndPoint;

	if( !ep->desc )
		return;

	nEndPoint = ep->index;

	dwIRSave	= HwIR;
	HwIR		= nEndPoint;

	if ( nEndPoint )
		HwECR		|= HwECR_ESS_SET;
	else
		HwEP0CR		|= HwEP0CR_ESS_SET;

	HwIR = dwIRSave;
}

static void usbt_unhalt(struct usbt_ep_t *ep)
{
	if( ep->index )
	{
		u32 HwIRSave = HwIR;
		HwIR = ep->index;
		HwECR |= HwECR_CDP_CLR;
		HwIR = HwIRSave;
	}
	usbt_unstall(ep);
}

static void usbt_testmode(u16 wIndex)
{
	int i;
	static const u16 TEST_Packet[27] = {
		/*0x0000, USB Phy.'s bug fix*/ 0x0000, 0x0000, 0x0000, 0xAA00,
		0xAAAA, 0xAAAA, 0xAAAA, 0xEEAA, 0xEEEE, 0xEEEE, 0xEEEE, 0xFEEE,
		0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x7FFF, 0xDFBF, 0xF7EF,
		0xFDFB, 0x7EFC, 0xDFBF, 0xF7EF, 0xFDFB, 0x007E, 0x0000
	};
	//static const char test_packet_buffer[] = {
	//	/* JKJKJKJK * 9 */
	//	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	//	/* JJKKJJKK * 8 */
	//	0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA,
	//	/* JJKKJJKK * 8 */
	//	0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE,
	//	/* JJJJJJJKKKKKKK * 8 */
	//	0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	//	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	//	/* JJJJJJJK * 8 */
	//	0x7F, 0xBF, 0xDF, 0xEF, 0xF7, 0xFB, 0xFD,
	//	/* {JKKKKKKK * 10}, JK */
	//	0xFC, 0x7E, 0xBF, 0xDF, 0xEF, 0xF7, 0xFB, 0xFD, 0x7E
	//};

	udelay(100);	// 100us

	switch ( (wIndex>>8) & 0xFF )
	{
		case 0x01 :				// Test J
			HwUTST |= 0x10;
			HwUTST |= 0x02;
			break;

		case 0x02 : 			// Test K
			HwUTST |= 0x10;
			HwUTST |= 0x04;
			break;

		case 0x03 : 			// Test SE0 NAK
			HwUTST |= 0x10;
			HwUTST |= 0x01;
			break;

		case 0x04 : 			// Test Packet
			HwUTST |= 0x10;

			HwIR = 0;
			HwBWCR = 53;

			for ( i = 0; i < 27 ; ++i )
				HwEP0BUF = (unsigned long)TEST_Packet[i];

			HwUTST |= 0x08;
			break;
	}
}

static void usbt_flush_ep(struct usbt_ep_t *ep)
{
	u32 i,t,nEndPoint, dwIRSave;

	dwIRSave	= HwIR;
	nEndPoint	= ep->index;

	if( nEndPoint == 0 )
		return;

	HwIR		= nEndPoint;

	//_loc_SyncDelay();

	//dwIsDualFIFO = 1;

	//if ( nEndPoint == TCC7XX_USBDRV_BULK_OUT_EP )
	//{
	//	#ifdef USE_BULK_OUT_DUAL_FIFO
	//	dwIsDualFIFO = 2;
	//	#endif
	//
	//	#ifdef USE_QUEUE_BUFFER
	//	gUSB_QUEUE_OUT_FullFlag		= 0;
	//	gUSB_QUEUE_OUT_HeadPacket	= 0;
	//	gUSB_QUEUE_OUT_TailPacket	= 0;
	//	#endif
	//}
	//else if ( nEndPoint == TCC7XX_USBDRV_BULK_IN_EP )
	//{
	//	#ifdef USE_BULK_IN_DUAL_FIFO
	//	dwIsDualFIFO = 2;
	//	#endif
	//}

	for ( i = 0; i < 1; ++i )
	{
		if ( (HwESR & (HwESR_PSIF_1P|HwESR_PSIF_2P) ) == 0 )
			break;

		HwECR		|= HwECR_FLUSH_EN;

		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };	

		for ( t = 0; t < 0xFFFF; ++t )
		{
			if ( HwESR & HwESR_FFS )
			{
				HwESR = HwESR_FFS;
				break;
			}
		}

		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
		//__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };__asm { NOP };
	}

	//_loc_SyncDelay();

	HwIR = dwIRSave;
}

static u32 usbt_raw_read_ep0( void )
{
	u32	nLen, BRCR, IRSave;
	u16	*pTmp;

	IRSave	= HwIR;
	HwIR		= 0;

	BRCR	= HwBRCR;
	nLen	= BRCR<<1;		// Get the size of EP0 FIFO IN
	if( HwEP0SR & HwEP0SR_LWO )
		nLen--;
	s_usbt_ep0_length = nLen;

	pTmp	= (unsigned short *)s_usbt_ep0_buffer;
	while(BRCR--)
		*pTmp++	= HwEP0BUF;

	HwEP0SR	= HwEP0SR_RSR;

	HwIR	= IRSave;

	return nLen;
}

static int usbt_wait_read_ep0(void)
{
	volatile u32 i;
	for(i=0;i<0xffff;i++)
	{
		if(s_usbt_ep0_length!=0 && s_usbt_ep0_length!=8)
			return 1;
	}
	return 0;
}

static void usbt_read_ep0( struct usbt_ep_t *ep )
{
	struct usbt_udc_t *udc = ep->udc;
	struct usbt_request_t *req;

	if (list_empty(&ep->queue)) {
		//DBG(DBG_ERROR,"usbt_write_ep list empty error!");
		printk("usbt_read_ep list empty error\n");
		return;
	}

	req = list_entry(ep->queue.next, struct usbt_request_t, queue);

	while(req->req.length - req->req.actual)
	{
		if(usbt_wait_read_ep0())
		{
			if( req->req.length - req->req.actual < s_usbt_ep0_length )
				s_usbt_ep0_length = req->req.length - req->req.actual;
			memcpy(req->req.buf + req->req.actual, s_usbt_ep0_buffer, s_usbt_ep0_length);
			req->req.actual += s_usbt_ep0_length;
			s_usbt_ep0_length = 0;
		}
		else
			break;
	}

	//struct usbt_udc_t *udc = ep->udc;
	//DBG(DBG_DEBUG,"write complete");
	list_del_init(&req->queue);
	if( req->req.length - req->req.actual == 0 )
		req->req.status = 0;
	else
		req->req.status = -ETIMEDOUT;
	spin_unlock(&udc->lock);
	req->req.complete(&ep->ep, &req->req);
	//printk("read ep0 complete\n");
	spin_lock(&udc->lock);
}

static void usbt_raw_write_ep0( void *pData, u32 length )
{
	u32 IRSave, i, packetLength;
	u16 *pBuffer = pData;

	IRSave = HwIR;
	HwIR = 0;

	//DBG(DBG_DEBUG,"write ep0 length=%d",length);

	while( length > 0 )
	{
		if( length > 64 )
			packetLength = 64;
		else
			packetLength = length;

		HwBWCR = packetLength;
		for(i=0;i<packetLength;i+=2)
			HwEP0BUF = *pBuffer++;

		length -= packetLength;
		if(length > 0 )
		{
			while( !(HwEP0SR & HwEP0SR_TST) );
		}
	}

	HwIR = IRSave;
}

static void usbt_write_ep0(struct usbt_ep_t *ep)
{
	struct usbt_udc_t *udc = ep->udc;
	struct usbt_request_t *req;

	if (list_empty(&ep->queue)) {
		//DBG(DBG_ERROR,"usbt_write_ep list empty error!");
		printk("usbt_write_ep list empty error\n");
		return;
	}

	req = list_entry(ep->queue.next, struct usbt_request_t, queue);

	usbt_raw_write_ep0(req->req.buf,req->req.length);
	req->req.actual += req->req.length;

	if( req->req.length - req->req.actual == 0 )
	{
		//struct usbt_udc_t *udc = ep->udc;
		//DBG(DBG_DEBUG,"write complete");
		list_del_init(&req->queue);
		req->req.status = 0;
		spin_unlock(&udc->lock);
		req->req.complete(&ep->ep, &req->req);
		//printk("write ep0 complete\n");
		spin_lock(&udc->lock);
	}
}
static void usbt_read_ep(struct usbt_ep_t *ep)
{
	struct usbt_udc_t *udc = ep->udc;
	//unsigned long status;
	unsigned int bytecount;
	//unsigned int nr_busy;
	u32 BRCR,IRSave = HwIR;
	#if !defined(USE_DMA_IRQ)
	struct usbt_request_t *req;
	int is_complete = 0;
	u16 *pBuffer;
	#endif

	HwIR = ep->index;

	while( HwESR & HwESR_RPS )
	{
		if (list_empty(&ep->queue)) {
			break;
		}

		BRCR = HwBRCR;
		bytecount = BRCR << 1;
		if( HwESR & HwESR_LWO )
			bytecount--;

		//DBG(DBG_DEBUG, "read (%d)", bytecount);

		#if !defined(USE_DMA_IRQ)
		req = list_entry(ep->queue.next, struct usbt_request_t, queue);

		if( bytecount < ( le16_to_cpu(ep->desc->wMaxPacketSize) & 0x7ff ) )
			is_complete = 1;
		if (req->req.actual + bytecount >= req->req.length) {
			is_complete = 1;
			//bytecount = req->req.length - req->req.actual;
		}
		pBuffer = req->req.buf + req->req.actual;
		#endif

		#if defined(USE_MCU_ACCESS)
		{
			u32 i;
			u16 *EPBUF = usbt_get_epbuf( ep->index );
			for( i = 0 ; i < BRCR ; i++ )
				*pBuffer++ = *EPBUF;
		}
		#elif defined(USE_DMA_ACCESS)
		{
			#if defined(CONFIG_ARCH_TCC83X)
			sHwDMA *pHwDMA = (sHwDMA*)&HwST_SADR_USB;
			#elif defined(CONFIG_ARCH_TCC79X)
			sHwDMA *pHwDMA	= (sHwDMA*)&HwGDMA_CH_BASE(USB_DMA_CHANNEL);
			#endif
			unsigned uCHCTRL =
								HwCHCTRL_BST_ARB		|
								HwCHCTRL_TYPE_SW		|
								HwCHCTRL_WSIZE_16		|
								HwCHCTRL_FLAG			|
								#if defined(USE_DMA_IRQ)
								HwCHCTRL_IEN_EN			|
								#endif
								HwCHCTRL_EN_EN;

			// Set Source Address & Source Parameter (mask + increment)
			pHwDMA->ST_SADR	= (unsigned)usbt_get_epbuf_physical( ep->index );
			pHwDMA->SPARAM = 0x00;

			if( (bytecount&0xF) == 0 )
			{
				uCHCTRL |= HwCHCTRL_BSIZE_8;
				pHwDMA->HCOUNT = bytecount>>4;
			}
			else
				pHwDMA->HCOUNT = (bytecount+1)>>1;

			// Set Dest Address & Dest Parameter
			pHwDMA->ST_DADR = (unsigned)udc->dtcm_pa;
			pHwDMA->DPARAM = 0x02;

			// Set Channel Control
			pHwDMA->CHCTRL		= uCHCTRL;

			#if defined(USE_DMA_IRQ)
			udc->dmaisr_ep = ep;
			udc->dmaisr_length = bytecount;
			ep->is_working = 1;

			HwEIER &= ~(1 << ep->index);
			break;
			#else
			while( !(pHwDMA->CHCTRL & HwCHCTRL_FLAG) );

			memcpy( (void*)pBuffer, (void*)udc->dtcm, bytecount );
			#endif
		}
		#endif

		#if !defined(USE_DMA_IRQ)
		req->req.actual += bytecount;

		if (is_complete)
		{
			//DBG(DBG_DEBUG, "read complete");
			req->req.status = 0;
			list_del_init(&req->queue);
			spin_unlock(&udc->lock);
			req->req.complete(&ep->ep, &req->req);
			spin_lock(&udc->lock);
			break;
		}
		#endif
	}

	HwIR = IRSave;
}

static void usbt_request_complete(struct usbt_ep_t *ep, struct usbt_request_t *req, int status)
{
	struct usbt_udc_t *udc = ep->udc;

	WARN_ON(!list_empty(&req->queue));

	if (req->req.status == -EINPROGRESS)
		req->req.status = status;

	//if (req->mapped) {
	//	dma_unmap_single(
	//		&udc->pdev->dev, req->req.dma, req->req.length,
	//		ep->is_in ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
	//	req->req.dma = DMA_ADDR_INVALID;
	//	req->mapped = 0;
	//}

	//DBG(DBG_GADGET | DBG_REQ,
	//	"%s: req %p complete: status %d, actual %u\n",
	//	ep->ep.name, req, req->req.status, req->req.actual);

	spin_unlock(&udc->lock);
	req->req.complete(&ep->ep, &req->req);
	spin_lock(&udc->lock);
}

static void usbt_request_complete_list(struct usbt_ep_t *ep, struct list_head *list, int status)
{
	struct usbt_request_t *req, *tmp_req;

	list_for_each_entry_safe(req, tmp_req, list, queue) {
		list_del_init(&req->queue);
		usbt_request_complete(ep, req, status);
	}
}

static int usbt_wait_bulkin_ready( void )
{
	u32 t,k,ESR,dwConfirmCount;
	int cRet;

	cRet = 0;
	dwConfirmCount = 0;

	for ( t = 0; t < 0x4FFFF && !cRet; t++ )
	{
		for ( k = 0; k < 100 && !cRet; k++ )
		{
			ESR = (HwESR&(Hw3+Hw2));

			if ( ESR == 0x00 )							// No Packet in FIFO
			{
				//cRet = 1;
				++dwConfirmCount;
			}
			else
			{
				dwConfirmCount = 0;
			}

			if ( dwConfirmCount > 20 )
			{
				cRet = 1;
			}
		}	
	}

	return cRet;
}

static void usbt_write_ep(struct usbt_ep_t *ep)
{
	struct usbt_udc_t *udc = ep->udc;
	struct usbt_request_t *req;
	u32 IRSave = HwIR;
	u16 *EPBUF, *pBuffer;
	u32 maxpacket = (u32)le16_to_cpu(ep->desc->wMaxPacketSize) & 0x7ff;
	u32 packetsize;

	if (list_empty(&ep->queue)) {
		//DBG(DBG_ERROR,"usbt_write_ep list empty error!");
		return;
	}

	HwIR = ep->index;

	#ifdef EP_WRITE_INT
	if( !( HwESR & HwESR_PSIF_1P ) )
	#else
	while(1)
	#endif
	{
		req = list_entry(ep->queue.next, struct usbt_request_t, queue);

		if( req->req.length - req->req.actual > maxpacket )
			packetsize = maxpacket;
		else
			packetsize = req->req.length - req->req.actual;
		pBuffer = req->req.buf + req->req.actual;

		if( packetsize )
		{
			#ifndef EP_WRITE_INT
			if(usbt_wait_bulkin_ready()==0)
			{
				req->req.status = -ETIMEDOUT;
				break;
			}
			#endif

			//DBG(DBG_DEBUG,"write (%d)",packetsize);
			HwBWCR = packetsize;
			#if defined(USE_MCU_ACCESS)
			{
				u32 i;
				EPBUF = usbt_get_epbuf( ep->index );
				for( i = 0 ; i < packetsize ; i += 2 )
					*EPBUF = *pBuffer++;
			}
			#elif defined(USE_DMA_ACCESS)
			{
				#if defined(CONFIG_ARCH_TCC83X)
				sHwDMA *pHwDMA = (sHwDMA*)&HwST_SADR_USB;
				#elif defined(CONFIG_ARCH_TCC79X)
				sHwDMA *pHwDMA	= (sHwDMA*)&HwGDMA_CH_BASE(USB_DMA_CHANNEL);
				#endif
				unsigned uCHCTRL =
									HwCHCTRL_BST_ARB		|
									HwCHCTRL_TYPE_SW		|
									HwCHCTRL_WSIZE_16		|
									HwCHCTRL_FLAG			|
									HwCHCTRL_EN_EN;
				EPBUF = usbt_get_epbuf_physical( ep->index );

				memcpy( (void*)udc->dtcm, (void*)pBuffer, packetsize );

				// Set Dest Address & Dest Parameter
				pHwDMA->ST_DADR = (unsigned)EPBUF;
				pHwDMA->DPARAM = 0x00;

				if( (packetsize&0xF) == 0 )
				{
					uCHCTRL |= HwCHCTRL_BSIZE_8;
					pHwDMA->HCOUNT = packetsize>>4;
				}
				else
					pHwDMA->HCOUNT = (packetsize+1)>>1;

				// Set Source Address & Source Parameter (mask + increment)
				pHwDMA->ST_SADR	= (unsigned)udc->dtcm_pa;
				pHwDMA->SPARAM = 0x02;

				// Set Channel Control
				pHwDMA->CHCTRL		= uCHCTRL;
			}
			#endif
			req->req.actual += packetsize;

			#ifdef EP_WRITE_INT
			ep->is_working = 1;
			#endif
		}

		if( req->req.length - req->req.actual == 0 )
		{
			//struct usbt_udc_t *udc = ep->udc;
			//DBG(DBG_DEBUG,"write complete");
			list_del_init(&req->queue);
			req->req.status = 0;
			spin_unlock(&udc->lock);
			req->req.complete(&ep->ep, &req->req);
			spin_lock(&udc->lock);

			#ifndef EP_WRITE_INT
			break;
			#endif
		}
	}

	HwIR = IRSave;
}

static int usbt_get_frame_number(struct usb_gadget *gadget)
{
	if( (HwFNR&HwFNR_FTL) != 0 )		// frame timer is locked?
	{
		if( (HwFNR&HwFNR_SM) == 0 )		// frame timer locking is failed?
		{
			return (HwFNR & HwFNR_FN_MASK);
		}
	}

	return 0;
}

/*-------------------------------------------------------------------------*/

static void nop_release(struct device *dev)
{
}

static const struct usb_gadget_ops usbt_udc_ops = {
	.get_frame		= usbt_get_frame_number,
	.wakeup			= NULL,					// N/A
	.set_selfpowered	= NULL,				// N/A
};

static int usbt_ep_enable(struct usb_ep *_ep, const struct usb_endpoint_descriptor *desc)
{
	struct usbt_ep_t *ep = to_usbt_ep(_ep);
	struct usbt_udc_t *udc = ep->udc;
	unsigned long flags, maxpacket;
	//unsigned long ept_cfg;
	//unsigned int nr_trans;
	u32 HwIRSave;
	int retVal= 0;

	DBG(DBG_GADGET, "%s: ep_enable: desc=%p", ep->ep.name, desc);

	maxpacket = le16_to_cpu(desc->wMaxPacketSize) & 0x7ff;

	if ( ((desc->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK) != ep->index)
			|| ep->index == 0
			|| desc->bDescriptorType != USB_DT_ENDPOINT
			|| maxpacket == 0
			|| maxpacket > ep->fifo_size
			|| ep->desc != NULL ) {
		DBG(DBG_ERR, "ep_enable: Invalid argument");
		return -EINVAL;
	}

	//ep->is_isoc = 0;

	spin_lock_irqsave(&ep->udc->lock, flags);

	HwIRSave = HwIR;
	HwIR = ep->index;

	HwMPR = maxpacket;

	DBG(DBG_HW, "%s: EPT_SIZE = maxpacket = %lu",
			ep->ep.name, maxpacket);

	if ((desc->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
		ep->is_in = 1;
		//ept_cfg |= USBA_EPT_DIR_IN;
		HwEDR |= ( 1 << ep->index );
	}
	else
	{
		ep->is_in = 0;
		HwEDR &= ~( 1 << ep->index );
	}
	ep->is_working = 0;

	switch (desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) {
	case USB_ENDPOINT_XFER_CONTROL:
		{
			DBG(DBG_ERR, "ep_enable: %s is not control capable",
					ep->ep.name);
			retVal = -EINVAL;
		}
		//ept_cfg |= USBA_BF(EPT_TYPE, USBA_EPT_TYPE_CONTROL);
		//ept_cfg |= USBA_BF(BK_NUMBER, USBA_BK_NUMBER_ONE);
		break;
	case USB_ENDPOINT_XFER_ISOC:
		//if (!ep->can_isoc)
		{
			DBG(DBG_ERR, "ep_enable: %s is not isoc capable",
					ep->ep.name);
			retVal = -EINVAL;
		}

		/*
		 * Bits 11:12 specify number of _additional_
		 * transactions per microframe.
		 */
		//nr_trans = ((le16_to_cpu(desc->wMaxPacketSize) >> 11) & 3) + 1;
		//if (nr_trans > 3)
		//	retVal = -EINVAL;
		//
		//ep->is_isoc = 1;
		//ept_cfg |= USBA_BF(EPT_TYPE, USBA_EPT_TYPE_ISO);
		//
		///*
		// * Do triple-buffering on high-bandwidth iso endpoints.
		// */
		//if (nr_trans > 1 && ep->nr_banks == 3)
		//	ept_cfg |= USBA_BF(BK_NUMBER, USBA_BK_NUMBER_TRIPLE);
		//else
		//	ept_cfg |= USBA_BF(BK_NUMBER, USBA_BK_NUMBER_DOUBLE);
		//ept_cfg |= USBA_BF(NB_TRANS, nr_trans);
		break;
	case USB_ENDPOINT_XFER_BULK:
		//ept_cfg |= USBA_BF(EPT_TYPE, USBA_EPT_TYPE_BULK);
		//ept_cfg |= USBA_BF(BK_NUMBER, USBA_BK_NUMBER_DOUBLE);
		if( ep->is_in )
			HwECR = HwECR_CDP_CLR;
		else
			HwECR = HwECR_DUEN_EN | HwECR_CDP_CLR;
		HwDCR = 0x00;
		break;
	case USB_ENDPOINT_XFER_INT:
		//ept_cfg |= USBA_BF(EPT_TYPE, USBA_EPT_TYPE_INT);
		//ept_cfg |= USBA_BF(BK_NUMBER, USBA_BK_NUMBER_DOUBLE);
		HwECR = HwECR_CDP_CLR;
		HwDCR = 0x00;
		break;
	}

	if( retVal != 0 )
	{
		HwIR = HwIRSave;
		spin_unlock_irqrestore(&udc->lock, flags);
		return retVal;
	}

	ep->desc = desc;
	ep->ep.maxpacket = maxpacket;

	//usba_ep_writel(ep, CFG, ept_cfg);
	//usba_ep_writel(ep, CTL_ENB, USBA_EPT_ENABLE);
	HwEIR = 1 << ep->index;
	if( ep->is_in )
		HwEIER |= 1 << ep->index;

	//if (ep->can_dma) {
	//	u32 ctrl;
	//
	//	usba_writel(udc, INT_ENB,
	//			(usba_readl(udc, INT_ENB)
	//				| USBA_BF(EPT_INT, 1 << ep->index)
	//				| USBA_BF(DMA_INT, 1 << ep->index)));
	//	ctrl = USBA_AUTO_VALID | USBA_INTDIS_DMA;
	//	usba_ep_writel(ep, CTL_ENB, ctrl);
	//} else {
	//	usba_writel(udc, INT_ENB,
	//			(usba_readl(udc, INT_ENB)
	//				| USBA_BF(EPT_INT, 1 << ep->index)));
	//}

	spin_unlock_irqrestore(&udc->lock, flags);

	//DBG(DBG_HW, "EPT_CFG%d after init: %#08lx\n", ep->index,
	//		(unsigned long)usba_ep_readl(ep, CFG));
	//DBG(DBG_HW, "INT_ENB after init: %#08lx\n",
	//		(unsigned long)usba_readl(udc, INT_ENB));

	return 0;
}

static int usbt_ep_disable(struct usb_ep *_ep)
{
	struct usbt_ep_t *ep = to_usbt_ep(_ep);
	struct usbt_udc_t *udc = ep->udc;
	LIST_HEAD(req_list);
	unsigned long flags;

	DBG(DBG_DEBUG, "ep_disable: %s\n", ep->ep.name);

	spin_lock_irqsave(&udc->lock, flags);

	if (!ep->desc) {
		spin_unlock_irqrestore(&udc->lock, flags);
		DBG(DBG_ERROR, "ep_disable: %s not enabled\n", ep->ep.name);
		return -EINVAL;
	}
	ep->desc = NULL;

	list_splice_init(&ep->queue, &req_list);
	//if (ep->can_dma) {
	//	usba_dma_writel(ep, CONTROL, 0);
	//	usba_dma_writel(ep, ADDRESS, 0);
	//	usba_dma_readl(ep, STATUS);
	//}
	//usba_ep_writel(ep, CTL_DIS, USBA_EPT_ENABLE);
	//usba_writel(udc, INT_ENB,
	//		usba_readl(udc, INT_ENB)
	//		& ~USBA_BF(EPT_INT, 1 << ep->index));

	usbt_request_complete_list(ep, &req_list, -ESHUTDOWN);

	HwEIER &= ~( 1 << ep->index );

	spin_unlock_irqrestore(&udc->lock, flags);

	return 0;
}

static struct usb_request *usbt_ep_alloc_request(struct usb_ep *_ep, gfp_t gfp_flags)
{
	struct usbt_request_t *req;

	DBG(DBG_GADGET, "ep_alloc_request: %p, 0x%x\n", _ep, gfp_flags);

	req = kzalloc(sizeof(*req), gfp_flags);
	if (!req)
		return NULL;

	INIT_LIST_HEAD(&req->queue);
	//req->req.dma = DMA_ADDR_INVALID;

	return &req->req;
}

static void usbt_ep_free_request(struct usb_ep *_ep, struct usb_request *_req)
{
	struct usbt_request_t *req = to_usbt_req(_req);

	DBG(DBG_GADGET, "ep_free_request: %p, %p\n", _ep, _req);

	kfree(req);
}

static int usbt_ep_queue(struct usb_ep *_ep, struct usb_request *_req, gfp_t gfp_flags)
{
	struct usbt_request_t *req = to_usbt_req(_req);
	struct usbt_ep_t *ep = to_usbt_ep(_ep);
	struct usbt_udc_t *udc = ep->udc;
	unsigned long flags;
	int ret;

	//DBG(DBG_DEBUG, "usbt ep queue(%s) size=%d",_ep->name,_req->length);

	if( (u32)_req->buf & 1 )
		DBG(DBG_ERROR,"ep%d buffer is wrongly aligned", ep->index);

	//DBG(DBG_GADGET | DBG_QUEUE | DBG_REQ, "%s: queue req %p, len %u\n",
	//		ep->ep.name, req, _req->length);

	if (!udc->driver || udc->gadget.speed == USB_SPEED_UNKNOWN || !ep->desc)
	{
		DBG(DBG_ERROR, "usbt ep queue driver=%p speed=%d desc=%p", udc->driver, udc->gadget.speed, ep->desc);
		return -ESHUTDOWN;
	}

	//req->submitted = 0;
	//req->using_dma = 0;
	//req->last_transaction = 0;

	_req->status = -EINPROGRESS;
	_req->actual = 0;

	//if (ep->can_dma)
	//	return queue_dma(udc, ep, req, gfp_flags);

	/* May have received a reset since last time we checked */
	ret = -ESHUTDOWN;
	spin_lock_irqsave(&udc->lock, flags);
	if (ep->desc) {
		list_add_tail(&req->queue, &ep->queue);

		//if (ep->is_in || (ep_is_control(ep)
		//		&& (ep->state == DATA_STAGE_IN
		//			|| ep->state == STATUS_STAGE_IN)))
		//	usba_ep_writel(ep, CTL_ENB, USBA_TX_PK_RDY);
		//else
		//	usba_ep_writel(ep, CTL_ENB, USBA_RX_BK_RDY);

		if( ep->is_in )
		{
			if( ep->index == 0 )
			{
				//printk("write ep0 queue ");
				usbt_write_ep0(ep);
			}
			else
			{
				//DBG(DBG_DEBUG, "IN queue req len %d", _req->length);
				if( ep->is_working == 0 )
				{
					//DBG(DBG_DEBUG, "IN queue now");
					usbt_write_ep(ep);
				}
			}
		}
		else
		{
			if( ep->index == 0 )
			{
				//printk("read ep0 queue ");
				usbt_read_ep0(ep);
			}
			if( ep->index && !ep->is_working )
				HwEIER |= 1 << ep->index;
		}

		ret = 0;
	}
	spin_unlock_irqrestore(&udc->lock, flags);

	return ret;
}

static int usbt_ep_dequeue(struct usb_ep *_ep, struct usb_request *_req)
{
	struct usbt_ep_t *ep = to_usbt_ep(_ep);
	struct usbt_udc_t *udc = ep->udc;
	struct usbt_request_t *req = to_usbt_req(_req);
	unsigned long flags;

	DBG(DBG_DEBUG,"usbt ep dequeue %s", ep->ep.name);

	spin_lock_irqsave(&udc->lock, flags);

	/*
	 * Errors should stop the queue from advancing until the
	 * completion function returns.
	 */
	list_del_init(&req->queue);

	usbt_request_complete(ep, req, -ECONNRESET);

	/* Process the next request if any */
	//submit_next_request(ep);
	spin_unlock_irqrestore(&udc->lock, flags);

	return 0;
}

static int usbt_ep_set_halt(struct usb_ep *_ep, int value)
{
	struct usbt_ep_t *ep = to_usbt_ep(_ep);
	struct usbt_udc_t *udc = ep->udc;
	unsigned long flags;
	int ret = 0;

	if (!ep->desc) {
		DBG(DBG_ERR, "Attempted to halt uninitialized ep %s\n",
				ep->ep.name);
		return -ENODEV;
	}
	//if (ep->is_isoc) {
	//	DBG(DBG_ERR, "Attempted to halt isochronous ep %s\n",
	//			ep->ep.name);
	//	return -ENOTTY;
	//}

	spin_lock_irqsave(&udc->lock, flags);

	/*
	 * We can't halt IN endpoints while there are still data to be
	 * transferred
	 */
	if (!list_empty(&ep->queue) )
			//|| ((value && ep->is_in && (usba_ep_readl(ep, STA)
			//		& USBA_BF(BUSY_BANKS, -1L)))))
	{
		ret = -EAGAIN;
	}
	else
	{
		if (value)
		{
			//usba_ep_writel(ep, SET_STA, USBA_FORCE_STALL);
			if( ep->is_in )
			{
				u32 IRSave = HwIR;
				HwIR = ep->index;
				usbt_wait_bulkin_ready();
				HwIR = IRSave;
			}
			usbt_stall(ep);
		}
		else
		{
			//usba_ep_writel(ep, CLR_STA, USBA_FORCE_STALL | USBA_TOGGLE_CLR);
			usbt_unhalt(ep);
		}
		//usba_ep_readl(ep, STA);
		//DBG(DBG_DEBUG,"usbt ep set halt value=%d", value);
	}

	spin_unlock_irqrestore(&udc->lock, flags);

	return ret;

	//struct usbt_ep_t *ep = to_usbt_ep(_ep);
	//DBG(DBG_DEBUG,"usbt ep set halt value=%d", value);
	//usbt_stall(ep);
	//return 0;
}

static int usbt_ep_fifo_status(struct usb_ep *_ep)
{
	struct usbt_ep_t *ep = to_usbt_ep(_ep);
	u32 IRSave = HwIR;

	DBG(DBG_DEBUG,"usbt ep fifo status");

	HwIR = ep->index;
	if( ep->is_in )
		return (int)HwBWCR;
	else
		return (int)(HwBRCR<<1);
	HwIR = IRSave;
}

static void usbt_ep_fifo_flush(struct usb_ep *_ep)
{
	struct usbt_ep_t *ep = to_usbt_ep(_ep);

	DBG(DBG_DEBUG,"usbt ep fifo flush");

	usbt_flush_ep(ep);
}

static const struct usb_ep_ops usbt_ep_ops = {
	.enable		= usbt_ep_enable,
	.disable	= usbt_ep_disable,
	.alloc_request	= usbt_ep_alloc_request,
	.free_request	= usbt_ep_free_request,
	.queue		= usbt_ep_queue,
	.dequeue	= usbt_ep_dequeue,
	.set_halt	= usbt_ep_set_halt,
	.fifo_status	= usbt_ep_fifo_status,
	.fifo_flush	= usbt_ep_fifo_flush,
};

static struct usbt_udc_t the_udc;

#define EP(nam, idx, maxpkt)			\
{										\
	.ep	= {								\
		.ops		= &usbt_ep_ops,	\
		.name		= nam,				\
		.maxpacket	= maxpkt,			\
	},									\
	.udc			= &the_udc,		\
	.queue		= LIST_HEAD_INIT(usbt_ep[idx].queue),	\
	.fifo_size		= maxpkt,			\
	.index			= idx,				\
	.is_working		= 0,				\
}

static struct usbt_ep_t usbt_ep[] = {
	EP("ep0",			0,	64),
	EP("ep1in-bulk",	1,	512),
	EP("ep2out-bulk",	2,	512),
	EP("ep3in-int",		3,	64),
};
#undef EP

static struct usb_endpoint_descriptor usbt_ep0_desc = {
	.bLength = USB_DT_ENDPOINT_SIZE,
	.bDescriptorType = USB_DT_ENDPOINT,
	.bEndpointAddress = 0,
	.bmAttributes = USB_ENDPOINT_XFER_CONTROL,
	.wMaxPacketSize = __constant_cpu_to_le16(64),
	.bInterval = 0,
};

static struct usbt_udc_t the_udc = {
	.gadget	= {
		.ops			= &usbt_udc_ops,
		.ep0			= &usbt_ep[0].ep,
		.ep_list		= LIST_HEAD_INIT(the_udc.gadget.ep_list),
		.is_dualspeed	= 1,
		.name			= driver_name,
		.dev = {
			.bus_id		= "gadget",
			.release	= nop_release,
		},
	},

	.lock	= SPIN_LOCK_UNLOCKED,
};

static void usbt_reset(void)
{
	//--------------------------------------------------
	// Setting Control Registers
	//--------------------------------------------------

	/* Setting USB Device Type */
	HwUTST = 0x0000; 	// SET DEVICE AS USB 2.0 HIGH SPEED

	HwUPCFG0	= 0x0842;

	#if defined(CONFIG_ARCH_TCC83X)
	HwUPCFG1	=
					HwUPCFG1_TXFSLST(3)	// FS/LS Pull-Up Resistance Adjustment = Design default
					| HwUPCFG1_SQRXT(6)	// Squelch Threshold Tune = -5%
					| HwUPCFG1_OTGT(3)	// VBUS Valid Threshold Adjustment = Design default
					| HwUPCFG1_CDT(4);	// Disconnect Threshold Adjustment = Design default
	HwUPCFG2	=
					HwUPCFG2_TM_HS
					| HwUPCFG2_XCVRSEL_HS
					| HwUPCFG2_OPMODE_NORMAL
					//| HwUPCFG2_TXVRT(0xC)	// HS DC voltage level adjustment = 9%
					| HwUPCFG2_TXVRT(0xA)	// HS DC voltage level adjustment = 6%
					//| HwUPCFG2_TXVRT(5)	// HS DC voltage level adjustment = Design default
					| HwUPCFG2_TXRT(0)	// HS Transmitter Rise/Fall Time Adjustment = Design default
					| HwUPCFG2_TP_DIS;
	#elif defined(CONFIG_ARCH_TCC79X)
	HwUPCFG1	=
					HwUPCFG1_TXFSLST(7)	// FS/LS Pull-Up Resistance Adjustment = Design default
					| HwUPCFG1_SQRXT(3)	// Squelch Threshold Tune = -5%
					| HwUPCFG1_OTGT(4)	// VBUS Valid Threshold Adjustment = Design default
					| HwUPCFG1_CDT(3);	// Disconnect Threshold Adjustment = Design default
	HwUPCFG2	=
					HwUPCFG2_TM_HS
					| HwUPCFG2_XCVRSEL_HS
					| HwUPCFG2_OPMODE_NORMAL
					| HwUPCFG2_TXVRT(8)	// HS DC voltage level adjustment = Design default
					| HwUPCFG2_TXRT(0)	// HS Transmitter Rise/Fall Time Adjustment = Design default
					| HwUPCFG2_TP_DIS;
	#endif
	HwUPCFG3	= 0x9000;	// OTG disable
	HwPLICR		= 0x0000;

	HwSCR		=	0xA000 				|
					//HwSCR_DTZIEN_EN		|	// DMA Total Conter Zero Interrupt Enable
					HwSCR_DIEN_EN		|	// DUAL Interrupt Enable
					//HwSCR_EIE_EN		|	// Error Interrupt Enable
					HwSCR_SDE_EN		|	// Speed Detect End Interrupt Enable
					HwSCR_RRDE_EN		|	// Reverse Read Data Enable
					//HwSCR_HSUSPE_EN		|	// Suspend Enable
					HwSCR_HRESE_EN;			// Reset Enable

	HwFAR		= 0x00000000;

	//--------------------------------------------------
	// Clear Endpoint
	//--------------------------------------------------
	HwEIR		= HwEIR_EP0INT + HwEIR_EP1INT + HwEIR_EP2INT + HwEIR_EP3INT;		// Clear Endpoint Interrupt.
	HwEIER		= HwEIER_EP0INT_EN;

	//--------------------------------------------------
	// Initialize relevant variables
	//--------------------------------------------------
	usbt_ep[0].desc = &usbt_ep0_desc;
}

static void usbt_detach(void)
{
	/* forced disconnect USB from host */
	HwSWRESET |= HwSWRESET_UB_ON;
//	udelay(5);								// SE0 for >=2.5us
	HwSWRESET &= ~HwSWRESET_UB_ON;
}

static void usbt_reset_all_endpoints(struct usbt_udc_t *udc)
{
	struct usbt_ep_t *ep;
	struct usbt_request_t *req, *tmp_req;

	//usba_writel(udc, EPT_RST, ~0UL);
	//printk("usbt_reset_all_endpoints\n");

	ep = to_usbt_ep(udc->gadget.ep0);
	list_for_each_entry_safe(req, tmp_req, &ep->queue, queue) {
		list_del_init(&req->queue);
		//printk("usbt_request_complete\n");
		usbt_request_complete(ep, req, -ECONNRESET);
	}

	list_for_each_entry(ep, &udc->gadget.ep_list, ep.ep_list) {
		if (ep->desc) {
			spin_unlock(&udc->lock);
			//printk("usbt_ep_disable %d\n",ep->index);
			usbt_ep_disable(&ep->ep);
			spin_lock(&udc->lock);
		}
	}
}

static struct usbt_ep_t *usbt_get_ep_by_addr(struct usbt_udc_t *udc, u16 wIndex)
{
	struct usbt_ep_t *ep;

	if ((wIndex & USB_ENDPOINT_NUMBER_MASK) == 0)
		return to_usbt_ep(udc->gadget.ep0);

	list_for_each_entry (ep, &udc->gadget.ep_list, ep.ep_list) {
		u8 bEndpointAddress;

		if (!ep->desc)
			continue;
		bEndpointAddress = ep->desc->bEndpointAddress;
		//if ((wIndex ^ bEndpointAddress) & USB_DIR_IN)
		//	continue;
		if ((bEndpointAddress & USB_ENDPOINT_NUMBER_MASK)
				== (wIndex & USB_ENDPOINT_NUMBER_MASK))
			return ep;
	}

	return NULL;
}

/* Avoid overly long expressions */
static inline bool feature_is_dev_remote_wakeup(struct usb_ctrlrequest *crq)
{
	if (crq->wValue == __constant_cpu_to_le16(USB_DEVICE_REMOTE_WAKEUP))
		return true;
	return false;
}

static inline bool feature_is_dev_test_mode(struct usb_ctrlrequest *crq)
{
	if (crq->wValue == __constant_cpu_to_le16(USB_DEVICE_TEST_MODE))
		return true;
	return false;
}

static inline bool feature_is_ep_halt(struct usb_ctrlrequest *crq)
{
	if (crq->wValue == __constant_cpu_to_le16(USB_ENDPOINT_HALT))
		return true;
	return false;
}

static int usbt_handle_ep0_setup(struct usbt_udc_t *udc, struct usbt_ep_t *ep,
		struct usb_ctrlrequest *crq)
{
	int retval = 0;;

	switch (crq->bRequest) {
	case USB_REQ_GET_STATUS: {
		u16 status;

		if (crq->bRequestType == (USB_DIR_IN | USB_RECIP_DEVICE)) {
			status = cpu_to_le16(udc->devstatus);
		} else if (crq->bRequestType
				== (USB_DIR_IN | USB_RECIP_INTERFACE)) {
			status = __constant_cpu_to_le16(0);
		} else if (crq->bRequestType
				== (USB_DIR_IN | USB_RECIP_ENDPOINT)) {
			struct usbt_ep_t *target;

			target = usbt_get_ep_by_addr(udc, le16_to_cpu(crq->wIndex));
			if (!target)
				goto stall;

			status = 0;
			if (usbt_is_stalled(target))
				status |= __constant_cpu_to_le16(1);
		} else
			goto delegate;

		/* Write directly to the FIFO. No queueing is done. */
		if (crq->wLength != __constant_cpu_to_le16(sizeof(status)))
			goto stall;
		//ep->state = DATA_STAGE_IN;
		//__raw_writew(status, ep->fifo);
		usbt_raw_write_ep0(&status, 2);
		//usba_ep_writel(ep, SET_STA, USBA_TX_PK_RDY);
		break;
	}

	case USB_REQ_CLEAR_FEATURE: {
		//DBG(DBG_DEBUG,"USB_REQ_CLEAR_FEATURE");
		if (crq->bRequestType == USB_RECIP_DEVICE) {
			if (feature_is_dev_remote_wakeup(crq))
				udc->devstatus
					&= ~(1 << USB_DEVICE_REMOTE_WAKEUP);
			else
			//	/* Can't CLEAR_FEATURE TEST_MODE */
				goto stall;
		} else if (crq->bRequestType == USB_RECIP_ENDPOINT) {
			struct usbt_ep_t *target;

			if (crq->wLength != __constant_cpu_to_le16(0)
					|| !feature_is_ep_halt(crq))
			{
				goto stall;
			}
			target = usbt_get_ep_by_addr(udc, le16_to_cpu(crq->wIndex));
			if (!target)
			{
				goto stall;
			}

			//usba_ep_writel(target, CLR_STA, USBA_FORCE_STALL);
			//if (target->index != 0)
			//	usba_ep_writel(target, CLR_STA,
			//			USBA_TOGGLE_CLR);
			usbt_unhalt(target);
		} else {
			goto delegate;
		}

		//send_status(udc, ep);
		break;
	}

	case USB_REQ_SET_FEATURE: {
		if (crq->bRequestType == USB_RECIP_DEVICE) {
			if (feature_is_dev_test_mode(crq)) {
				//send_status(udc, ep);
				//ep->state = STATUS_STAGE_TEST;
				//udc->test_mode = le16_to_cpu(crq->wIndex);
				//return 0;
				usbt_testmode(le16_to_cpu(crq->wIndex));
			} else if (feature_is_dev_remote_wakeup(crq)) {
				udc->devstatus |= 1 << USB_DEVICE_REMOTE_WAKEUP;
			} else {
				goto stall;
			}
		} else if (crq->bRequestType == USB_RECIP_ENDPOINT) {
			struct usbt_ep_t *target;

			if (crq->wLength != __constant_cpu_to_le16(0)
					|| !feature_is_ep_halt(crq))
				goto stall;

			target = usbt_get_ep_by_addr(udc, le16_to_cpu(crq->wIndex));
			if (!target)
				goto stall;

			//usba_ep_writel(target, SET_STA, USBA_FORCE_STALL);
			usbt_stall(target);
		} else
			goto delegate;

		//send_status(udc, ep);
		break;
	}

	case USB_REQ_SET_ADDRESS:
		if (crq->bRequestType != (USB_DIR_OUT | USB_RECIP_DEVICE))
			goto delegate;

		//set_address(udc, le16_to_cpu(crq->wValue));
		//send_status(udc, ep);
		//ep->state = STATUS_STAGE_ADDR;
		break;

	default:
delegate:
		spin_unlock(&udc->lock);
		retval = udc->driver->setup(&udc->gadget, crq);
		spin_lock(&udc->lock);
	}

	return retval;

stall:
	printk(KERN_ERR
		"udc: %s: Invalid setup request: %02x.%02x v%04x i%04x l%d, "
		"halting endpoint...\n",
		ep->ep.name, crq->bRequestType, crq->bRequest,
		le16_to_cpu(crq->wValue), le16_to_cpu(crq->wIndex),
		le16_to_cpu(crq->wLength));
	//set_protocol_stall(udc, ep);
	usbt_stall(ep);
	return -1;
}

#if defined(USE_DMA_IRQ)
static irqreturn_t usbt_dma_irq(int irq, void *devid)
{
	struct usbt_udc_t *udc = devid;
	struct usbt_ep_t *ep = udc->dmaisr_ep;
	struct usbt_request_t *req;
	u32 length = udc->dmaisr_length;
	void *pBuffer;
	int is_complete=0;
	#if defined(CONFIG_ARCH_TCC83X)
	sHwDMA *pHwDMA = (sHwDMA*)&HwST_SADR_USB;
	#endif

	spin_lock(&udc->lock);

	pHwDMA->CHCTRL = HwCHCTRL_FLAG;

	req = list_entry(ep->queue.next, struct usbt_request_t, queue);

	if( length < ( le16_to_cpu(ep->desc->wMaxPacketSize) & 0x7ff ) )
		is_complete = 1;
	if (req->req.actual + length >= req->req.length) {
		is_complete = 1;
	}
	pBuffer = (void*)( req->req.buf + req->req.actual );

	memcpy( pBuffer, (void*)udc->dtcm, length );

	req->req.actual += length;

	if (is_complete)
	{
		req->req.status = 0;
		list_del_init(&req->queue);
		spin_unlock(&udc->lock);
		req->req.complete(&ep->ep, &req->req);
		spin_lock(&udc->lock);
	}

	ep->is_working = 0;
	HwEIER |= 1 << ep->index;

	spin_unlock(&udc->lock);

	return IRQ_HANDLED;
}
#endif

static void usbt_control_irq(struct usbt_udc_t *udc, struct usbt_ep_t *ep)
{
	//struct usba_request *req;
	//u32 epstatus;
	//u32 epctrl;
	u32 IR_Save;

	IR_Save	= HwIR;
	HwIR	= 0;

	//----------------------------------
	// STALL
	//----------------------------------
	if ( HwEP0SR & HwEP0SR_SHT )
	{
		HwEP0SR		= HwEP0SR_SHT;
		//gTCC7XXUSB_EP0TransmitLen = 0;
	}

//restart:
	//epstatus = usba_ep_readl(ep, STA);
	//epctrl = usba_ep_readl(ep, CTL);
	//
	//DBG(DBG_INT, "%s [%d]: s/%08x c/%08x\n",
	//		ep->ep.name, ep->state, epstatus, epctrl);

	//req = NULL;
	//if (!list_empty(&ep->queue))
	//	req = list_entry(ep->queue.next,
	//			 struct usba_request, queue);

	//if ((epctrl & USBA_TX_PK_RDY) && !(epstatus & USBA_TX_PK_RDY)) {
	//	if (req->submitted)
	//		next_fifo_transaction(ep, req);
	//	else
	//		submit_request(ep, req);
	//
	//	if (req->last_transaction) {
	//		usba_ep_writel(ep, CTL_DIS, USBA_TX_PK_RDY);
	//		usba_ep_writel(ep, CTL_ENB, USBA_TX_COMPLETE);
	//	}
	//	goto restart;
	//}
	//if ((epstatus & epctrl) & USBA_TX_COMPLETE) {
	//	usba_ep_writel(ep, CLR_STA, USBA_TX_COMPLETE);
	//
	//	switch (ep->state) {
	//	case DATA_STAGE_IN:
	//		usba_ep_writel(ep, CTL_ENB, USBA_RX_BK_RDY);
	//		usba_ep_writel(ep, CTL_DIS, USBA_TX_COMPLETE);
	//		ep->state = STATUS_STAGE_OUT;
	//		break;
	//	case STATUS_STAGE_ADDR:
	//		/* Activate our new address */
	//		usba_writel(udc, CTRL, (usba_readl(udc, CTRL)
	//					| USBA_FADDR_EN));
	//		usba_ep_writel(ep, CTL_DIS, USBA_TX_COMPLETE);
	//		ep->state = WAIT_FOR_SETUP;
	//		break;
	//	case STATUS_STAGE_IN:
	//		if (req) {
	//			list_del_init(&req->queue);
	//			request_complete(ep, req, 0);
	//			submit_next_request(ep);
	//		}
	//		usba_ep_writel(ep, CTL_DIS, USBA_TX_COMPLETE);
	//		ep->state = WAIT_FOR_SETUP;
	//		break;
	//	case STATUS_STAGE_TEST:
	//		usba_ep_writel(ep, CTL_DIS, USBA_TX_COMPLETE);
	//		ep->state = WAIT_FOR_SETUP;
	//		if (do_test_mode(udc))
	//			set_protocol_stall(udc, ep);
	//		break;
	//	default:
	//		printk(KERN_ERR
	//			"udc: %s: TXCOMP: Invalid endpoint state %d, "
	//			"halting endpoint...\n",
	//			ep->ep.name, ep->state);
	//		set_protocol_stall(udc, ep);
	//		break;
	//	}
	//
	//	goto restart;
	//}
	//if ((epstatus & epctrl) & USBA_RX_BK_RDY) {
	//	switch (ep->state) {
	//	case STATUS_STAGE_OUT:
	//		usba_ep_writel(ep, CLR_STA, USBA_RX_BK_RDY);
	//		usba_ep_writel(ep, CTL_DIS, USBA_RX_BK_RDY);
	//
	//		if (req) {
	//			list_del_init(&req->queue);
	//			request_complete(ep, req, 0);
	//		}
	//		ep->state = WAIT_FOR_SETUP;
	//		break;
	//
	//	case DATA_STAGE_OUT:
	//		receive_data(ep);
	//		break;
	//
	//	default:
	//		usba_ep_writel(ep, CLR_STA, USBA_RX_BK_RDY);
	//		usba_ep_writel(ep, CTL_DIS, USBA_RX_BK_RDY);
	//		printk(KERN_ERR
	//			"udc: %s: RXRDY: Invalid endpoint state %d, "
	//			"halting endpoint...\n",
	//			ep->ep.name, ep->state);
	//		set_protocol_stall(udc, ep);
	//		break;
	//	}
	//
	//	goto restart;
	//}

	if ( HwEP0SR & HwEP0SR_RSR )
	{
		union {
			struct usb_ctrlrequest crq;
			unsigned long data[2];
		} crq;
		u32 pkt_len;
		int ret;

		//if (ep->state != WAIT_FOR_SETUP) {
			/*
			 * Didn't expect a SETUP packet at this
			 * point. Clean up any pending requests (which
			 * may be successful).
			 */
			//int status = -EPROTO;

			/*
			 * RXRDY and TXCOMP are dropped when SETUP
			 * packets arrive.  Just pretend we received
			 * the status packet.
			 */
			//if (ep->state == STATUS_STAGE_OUT
			//		|| ep->state == STATUS_STAGE_IN) {
			//	usba_ep_writel(ep, CTL_DIS, USBA_RX_BK_RDY);
			//	status = 0;
			//}

			//if (req) {
			//	list_del_init(&req->queue);
			//	request_complete(ep, req, status);
			//}
		//}

		usbt_unstall(ep);
		pkt_len = usbt_raw_read_ep0();//USBA_BFEXT(BYTE_COUNT, usba_ep_readl(ep, STA));
		//DBG(DBG_HW, "Packet length: %u\n", pkt_len);
		if (pkt_len == sizeof(crq))
		{
			//DBG(DBG_FIFO, "Copying ctrl request");
			memcpy((void*)crq.data, (const void*)s_usbt_ep0_buffer, sizeof(crq));

			/* Free up one bank in the FIFO so that we can
			 * generate or receive a reply right away. */
			//usba_ep_writel(ep, CLR_STA, USBA_RX_SETUP);

			/* printk(KERN_DEBUG "setup: %d: %02x.%02x\n",
				ep->state, crq.crq.bRequestType,
				crq.crq.bRequest); */

			if (crq.crq.bRequestType & USB_DIR_IN) {
				/*
				 * The USB 2.0 spec states that "if wLength is
				 * zero, there is no data transfer phase."
				 * However, testusb #14 seems to actually
				 * expect a data phase even if wLength = 0...
				 */
				//ep->state = DATA_STAGE_IN;
				ep->is_in = 1;
			} else {
				//if (crq.crq.wLength != __constant_cpu_to_le16(0))
					//ep->state = DATA_STAGE_OUT;
				//else
					//ep->state = STATUS_STAGE_IN;
				ep->is_in = 0;
			}

			ret = -1;
			if (ep->index == 0)
				ret = usbt_handle_ep0_setup(udc, ep, &crq.crq);
			else {
				spin_unlock(&udc->lock);
				ret = udc->driver->setup(&udc->gadget, &crq.crq);
				spin_lock(&udc->lock);
			}

			//DBG(DBG_BUS, "req %02x.%02x, length %d, ret %d\n",
			//	crq.crq.bRequestType, crq.crq.bRequest,
			//	le16_to_cpu(crq.crq.wLength), ret);

			if (ret < 0) {
				/* Let the host know that we failed */
				usbt_stall(ep);
			}
		}
		else
		{
			//printk(KERN_WARNING "udc: Invalid packet length %d "
			//	"(expected %d)\n", pkt_len, sizeof(crq));
			//usbt_stall(ep);
		}
	}
	else if ( HwEP0SR & HwEP0SR_TST )
	{
		HwEP0SR = HwEP0SR_TST;
	}

	HwIR	= IR_Save;
}

static void usbt_ep_irq(struct usbt_udc_t *udc, struct usbt_ep_t *ep)
{
	//struct usba_request *req;
	//u32 epstatus;
	//u32 epctrl;

	//epstatus = usba_ep_readl(ep, STA);
	//epctrl = usba_ep_readl(ep, CTL);

	//DBG(DBG_INT, "%s: interrupt, status: 0x%08x\n", ep->ep.name, epstatus);

	if( ep->is_in )
	{
		u32 IRSave;

		IRSave = HwIR;
		HwIR = ep->index;
		if( HwESR & HwESR_TPS )
			HwESR = HwESR_TPS;
		HwIR = IRSave;
 	}

	if (list_empty(&ep->queue)) {
		//DBG(DBG_DEBUG,"ep irq - queue empty");
		if( ep->is_in == 0 )
			HwEIER &= ~(1 << ep->index);
		if( ep->is_in == 1 )
		{
			//DBG(DBG_DEBUG,"is in now = 0");
			ep->is_working = 0;
		}
		return;
	}

	if( ep->is_in )
	{
	//while ((epctrl & USBA_TX_PK_RDY) && !(epstatus & USBA_TX_PK_RDY)) {
	//	DBG(DBG_BUS, "%s: TX PK ready\n", ep->ep.name);
	//
	//	if (list_empty(&ep->queue)) {
	//		dev_warn(&udc->pdev->dev, "ep_irq: queue empty\n");
	//		usba_ep_writel(ep, CTL_DIS, USBA_TX_PK_RDY);
	//		return;
	//	}
	//
	//	req = list_entry(ep->queue.next, struct usba_request, queue);
	//
	//	if (req->using_dma) {
	//		/* Send a zero-length packet */
	//		usba_ep_writel(ep, SET_STA,
	//				USBA_TX_PK_RDY);
	//		usba_ep_writel(ep, CTL_DIS,
	//				USBA_TX_PK_RDY);
	//		list_del_init(&req->queue);
	//		submit_next_request(ep);
	//		request_complete(ep, req, 0);
	//	} else {
	//		if (req->submitted)
	//			next_fifo_transaction(ep, req);
	//		else
	//			submit_request(ep, req);
	//
	//		if (req->last_transaction) {
	//			list_del_init(&req->queue);
	//			submit_next_request(ep);
	//			request_complete(ep, req, 0);
	//		}
	//	}
	//
	//	epstatus = usba_ep_readl(ep, STA);
	//	epctrl = usba_ep_readl(ep, CTL);
	//}
		//DBG(DBG_DEBUG,"%s: TX data", ep->ep.name);
		//if( ep->req->status == -EINPROGRESS )
			#ifdef EP_WRITE_INT
			usbt_write_ep(ep);
			#endif
	}
	//if ((epstatus & epctrl) & USBA_RX_BK_RDY)
	else
	{
		//DBG(DBG_DEBUG, "%s: RX data ready", ep->ep.name);
		//if( HwEIER & (1<<ep->index) )
			usbt_read_ep(ep);
		//usba_ep_writel(ep, CLR_STA, USBA_RX_BK_RDY);
	}
}

//#define __TEST_GPIO__	/* XXX: twkwon */
#ifdef __TEST_GPIO__
static void test_gpio_set(void)
{
#ifdef CONFIG_ARCH_TCC83X
	// GPIO_B0 
	BITCLR(HwGSEL_B, Hw14);	
	BITSET(HwNGSEL_B, Hw0);
	BITSET(HwGIOCON_B, Hw0);
	BITCLR(HwGDATA_B, Hw0);
#elif CONFIG_ARCH_TCC79X
	// GPIO_D7
	BITCSET(HwPORTCFG4, Hw28-Hw24, Hw24);	
	BITSET(HwGPDEN, Hw7);
	BITCLR(HwGPDDAT, Hw7);
#endif
}

#define gpio_high 1
#define gpio_low  0
static void test_gpio_ctr(int level)
{
#ifdef CONFIG_ARCH_TCC83X
	if (level == gpio_high) {
		BITSET(HwGDATA_B, Hw0);
	} else if (level == gpio_low) { 
		BITCLR(HwGDATA_B, Hw0);
	}
#elif CONFIG_ARCH_TCC79X
	if (level == gpio_high) {
		BITSET(HwGPDDAT, Hw7);
	} else if (level == gpio_low) { 
		BITCLR(HwGPDDAT, Hw7);
	}
#endif
}
#endif

static irqreturn_t usbt_udc_irq(int irq, void *devid)
{
	struct usbt_udc_t *udc = devid;
	//u32 status;
	//u32 dma_status;
	//u32 ep_status;
	u32 SSR;
	u32 EIR;

#ifdef __TEST_GPIO__
	test_gpio_ctr(gpio_high);
#endif
	spin_lock(&udc->lock);

	//status = usba_readl(udc, INT_STA);
	//DBG(DBG_INT, "irq, status=%#08x\n", status);
	SSR = HwSSR;

	if (SSR & HwSSR_HFRES) {
		//struct usbt_ep_t *ep0;

		//usba_writel(udc, INT_CLR, USBA_END_OF_RESET);
		HwSSR	= HwSSR_HFRES;				// Clear the Reset interrupt flag
		usbt_reset();
		usbt_reset_all_endpoints(udc);
		HwSCR	|= HwSCR_HSUSPE_EN;

		//ep0 = &usbt_ep[0];
		//ep0->desc = &usbt_ep0_desc;
		//ep0->state = WAIT_FOR_SETUP;
		//usba_ep_writel(ep0, CFG,
		//		(USBA_BF(EPT_SIZE, EP0_EPT_SIZE)
		//		| USBA_BF(EPT_TYPE, USBA_EPT_TYPE_CONTROL)
		//		| USBA_BF(BK_NUMBER, USBA_BK_NUMBER_ONE)));
		//usba_ep_writel(ep0, CTL_ENB,
		//		USBA_EPT_ENABLE | USBA_RX_SETUP);
		//usba_writel(udc, INT_ENB,
		//		(usba_readl(udc, INT_ENB)
		//		| USBA_BF(EPT_INT, 1)
		//		| USBA_DET_SUSPEND
		//		| USBA_END_OF_RESUME));

		//if (!(usba_ep_readl(ep0, CFG) & USBA_EPT_MAPPED))
		//	dev_warn(&udc->pdev->dev,
		//		 "WARNING: EP0 configuration is invalid!\n");
	}

	if( SSR & HwSSR_HFSUSP )
	{
		HwSSR	= HwSSR_HFSUSP;			// Clear the Suspend interrupt
		HwSCR	|= HwSCR_HRESE_EN;
		HwSCR	&= HwSCR_HSUSPE_DIS;		// Disable Suspend interrupt
		DBG(DBG_BUS, "Suspend detected");
		if (udc->gadget.speed != USB_SPEED_UNKNOWN
				&& udc->driver && udc->driver->suspend) {
			spin_unlock(&udc->lock);
			udc->driver->suspend(&udc->gadget);
			spin_lock(&udc->lock);
		}
	}

	//if (status & USBA_WAKE_UP) {
	//	usba_writel(udc, INT_CLR, USBA_WAKE_UP);
	//	DBG(DBG_BUS, "Wake Up CPU detected\n");
	//}

	if ( SSR & HwSSR_HFRM )
	{
		HwSSR	= HwSSR_HFRM;
		usbt_reset();
		HwSCR	|= HwSCR_HSUSPE_EN;
		DBG(DBG_BUS, "Resume detected");
		if (udc->gadget.speed != USB_SPEED_UNKNOWN
				&& udc->driver && udc->driver->resume) {
			spin_unlock(&udc->lock);
			udc->driver->resume(&udc->gadget);
			spin_lock(&udc->lock);
		}
	}

	//dma_status = USBA_BFEXT(DMA_INT, status);
	//if (dma_status) {
	//	int i;
	//
	//	for (i = 1; i < USBA_NR_ENDPOINTS; i++)
	//		if (dma_status & (1 << i))
	//			usba_dma_irq(udc, &usba_ep[i]);
	//}

	EIR = HwEIR;

	if ( EIR & ( HwEIR_EP0INT + HwEIR_EP1INT + HwEIR_EP2INT + HwEIR_EP3INT ) )
	{
		int i;
		for (i = 0; i < N_ENDPOINT ; i++)
		{
			if( EIR & (1<<i) )
			{
				if( i == 0 )
					usbt_control_irq(udc, &usbt_ep[0]);
				else
				{
					usbt_ep_irq(udc, &usbt_ep[i]);
				}
				HwEIR = 1<<i;
			}
		}
	}

	if ( SSR & HwSSR_SDE_END )
	{
		if ( HwSSR & HwSSR_HSP_HIGH )
		{
			//if ( USB_GetIsDeviceType(USB_GetIsCurrentClassType()) == USB_HIGHSPEED )
			HwUTST = 0x0000; 	// SET DEVICE AS USB 2.0 HIGH SPEED
			DBG(DBG_BUS, "High-speed bus reset detected");
			udc->gadget.speed = USB_SPEED_HIGH;
		}
		else
		{
			//gTCC7XXUSB_IsHighOrFullSpeed = USB_FULLSPEED;
			HwUTST = 0x0080; 	// SET DEVICE AS USB 1.1 FULLSPEED
			DBG(DBG_BUS, "Full-speed bus reset detected");
			udc->gadget.speed = USB_SPEED_FULL;
		}

		//TCC7XX_USBDRV_SetSpeed();

		HwSSR	= HwSSR_SDE_END;			// Clear the Suspend interrupt
	}

	if ( HwSSR & HwSSR_HSP_HIGH )
	{
		//if ( gTCC7XXUSB_IsHighOrFullSpeed != USB_HIGHSPEED )
		//{
		//	gTCC7XXUSB_IsHighOrFullSpeed = USB_HIGHSPEED;
		//	TCC7XX_USBDRV_SetSpeed();
		//	DBG(DBG_BUS, "High-speed bus reset detected\n");
			udc->gadget.speed = USB_SPEED_HIGH;
		//}
	}
	else
	{
		//if ( gTCC7XXUSB_IsHighOrFullSpeed != USB_FULLSPEED )
		//{
		//	gTCC7XXUSB_IsHighOrFullSpeed = USB_FULLSPEED;
		//	TCC7XX_USBDRV_SetSpeed();
		//	DBG(DBG_BUS, "Full-speed bus reset detected\n");
			udc->gadget.speed = USB_SPEED_FULL;
		//}
	}

	spin_unlock(&udc->lock);
#ifdef __TEST_GPIO__
	test_gpio_ctr(gpio_low);
#endif

	return IRQ_HANDLED;
}

int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	struct usbt_udc_t *udc = &the_udc;
	unsigned long flags;
	int ret;

	DBG(DBG_DEBUG, "usb gadget register driver start...");

	if (!udc->pdev)
	{
		DBG(DBG_ERROR, "udc->pdev is not exist");
		return -ENODEV;
	}

	spin_lock_irqsave(&udc->lock, flags);
	if (udc->driver) {
		spin_unlock_irqrestore(&udc->lock, flags);
		DBG(DBG_ERROR, "udc->driver is exist");
		return -EBUSY;
	}

	udc->devstatus = 1 << USB_DEVICE_SELF_POWERED;
	udc->driver = driver;
	udc->gadget.dev.driver = &driver->driver;
	spin_unlock_irqrestore(&udc->lock, flags);

	//clk_enable(udc->pclk);
	//clk_enable(udc->hclk);

	ret = driver->bind(&udc->gadget);
	if (ret) {
		DBG(DBG_ERR, "Could not bind to driver %s: error %d",
			driver->driver.name, ret);
		goto err_driver_bind;
	}

	DBG(DBG_GADGET, "registered driver `%s'", driver->driver.name);

	//udc->vbus_prev = 0;
	//if (udc->vbus_pin != -1)
	//	enable_irq(gpio_to_irq(udc->vbus_pin));

	/* If Vbus is present, enable the controller and wait for reset */
	spin_lock_irqsave(&udc->lock, flags);
	//if (vbus_is_present(udc) && udc->vbus_prev == 0) {
	//	usba_writel(udc, CTRL, USBA_EN_USBA);
	//	usba_writel(udc, INT_ENB, USBA_END_OF_RESET);
	//}
	spin_unlock_irqrestore(&udc->lock, flags);

	return 0;

err_driver_bind:
	udc->driver = NULL;
	udc->gadget.dev.driver = NULL;
	return ret;
}
EXPORT_SYMBOL(usb_gadget_register_driver);

int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
	struct usbt_udc_t *udc = &the_udc;
	unsigned long flags;

	if (!udc->pdev)
		return -ENODEV;
	if (driver != udc->driver)
		return -EINVAL;

	//if (udc->vbus_pin != -1)
	//	disable_irq(gpio_to_irq(udc->vbus_pin));

	spin_lock_irqsave(&udc->lock, flags);
	udc->gadget.speed = USB_SPEED_UNKNOWN;
	usbt_reset_all_endpoints(udc);
	spin_unlock_irqrestore(&udc->lock, flags);

	/* This will also disable the DP pullup */
	//usba_writel(udc, CTRL, 0);

	driver->unbind(&udc->gadget);
	udc->gadget.dev.driver = NULL;
	udc->driver = NULL;

	//clk_disable(udc->hclk);
	//clk_disable(udc->pclk);

	usbt_detach();

	DBG(DBG_GADGET, "unregistered driver `%s'\n", driver->driver.name);

	return 0;
}
EXPORT_SYMBOL(usb_gadget_unregister_driver);

static int usbt_vbus_thread(void* _udc)
{
	struct usbt_udc_t *udc = _udc;

	DBG(DBG_DEBUG,"usbt_vbus_thread start!");

	// Setting ARM Interrupt relevant to VBUS On
	#if defined(CONFIG_ARCH_TCC83X)
	BITSET( HwSEL_B, HwSEL_VBON );			// Considered as IRQ interrupt
	BITSET( HwMODE_B, HwMODE_VBON );		// level trigger type for VBUS On interrupt
	BITCLR( HwPOL_B, HwPOL_VBON );			// Active High for VBUS On interrupt polarity

	#define IsUSBDET()	( HwSTS_B & HwSTS_VBON )
	#elif defined(CONFIG_ARCH_TCC79X)
	BITCSET(HwEINTSEL1, HwEINTSEL1_EINT7SEL_MASK, HwEINTSEL1_EINT7(31));
	BITCSET(HwIRQSEL, HwIRQSEL_SEL4_MASK, HwIRQSEL_SEL4_EINT7);
	BITSET(HwMODE, HwINT_SEL4);				// level trigger type for VBUS ON

	#define IsUSBDET()	( HwSRC & HwINT_SEL4)
	#endif

	udc->vbus_prev = 0;

	while(1)
	{
		msleep_interruptible(200);
		
		if( IsUSBDET() && udc->driver )
		{
			if( udc->vbus_prev == 0 )
			{
				spin_lock(&udc->lock);

				DBG(DBG_DEBUG,"vbus on");
				udc->vbus_prev = 1;
				//HwBCLKCTR &= HwBCLKCTR_UBH_OFF;	// USB Host clock off
				//HwBCLKCTR |= HwBCLKCTR_UB_ON;	// USB Device clock on

				// USB Host interrupt disable
				#if defined(CONFIG_ARCH_TCC79X)
				BITCLR(HwIRQMSK, HwINT_UH);
				#elif defined(CONFIG_ARCH_TCC83X)
				BITCLR(HwIEN, HwIEN_USBH);
				#endif

				usbt_reset();
				//enable_irq(udc->irq);

				/* 
				 * Send "online" event to user-level via /sys/devices/platform/tcc-udc.0 
				 *    msg: online@/devices/platform/tcc-udc.0
				 */
				kobject_uevent(&udc->pdev->dev.kobj, KOBJ_ONLINE);

				spin_unlock(&udc->lock);
			}
			else if (udc->vbus_prev == 3) 
			{
				/* For Suspend/Resume */
				DBG(DBG_DEBUG,"resume vbus on");
				udc->vbus_prev = 1;
				//HwBCLKCTR |= HwBCLKCTR_UB_ON;
				usbt_detach();
				udelay(5);
				usbt_reset();
			}
		}
		else
		{
			if( udc->vbus_prev != 0 )
			{
				spin_lock(&udc->lock);

				DBG(DBG_DEBUG,"vbus off");
				udc->vbus_prev = 0;
				usbt_reset_all_endpoints(udc);
				if(udc->driver)
				{
					spin_unlock(&udc->lock);
					udc->driver->disconnect(&udc->gadget);
					spin_lock(&udc->lock);
				}

				usbt_detach();

				HwUPCFG0    = 0x08B0;
				#if defined(CONFIG_ARCH_TCC83X)
				HwUPCFG1	= 0x3534;
				HwUPCFG2	= 0x51A0;	// HS DC voltage level adjustment = 10%
				#elif defined(CONFIG_ARCH_TCC79X)
				HwUPCFG1	= 0x7343;	// design default
				HwUPCFG2	= 0x5100;	// design default
				#endif
				HwUPCFG3	= 0x9000;	// OTG disable
				HwPLICR		= 0x0000;

				//HwBCLKCTR &= HwBCLKCTR_UB_OFF;	// USB Device clock off
				//HwBCLKCTR |= HwBCLKCTR_UBH_ON;	// USB Host clock on

				// USB Host interrupt enable
				#if defined(CONFIG_ARCH_TCC79X)
				BITSET(HwIRQMSK, HwINT_UH);
				#elif defined(CONFIG_ARCH_TCC83X)
				BITSET(HwIEN, HwIEN_USBH);
				#endif
				
				//disable_irq(udc->irq);

				/* 
				 * Send "offline" event to user-level via /sys/devices/platform/tcc-udc.0
				 *    msg: offline@/devices/platform/tcc-udc.0
				 */
				kobject_uevent(&udc->pdev->dev.kobj, KOBJ_OFFLINE);
				
				spin_unlock(&udc->lock);
			}
		}
	}
}

static int __init usbt_udc_probe(struct platform_device *pdev)
{
		//struct usba_platform_data *pdata = pdev->dev.platform_data;
		//struct resource *regs, *fifo;
		//struct clk *pclk, *hclk;
	struct usbt_udc_t *udc = &the_udc;
	int ret, irq, i;
	#if defined(USE_DMA_IRQ)
	int dma_irq;
	#endif

	DBG(DBG_DEBUG, "usbt udc probe start...");

		//regs = platform_get_resource(pdev, IORESOURCE_MEM, CTRL_IOMEM_ID);
		//fifo = platform_get_resource(pdev, IORESOURCE_MEM, FIFO_IOMEM_ID);
		//if (!regs || !fifo)
		//	return -ENXIO;

	irq = platform_get_irq(pdev, 0);
	if (irq < 0)
		return irq;

	#if defined(USE_DMA_IRQ)
	dma_irq= platform_get_irq(pdev, 1);
	if (dma_irq < 0)
		return dma_irq;
	#endif

		//pclk = clk_get(&pdev->dev, "pclk");
		//if (IS_ERR(pclk))
		//	return PTR_ERR(pclk);
		//hclk = clk_get(&pdev->dev, "hclk");
		//if (IS_ERR(hclk)) {
		//	ret = PTR_ERR(hclk);
		//	goto err_get_hclk;
		//}

	udc->pdev = pdev;
		//udc->pclk = pclk;
		//udc->hclk = hclk;
	//HwBCLKCTR |= HwBCLKCTR_UB_ON;
		//udc->vbus_pin = -1;

	ret = -ENOMEM;
		//udc->fifo = ioremap(fifo->start, fifo->end - fifo->start + 1);
		//if (!udc->fifo) {
		//	dev_err(&pdev->dev, "Unable to map FIFO, aborting.\n");
		//	goto err_map_fifo;
		//}
	udc->dtcm = ioremap(IO_USB_BUFFER0_BASE_PA, 512);
	if( !udc->dtcm )
		goto err_request_irq;
	udc->dtcm_pa = IO_USB_BUFFER0_BASE_PA;

	device_initialize(&udc->gadget.dev);
	udc->gadget.dev.parent = &pdev->dev;
	udc->gadget.dev.dma_mask = pdev->dev.dma_mask;

	platform_set_drvdata(pdev, udc);

	/* Make sure we start from a clean slate */
		//clk_enable(pclk);
		//usba_writel(udc, CTRL, 0);
		//clk_disable(pclk);

	INIT_LIST_HEAD(&usbt_ep[0].ep.ep_list);
	for (i = 1; i < ARRAY_SIZE(usbt_ep); i++) {
		struct usbt_ep_t *ep = &usbt_ep[i];

		list_add_tail(&ep->ep.ep_list, &udc->gadget.ep_list);
	}

	BITSET( HwMODE, HwINT_UD );		// level trigger type for USB interrupt
	BITSET( HwSEL, HwINT_UD );

	#if defined(CONFIG_ARCH_TCC83X)
	BITCLR( HwPOL, HwINT_UD );		// active high (usb's active low interrupt is inverted.)
	#elif defined(CONFIG_ARCH_TCC79X)
	BITSET( HwPOL, HwINT_UD );		// active low (usb's interrupt is active low)
	BITSET( HwIRQMSK, HwINT_UD );
	BITCLR( HwMODEA, HwINT_UD);
	#endif

	//ret = request_irq(irq, usbt_udc_irq, 0, driver_name, udc);
	ret = request_irq(irq, usbt_udc_irq, IRQF_DISABLED, driver_name, udc);
	if (ret) {
		dev_err(&pdev->dev, "Cannot request irq %d (error %d)",
			irq, ret);
		goto err_request_irq;
	}
	udc->irq = irq;

	#if defined(USE_DMA_IRQ)
	ret = request_irq(dma_irq, usbt_dma_irq, IRQF_DISABLED, driver_name, udc);
	if (ret) {
		dev_err(&pdev->dev, "Cannot request dma irq %d (error %d)",
			dma_irq, ret);
		goto err_request_dma_irq;
	}
	udc->dma_irq = dma_irq;
	#endif

	ret = device_add(&udc->gadget.dev);
	if (ret) {
		dev_dbg(&pdev->dev, "Could not add gadget: %d", ret);
		goto err_device_add;
	}

	//usbt_reset();
	//HwBCLKCTR &= HwBCLKCTR_UB_OFF;
	HwBCLKCTR |= HwBCLKCTR_UB_ON;
	//disable_irq(udc->irq);

	udc->thread_task = kthread_create(usbt_vbus_thread, udc, "usbt-vbus-thread");
	if (IS_ERR(udc->thread_task)) {
		printk("\nPTR ERR %p", udc->thread_task);
		goto err_device_add;
	}

	/* Tell the thread to start working */
	wake_up_process(udc->thread_task);

	DBG(DBG_DEBUG, "usbt udc probe succeeds");

	return 0;

err_device_add:
	free_irq(udc->dma_irq, udc);
#if defined(USE_DMA_IRQ)
err_request_dma_irq:
	free_irq(udc->irq, udc);
#endif
err_request_irq:
//	iounmap(udc->fifo);
//err_map_fifo:
//	iounmap(udc->regs);
//err_map_regs:
	//clk_put(hclk);
//err_get_hclk:
	//clk_put(pclk);
	if( udc->dtcm )
	{
		iounmap(udc->dtcm);
	}

	//HwBCLKCTR &= HwBCLKCTR_UB_OFF;

	platform_set_drvdata(pdev, NULL);

	return ret;
}

static int __exit usbt_udc_remove(struct platform_device *pdev)
{
	struct usbt_udc_t *udc;
	//int i;

	DBG(DBG_DEBUG, "usbt udc remove start...");

	udc = platform_get_drvdata(pdev);

	//for (i = 1; i < ARRAY_SIZE(usba_ep); i++)
	//	usba_ep_cleanup_debugfs(&usba_ep[i]);
	//usba_cleanup_debugfs(udc);

	//if (udc->vbus_pin != -1)
	//	gpio_free(udc->vbus_pin);

	free_irq(udc->irq, udc);
	#if defined(USE_DMA_IRQ)
	free_irq(udc->dma_irq, udc);
	#endif
	//iounmap(udc->fifo);
	//iounmap(udc->regs);
	//clk_put(udc->hclk);
	//clk_put(udc->pclk);
	iounmap(udc->dtcm);
	//HwBCLKCTR &= HwBCLKCTR_UB_OFF;

	device_unregister(&udc->gadget.dev);

	printk("%s: remove\n", driver_name);
	return 0;
}

/*-------------------------------------------------------------------------*/

#ifdef  CONFIG_PM
static int usbt_udc_suspend(struct platform_device *dev, pm_message_t state)
{
	struct usbt_udc_t *udc = platform_get_drvdata(dev);
	HwUSBPHYCFG = 0x2940;
	if (udc->vbus_prev == 1) {
	/* Reset USB device */
	HwSWRESET |= HwSWRESET_UB_ON;
	HwSWRESET &= ~HwSWRESET_UB_ON;
	}

    return 0;
}

static int usbt_udc_resume(struct platform_device *dev)
{
	struct usbt_udc_t *udc = platform_get_drvdata(dev);
	HwUSBPHYCFG = 0x2940;
	if (udc->vbus_prev == 1) {
		udc->vbus_prev = 3;
		//HwBCLKCTR |= HwBCLKCTR_UB_ON;
		//usbt_reset();
	}

    return 0;
}
#else
#define usbt_udc_suspend	NULL
#define usbt_udc_resume		NULL
#endif

/*-------------------------------------------------------------------------*/

static struct platform_driver udc_driver = {
	//.probe      = usbt_udc_probe,
	.remove		= __exit_p(usbt_udc_remove),
	.suspend	= usbt_udc_suspend,
	.resume		= usbt_udc_resume,
	.driver		= {
		.owner	= THIS_MODULE,
		.name	= driver_name,
	},
};

static int __init udc_init(void)
{
	printk("%s: init\n", driver_name);
#ifdef __TEST_GPIO__
	test_gpio_set();
#endif
	DBG(DBG_DEBUG, "udc_init start... %s %s",DRIVER_DESC,DRIVER_VERSION);
	//printk(KERN_INFO "%s: version %s\n", driver_name, DRIVER_VERSION);
	//return platform_driver_register(&udc_driver);
	return platform_driver_probe(&udc_driver, usbt_udc_probe);
}
module_init(udc_init);

static void __exit udc_exit(void)
{
	printk("%s: exit\n", driver_name);
	platform_driver_unregister(&udc_driver);
}
module_exit(udc_exit);

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR("Frank Becker, Robert Schwebel, David Brownell");
MODULE_LICENSE("GPL");

