/*
 * linux/drivers/usb/host/ohci-tcc.c
 *
 * Description: OHCI HCD (Host Controller Driver) for USB.
 *              Bus Glue for Telechips TCCXXX
 *
 * (C) Copyright 1999 Roman Weissgaerber <weissg@vienna.at>
 * (C) Copyright 2000-2002 David Brownell <dbrownell@users.sourceforge.net>
 * (C) Copyright 2002 Hewlett-Packard Company
 * (C) Copyright 2008-2009 Telechips <linux@telechips.com>
 * (C) Copyright 2008-2009 Telechips 
 *
 * Rewritten by <linux@telechips.com>
 * Written by Christopher Hoover <ch@hpl.hp.com>
 * Based on fragments of previous driver by Russell King et al.
 *
 * Modified for LH7A404 from ohci-sa1111.c
 *  by Durgesh Pattamatta <pattamattad@sharpsec.com>
 *
 * Modified for pxa27x from ohci-lh7a404.c
 *  by Nick Bane <nick@cecomputing.co.uk> 26-8-2004
 *
 * Modified for TCCXXX from ohci-pxa27x.c
 *  by Telechips Linux Team <linux@telechips.com> 03-9-2008
 *
 * This file is licenced under the GPL.
 */

#include <linux/device.h>
#include <linux/signal.h>
#include <linux/platform_device.h>

#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <mach/ohci.h>

#include <asm/io.h>
#include <asm/irq.h>

#if defined(CONFIG_ARCH_TCC83X)
#include <asm/arch/tcc83x_virt.h>
#elif defined(CONFIG_ARCH_TCC79X)
#include <mach/tcc79x.h>
#else
#error "Unknown Target"
#endif

#include <asm/plat-tcc/tcc_ckc.h>


/*
  PMM_NPS_MODE -- PMM Non-power switching mode
      Ports are powered continuously.

  PMM_GLOBAL_MODE -- PMM global switching mode
      All ports are powered at the same time.

  PMM_PERPORT_MODE -- PMM per port switching mode
      Ports are powered individually.
 */
static int tcc_ohci_select_pmm( int mode )
{
	switch ( mode ) {
	case PMM_NPS_MODE:
		HwRhDescriptorA |= USBOHCI_UHCRHDA_NPS;
		break;
	case PMM_GLOBAL_MODE:
		HwRhDescriptorA &= ~(USBOHCI_UHCRHDA_NPS & USBOHCI_UHCRHDA_PSM_PERPORT);
		break;
	case PMM_PERPORT_MODE:
		HwRhDescriptorA &= ~(USBOHCI_UHCRHDA_NPS);
		HwRhDescriptorA |= USBOHCI_UHCRHDA_PSM_PERPORT;

		/* Set port power control mask bits, only 1 ports. */
		HwRhDescriptorB |= (0x1<<17);
		break;
	default:
		printk(KERN_ERR "Invalid mode %d, set to non-power switch mode.\n", mode);
		HwRhDescriptorA |= USBOHCI_UHCRHDA_NPS;
	}

	return 0;
}

extern int usb_disabled(void);

/*-------------------------------------------------------------------------*/

void TurnOnUSBHostClocks(void)
{
    HwBCLKCTR |= HwBCLKCTR_UBH_ON;	// USB1.1 Host Controller
#if defined(CONFIG_ARCH_TCC83X)
	tca_ckc_setperi(PERI_USB11H, MHz, 48, CLKFROM48M);	// Select USB Host 48Mhz clock from USB PHY.
														// *(volatile unsigned long *)0xF10130A4 = 0x1;
#elif defined(CONFIG_ARCH_TCC79X)
	tca_ckc_setperi(PERI_USB11H, MHz, 48, CLKFROM48M);	// Select USB Host 48Mhz clock from USB PHY.
#endif
}

void TurnOffUSBHostClocks(void)
{
    HwBCLKCTR &= HwBCLKCTR_UBH_OFF;
}

void USBHostReset(void)
{
	//HwBCLKCTR |= HwBCLKCTR_UB_ON;	//--> USB2.0 on before USBH config
	
#if defined(CONFIG_ARCH_TCC83X)
	BITSET(HwSWRESET, Hw11); //S/W reset on
	//msleep(1);
	BITCLR(HwSWRESET, Hw11); //S/W reset off
#elif defined(CONFIG_ARCH_TCC79X)
	BITSET(HwSWRESET, Hw0); //S/W reset on
	//msleep(1);
	BITCLR(HwSWRESET, Hw0); //S/W reset off
#endif

	// USBH config
	HwUPCFG0 = 0x800008b0; //800008b0
	HwUPCFG2 = 0x00005500; //12MHz
	HwUPCFG1 = 0x3634;
	HwUPCFG3 = 0x9040;
	
	//HwBCLKCTR &= HwBCLKCTR_UB_OFF;	//<-- USB2.0 off after USBH config
}

int tcc_set_cken(int sw)
{
	if (sw){
	    TurnOnUSBHostClocks();
		USBHostReset();
	} else {
		TurnOffUSBHostClocks();
	}
	return 0;
}

static int tcc_start_hc(struct device *dev)
{
	int retval = 0;
	struct tccohci_platform_data *inf;

	inf = dev->platform_data;

	tcc_set_cken(1);

	if (inf->init)
		retval = inf->init(dev);

	if (retval < 0)
		return retval;

	return 0;
}

static void tcc_stop_hc(struct device *dev)
{
	struct tccohci_platform_data *inf;

	inf = dev->platform_data;

	if (inf->exit)
		inf->exit(dev);

	tcc_set_cken(0);
}


/*-------------------------------------------------------------------------*/

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */


/**
 * usb_hcd_tcc_probe - initialize tcc-based HCDs
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 */
int usb_hcd_tcc_probe (const struct hc_driver *driver, struct platform_device *pdev)
{
	int retval;
	struct usb_hcd *hcd;

	struct tccohci_platform_data *inf;

	inf = pdev->dev.platform_data;

	if (!inf)
		return -ENODEV;

	if (pdev->resource[1].flags != IORESOURCE_IRQ) {
		pr_debug ("resource[1] is not IORESOURCE_IRQ");
		return -ENOMEM;
	}

	hcd = usb_create_hcd (driver, &pdev->dev, "tcc");
	if (!hcd)
		return -ENOMEM;
	hcd->rsrc_start = pdev->resource[0].start;
	hcd->rsrc_len = pdev->resource[0].end - pdev->resource[0].start + 1;

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed");
		retval = -EBUSY;
		goto err1;
	}

	hcd->regs = (void __iomem *)(int)(hcd->rsrc_start);

	if ((retval = tcc_start_hc(&pdev->dev)) < 0) {
		pr_debug("tcc_start_hc failed");
		goto err2;
	}

	/* Select Power Management Mode */
	tcc_ohci_select_pmm(inf->port_mode);

	if (inf->power_budget)
		hcd->power_budget = inf->power_budget;

	ohci_hcd_init(hcd_to_ohci(hcd));

	retval = usb_add_hcd(hcd, pdev->resource[1].start, IRQF_DISABLED);
	if (retval == 0)
		return retval;

	tcc_stop_hc(&pdev->dev);
err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
err1:
	usb_put_hcd(hcd);
	return retval;
}


/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_tcc_remove - shutdown processing for tcc-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_tcc_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
void usb_hcd_tcc_remove (struct usb_hcd *hcd, struct platform_device *pdev)
{
	usb_remove_hcd(hcd);
	tcc_stop_hc(&pdev->dev);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
}

/*-------------------------------------------------------------------------*/

static int __devinit
ohci_tcc_start (struct usb_hcd *hcd)
{
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int ret;

	TurnOnUSBHostClocks();
	USBHostReset();

	/* The value of NDP in roothub_a is incorrect on this hardware */
	ohci->num_ports = 1;

	if ((ret = ohci_init(ohci)) < 0)
	{
		return ret;
	}

	if ((ret = ohci_run (ohci)) < 0)
	{
		err ("can't start %s", hcd->self.bus_name);
		ohci_stop (hcd);
		return ret;
	}

	return 0;
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ohci_tcc_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"TCC OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			ohci_irq,
	.flags =		HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.start =		ohci_tcc_start,
	.stop =			ohci_stop,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
#ifdef  CONFIG_PM
	.bus_suspend = 		ohci_bus_suspend,
	.bus_resume =		ohci_bus_resume,
#endif
	.start_port_reset =	ohci_start_port_reset,
};

/*-------------------------------------------------------------------------*/

static int ohci_hcd_tcc_drv_probe(struct platform_device *pdev)
{
	if (usb_disabled())
		return -ENODEV;

	return usb_hcd_tcc_probe(&ohci_tcc_hc_driver, pdev);
}

static int ohci_hcd_tcc_drv_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	usb_hcd_tcc_remove(hcd, pdev);
	platform_set_drvdata(pdev, NULL);
	return 0;
}

#ifdef	CONFIG_PM
static int ohci_hcd_tcc_drv_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);

	if (time_before(jiffies, ohci->next_statechange))
		msleep(5);
	ohci->next_statechange = jiffies;

	tcc_stop_hc(&pdev->dev);
	hcd->state = HC_STATE_SUSPENDED;
	pdev->dev.power.power_state = PMSG_SUSPEND;

	return 0;
}

static int ohci_hcd_tcc_drv_resume(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int status;

	if (time_before(jiffies, ohci->next_statechange))
		msleep(5);
	ohci->next_statechange = jiffies;

	if ((status = tcc_start_hc(&pdev->dev)) < 0)
		return status;

	pdev->dev.power.power_state = PMSG_ON;
	usb_hcd_resume_root_hub(hcd);

	return 0;
}
#endif


static struct platform_driver ohci_hcd_tcc_driver = {
	.probe		= ohci_hcd_tcc_drv_probe,
	.remove		= ohci_hcd_tcc_drv_remove,
#ifdef CONFIG_PM
	.suspend	= ohci_hcd_tcc_drv_suspend,
	.resume		= ohci_hcd_tcc_drv_resume,
#endif
	.driver		= {
		.name	= "tcc-ohci",
		.owner	= THIS_MODULE,
	},
};

static int __init ohci_hcd_tcc_init (void)
{
	pr_debug (DRIVER_INFO " (ohci-tcc)");
	pr_debug ("block sizes: ed %d td %d\n",
		sizeof (struct ed), sizeof (struct td));

	HwBCLKCTR |= HwBCLKCTR_UB_ON;	// always USB2.0 device clock ON
	USBHostReset();

	return platform_driver_register(&ohci_hcd_tcc_driver);
}

static void __exit ohci_hcd_tcc_cleanup (void)
{
	platform_driver_unregister(&ohci_hcd_tcc_driver);
}

module_init (ohci_hcd_tcc_init);
module_exit (ohci_hcd_tcc_cleanup);

